/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <pthread.h>
#include <sys/timerfd.h>
#include <sys/eventfd.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>
#ifdef DEBUG
#include <sys/prctl.h>
#include <stdio.h>
#endif
#include "misc.h"
#include "crypto.h"
#include "tcp.h"
#include "sctp_oo.h"
#include "sctp_om.h"
#include "udp.h"
#include "proto.h"
#include "rxtx.h"
#include "keyio.h"
#include "parse.h"
#include "tun.h"
#include "dns.h"
#include "helper.h"
#include "rtp.h"
#include "client.h"

#ifdef DEBUG
#define ERR(a) do							\
{									\
	fprintf(stderr,"Client abort: %s line %d\n",__FILE__,__LINE__);	\
	goto a;								\
} while(0)
#else
#define ERR(a)	goto a
#endif

struct client
{
	int curve;
	int sigcurve;
	int port;
	int proxy;
	int dstlen;
	struct config *cfg;
	void *host;
	void *rtp;
	void *dst;
	void *auth;
	void *user;
	struct crypto_conf *cnf;
	int (*checkserver)(void *data,int len,void *user);
};

struct connect
{
	struct client cln;
	struct rxtx rt;
	void *pub[2];
	int publen[2];
	int wfd;
	int pfd;
	pthread_t th1;
	pthread_t th2;
};

struct clnctx
{
	int total;
	void *h;
	struct connect c[0];
};

static int client_recv(int fd,struct client *c,void *bfr,int max)
{       
	switch(c->cfg->proto)
	{
	case TCP:
		return tcprecv(fd,bfr,max);
	case SCTP_OO:
		return sctpoorecv(fd,bfr,max);
	case SCTP_OM:
		return sctpomrecv(fd,bfr,max);
	case UDP:
	case UDPLITE:
		return udprecv(fd,bfr,max);
	case HELPER:
		return helperrecv(fd,bfr,max);
	case PKTHELPER:
		return helperpktrecv(fd,bfr,max);
	default:return -1;
	}
}

static int client_send(int fd,struct client *c,void *bfr,int len)
{
	switch(c->cfg->proto)
	{
	case TCP:
		return tcpsend(fd,bfr,len);
	case SCTP_OO:
		return sctpoosend(fd,bfr,len);
	case SCTP_OM:
		return sctpomsend(fd,c->dst,c->dstlen,bfr,len);
	case UDP:
	case UDPLITE:
		return udpsend(fd,bfr,len);
	case HELPER:
		return helpersend(fd,bfr,len);
	case PKTHELPER:
		return helperpktsend(fd,bfr,len);
	default:return -1;
	}
}

static struct conn *client_connect(void *h,struct client *c)
{
	int len;
	int tot;
	int l;
	int fd;
	int tfd;
	int siglen;
	int bits;
	int tmplen;
	int max;
	int minhdr;
	void *msg;
	void *sig;
	void *bfr;
	void *tmp;
	struct crypto_conn *conn;
	struct conn *res;
	struct crypto_x25519 *x25519;
	struct crypto_setup *setup;
	struct itimerspec it;
	struct pollfd p[3];

	switch(c->cfg->proto)
	{
	case TCP:
		if(c->proxy)fd=tcpproxyconnect(c->dst,c->host,c->port,
			c->cfg->conntimeout,c->cfg->global->keepalive,
			c->cfg->global->cefd);
		else if(c->host)fd=tcpsocks5connect(c->dst,c->host,c->port,
			c->cfg->conntimeout,c->cfg->global->keepalive,
			c->cfg->global->cefd);
		else fd=tcpconnect(c->dst,c->cfg->conntimeout,
			c->cfg->global->keepalive,c->cfg->sendbuf,
			c->cfg->maxseg,c->cfg->tcpdefer,c->cfg->global->tcpuser,
			c->cfg->global->cefd);
		break;
	case SCTP_OO:
		fd=sctpooconnect(c->dst,c->cfg->conntimeout,
			c->cfg->global->keepalive,c->cfg->sendbuf,
			c->cfg->maxseg,c->cfg->srcaddr,c->cfg->global->cefd);
		break;
	case SCTP_OM:
		fd=sctpomconnect(c->dst,c->cfg->conntimeout,
			c->cfg->global->keepalive,c->cfg->sendbuf,
			c->cfg->maxseg,c->cfg->srcaddr,c->cfg->global->cefd);
		break;
	case UDP:
		fd=udp_connect(h,c->dst,c->cfg->conntimeout,c->cfg->sendbuf,0,
			c->cfg->frag,c->rtp!=NULL,c->cfg->global->alt,
			c->cfg->global->portmin,c->cfg->global->portmax,
			c->cfg->global->cefd);
		break;
	case UDPLITE:
		fd=udp_connect(h,c->dst,c->cfg->conntimeout,c->cfg->sendbuf,
			c->rtp!=NULL?20:8,c->cfg->frag,c->rtp!=NULL,
			c->cfg->global->alt,c->cfg->global->portmin,
			c->cfg->global->portmax,c->cfg->global->cefd);
		break;
	case HELPER:
		fd=helperconnect(c->cfg->helpercmd,c->cfg->helpercnf,c->host,
			c->port,c->cfg->conntimeout,0,c->cfg->global->cefd);
		break;
	case PKTHELPER:
		fd=helperconnect(c->cfg->helpercmd,c->cfg->helpercnf,c->host,
			c->port,c->cfg->conntimeout,1,c->cfg->global->cefd);
		break;
	default:fd=-1;
		break;
	}

	if(fd==-1)ERR(err1);

	if(c->rtp)max=(c->cfg->setuplimit<1330)?1330:c->cfg->setuplimit;
	else max=c->cfg->setuplimit;
	if(!(bfr=malloc(max)))ERR(err2);

	if((tfd=timerfd_create(CLOCK_MONOTONIC,TFD_CLOEXEC|TFD_NONBLOCK))==-1)
		ERR(err3);
	memset(&it,0,sizeof(it));
	it.it_value.tv_sec=c->cfg->global->setuptimeout;
	if(timerfd_settime(tfd,0,&it,NULL))ERR(err4);

	p[0].fd=c->cfg->global->cefd;
	p[0].events=POLLIN;
	p[1].fd=tfd;
	p[1].events=POLLIN;
	p[2].fd=fd;
	p[2].events=POLLIN;

	if(!(x25519=crypto_initialrequest(h,c->cnf,&msg,&len,c->rtp!=NULL)))
		ERR(err4);
	if(c->rtp)
	{
		if(rtp_pack(c->rtp,msg,len,&tmp,&tmplen))ERR(err5);
		memcpy(bfr,tmp,tmplen);
		free(tmp);
		msg=bfr;
		len=tmplen;
	}
	if(client_send(fd,c,msg,len)!=len)ERR(err5);

	while(1)
	{
		if(poll(p,3,-1)<=0)continue;
		if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))ERR(err5);
		if(!(p[2].revents&POLLIN))continue;
		if((len=client_recv(fd,c,bfr,max))<=0)
		{
			if(len==-2)continue;
			ERR(err5);
		}
		if(c->rtp)
		{
			switch(c->cfg->proto)
			{
			case TCP:
			case SCTP_OO:
			case SCTP_OM:
				if(len<2)ERR(err5);
				if((tmplen=(((unsigned char *)bfr)[0]<<8)|
					((unsigned char *)bfr)[1])!=len-2)
						ERR(err5);
				memmove(bfr,bfr+2,tmplen);
				len=tmplen;
			}
			if(rtp_unpack(c->rtp,bfr,len,&tmp,&tmplen))ERR(err5);
			if(tmplen>max)
			{
				free(tmp);
				ERR(err5);
			}
			memcpy(bfr,tmp,tmplen);
			free(tmp);
			len=tmplen;
		}
		break;
	}

	if(crypto_initialresponse(h,c->cnf,x25519,bfr,len,c->rtp!=NULL))
		ERR(err5);
	switch(c->cfg->proto)
	{
	case UDP:
	case UDPLITE:
	case PKTHELPER:
		if(!(conn=crypto_initialderive(h,c->cnf,x25519,1,0,
			c->rtp!=NULL,c->cfg->proto==UDPLITE?69:0,c->cfg->small,
			c->cfg->strip,c->cfg->global->alt)))ERR(err5);
		break;

	default:if(!(conn=crypto_initialderive(h,c->cnf,x25519,0,0,
			c->rtp!=NULL,0,c->cfg->small,c->cfg->strip,
			c->cfg->global->alt)))ERR(err5);
		break;
	}
	minhdr=crypto_hdrsize(h,conn);
	if(!(setup=crypto_realrequest(h,c->cnf,c->curve,&msg,&len)))ERR(err6);
	if(!(msg=crypto_encode(h,conn,CONTROL,msg,len,0,&len,c->rtp!=NULL)))
		ERR(err7);
	if(c->rtp)
	{
		if(rtp_pack(c->rtp,msg,len,&tmp,&tmplen))ERR(err8);
		memcpy(bfr,tmp,tmplen);
		free(msg);
		msg=tmp;
		len=tmplen;
	}
	if(client_send(fd,c,msg,len)!=len)ERR(err8);
	free(msg);

	switch(c->cfg->proto)
	{
	case SCTP_OM:
	case UDP:
	case UDPLITE:
	case PKTHELPER:
rep1:
		while(poll(p,3,-1)<=0);
		if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))ERR(err7);
		if(!(p[2].revents&POLLIN))goto rep1;
		if((len=client_recv(fd,c,bfr,max))<=minhdr)
		{
			if(len==-2)goto rep1;
			ERR(err7);
		}
		if(c->rtp)
		{
			switch(c->cfg->proto)
			{
			case TCP:
			case SCTP_OO:
			case SCTP_OM:
				if(len<2)ERR(err5);
				if((tmplen=(((unsigned char *)bfr)[0]<<8)|
					((unsigned char *)bfr)[1])!=len-2)
						ERR(err7);
				memmove(bfr,bfr+2,tmplen);
				len=tmplen;
			}
			if(rtp_unpack(c->rtp,bfr,len,&tmp,&tmplen))ERR(err5);
			if(tmplen>max)
			{
				free(tmp);
				ERR(err5);
			}
			memcpy(bfr,tmp,tmplen);
			free(tmp);
			len=tmplen;
		}
		if(crypto_unmask(conn,bfr,len,&tot,&bits)||tot!=len)ERR(err7);
		break;

	default:if(c->rtp)
		{
			for(len=0;len<2;len+=l)
			{
				if(poll(p,3,-1)<=0)continue;
				if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))
					ERR(err7);
				if(!(p[2].revents&POLLIN))continue;
				if((l=client_recv(fd,c,bfr+len,2-len))<=0)
					ERR(err7);
			}
			tot=(((unsigned char *)bfr)[0]<<8)|
				((unsigned char *)bfr)[1];
			if(tot<200||tot>max)ERR(err7);
			for(len=0;len<tot;len+=l)
			{
				if(poll(p,3,-1)<=0)continue;
				if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))
					ERR(err7);
				if(!(p[2].revents&POLLIN))continue;
				if((l=client_recv(fd,c,bfr+len,tot-len))<=0)
					ERR(err7);
			}
			if(rtp_unpack(c->rtp,bfr,len,&tmp,&tmplen))ERR(err7);
			if(tmplen>max)
			{
				free(tmp);
				ERR(err7);
			}
			memcpy(bfr,tmp,tmplen);
			free(tmp);
			tot=len=tmplen;
		}
		else for(len=0;len<minhdr-2;len+=l)
		{
			if(poll(p,3,-1)<=0)continue;
			if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))
				ERR(err7);
			if(!(p[2].revents&POLLIN))continue;
			if((l=client_recv(fd,c,bfr+len,minhdr-2-len))<=0)
				ERR(err7);
		}
		if(crypto_unmask(conn,bfr,len,&tot,&bits)||tot<=minhdr||
			tot>max)ERR(err7);
		for(;len<tot;len+=l)
		{
			if(poll(p,3,-1)<=0)continue;
			if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))
				ERR(err7);
			if(!(p[2].revents&POLLIN))continue;
			if((l=client_recv(fd,c,bfr+len,tot-len))<=0)ERR(err7);
		}
		break;
	}
	if(!(msg=crypto_decode(h,conn,&l,bfr,len,&len,NULL))||l!=CONTROL)
		ERR(err7);
	if(len>c->cfg->setuplimit)ERR(err8);
	if(crypto_realresponse(h,c->cnf,setup,msg,len,&sig,&siglen))ERR(err8);
	crypto_memclear(h,msg,len);
	free(msg);

	if(c->checkserver(sig,siglen,c->user))ERR(err7);

	crypto_freeconn(h,conn);
	switch(c->cfg->proto)
	{
	case UDP:
	case UDPLITE:
	case PKTHELPER:
		if(!(conn=crypto_derivekey(h,c->cnf,setup,1,c->rtp!=NULL,
			c->cfg->proto==UDPLITE?69:0,c->cfg->small,c->cfg->strip,
			c->cfg->global->alt)))ERR(err7);
		break;

	default:if(!(conn=crypto_derivekey(h,c->cnf,setup,0,c->rtp!=NULL,
			0,c->cfg->small,c->cfg->strip,c->cfg->global->alt)))
				ERR(err7);
		break;
	}
	minhdr=crypto_hdrsize(h,conn);

	if(crypto_authrequest(h,c->cnf,setup,c->auth,&msg,&len,
		c->cfg->paranoia))ERR(err7);
	if(!(msg=crypto_encode(h,conn,CONTROL,msg,len,0,&len,c->rtp!=NULL)))
		ERR(err7);
	if(c->rtp)
	{
		if(rtp_pack(c->rtp,msg,len,&tmp,&tmplen))ERR(err8);
		free(msg);
		msg=tmp;
		len=tmplen;
	}
	if(client_send(fd,c,msg,len)!=len)ERR(err8);
	free(msg);

	switch(c->cfg->proto)
	{
	case SCTP_OM:
	case UDP:
	case UDPLITE:
	case PKTHELPER:
rep2:
		while(poll(p,3,-1)<=0);
		if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))ERR(err7);
		if(!(p[2].revents&POLLIN))goto rep2;
		if((len=client_recv(fd,c,bfr,max))<=minhdr)
		{
			if(len==-2)goto rep2;
			ERR(err7);
		}
		if(c->rtp)
		{
			switch(c->cfg->proto)
			{
			case TCP:
			case SCTP_OO:
			case SCTP_OM:
				if(len<2)ERR(err5);
				if((tmplen=(((unsigned char *)bfr)[0]<<8)|
					((unsigned char *)bfr)[1])!=len-2)
						ERR(err7);
				memmove(bfr,bfr+2,tmplen);
				len=tmplen;
			}
			if(rtp_unpack(c->rtp,bfr,len,&tmp,&tmplen))ERR(err7);
			if(tmplen>max)
			{
				free(tmp);
				ERR(err7);
			}
			memcpy(bfr,tmp,tmplen);
			free(tmp);
			len=tmplen;
		}
		if(crypto_unmask(conn,bfr,len,&tot,&bits)||tot!=len)ERR(err7);
		break;

	default:if(c->rtp)
		{
			for(len=0;len<2;len+=l)
			{
				if(poll(p,3,-1)<=0)continue;
				if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))
					ERR(err7);
				if(!(p[2].revents&POLLIN))continue;
				if((l=client_recv(fd,c,bfr+len,2-len))<=0)
					ERR(err7);
			}
			tot=(((unsigned char *)bfr)[0]<<8)|
				((unsigned char *)bfr)[1];
			if(tot<200||tot>max)ERR(err7);
			for(len=0;len<tot;len+=l)
			{
				if(poll(p,3,-1)<=0)continue;
				if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))
					ERR(err7);
				if(!(p[2].revents&POLLIN))continue;
				if((l=client_recv(fd,c,bfr+len,tot-len))<=0)
					ERR(err7);
			}
			if(rtp_unpack(c->rtp,bfr,len,&tmp,&tmplen))ERR(err7);
			if(tmplen>max)
			{
				free(tmp);
				ERR(err7);
			}
			memcpy(bfr,tmp,tmplen);
			free(tmp);
			tot=len=tmplen;
		}
		else for(len=0;len<minhdr-2;len+=l)
		{
			if(poll(p,3,-1)<=0)continue;
			if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))
				ERR(err7);
			if(!(p[2].revents&POLLIN))continue;
			if((l=client_recv(fd,c,bfr+len,minhdr-2-len))<=0)
				ERR(err7);
		}
		if(crypto_unmask(conn,bfr,len,&tot,&bits)||tot<=minhdr||
			tot>max)ERR(err7);
		for(;len<tot;len+=l)
		{
			if(poll(p,3,-1)<=0)continue;
			if((p[0].revents&POLLIN)||(p[1].revents&POLLIN))
				ERR(err7);
			if(!(p[2].revents&POLLIN))continue;
			if((l=client_recv(fd,c,bfr+len,tot-len))<=0)ERR(err7);
		}
		break;
	}
	if(!(msg=crypto_decode(h,conn,&l,bfr,len,&len,NULL))||l!=CONTROL)
		ERR(err7);
	if(len>c->cfg->setuplimit)ERR(err8);
	if(crypto_authresponse(h,c->cnf,setup,msg,len))ERR(err8);
	crypto_memclear(h,msg,len);
	free(msg);

	if(c->cfg->paranoia)
	{
		crypto_freeconn(h,conn);
		switch(c->cfg->proto)
		{
		case UDP:
		case UDPLITE:
		case PKTHELPER:
			if(!(conn=crypto_authderive(h,c->cnf,setup,1,
				c->rtp!=NULL,c->cfg->proto==UDPLITE?69:0,
				c->cfg->small,c->cfg->strip,
				c->cfg->global->alt)))ERR(err7);
			break;

		default:if(!(conn=crypto_authderive(h,c->cnf,setup,0,
				c->rtp!=NULL,0,c->cfg->small,c->cfg->strip,
				c->cfg->global->alt)))ERR(err7);
			break;
		}
	}

	if(!(res=malloc(sizeof(struct conn))))ERR(err7);
	res->proto=c->cfg->proto;
	res->rtp=c->rtp;
	res->fd=fd;
	res->rekey=c->cfg->rekey;
	res->noop=c->cfg->noop;
	res->curve=crypto_getcurve(h,c->cnf,setup);
	res->keepalive=c->cfg->global->keepalive;
	res->dst=c->dst;
	res->dstlen=c->dstlen;
	res->conn[0]=conn;
	res->conn[1]=conn;

	crypto_setupfree(h,c->cnf,setup,&res->keys);
	crypto_initialfree(h,c->cnf,x25519);
	close(tfd);
	crypto_memclear(h,bfr,c->cfg->setuplimit);
	free(bfr);
	return res;

err8:	crypto_memclear(h,msg,len);
	free(msg);
err7:	crypto_setupfree(h,c->cnf,setup,NULL);
err6:	if(conn)crypto_freeconn(h,conn);
err5:	crypto_initialfree(h,c->cnf,x25519);
err4:	close(tfd);
err3:	crypto_memclear(h,bfr,c->cfg->setuplimit);
	free(bfr);
err2:	switch(c->cfg->proto)
	{
	case TCP:
		tcpclose(fd);
		break;
	case SCTP_OO:
		sctpooclose(fd);
		break;
	case SCTP_OM:
		sctpomclose(fd);
		break;
	case UDP:
	case UDPLITE:
		udpclose(fd);
		break;
	case HELPER:
	case PKTHELPER:
		helperclose(fd);
		break;
	default:close(fd);
		break;
	}
err1:	return NULL;
}

static int checkserver(void *data,int len,void *user)
{
	struct connect *c=user;

	if(!data)return -1;
	if(c->pub[0])if(len==c->publen[0]&&!memcmp(data,c->pub[0],len))return 0;
	if(c->pub[1])if(len==c->publen[1]&&!memcmp(data,c->pub[1],len))return 0;
	return -1;
}

static void notify(void *h,void *user,int update)
{
	uint64_t dummy;
	struct connect *c=user;

	if(update)
	{
		if(c->pub[0])free(c->pub[0]);
		if(c->pub[1])free(c->pub[1]);
		c->pub[0]=loadpub(h,c->cln.cfg->authfile[0],&c->publen[0]);
		c->pub[1]=loadpub(h,c->cln.cfg->authfile[1],&c->publen[1]);
	}

	dummy=1;
	dummy=write(c->wfd,&dummy,sizeof(dummy));
}

static void *connector(void *data)
{
	int tfd;
	uint64_t dummy;
	void *h;
	void *server;
	int port;
	struct conn *conn;
	struct connect *c=data;
	struct pollfd p[3];
	struct itimerspec it;
	sigset_t set;

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,NULL);
#ifdef DEBUG
	prctl(PR_SET_NAME,"connector",0,0,0);
#endif

	if((tfd=timerfd_create(CLOCK_MONOTONIC,TFD_CLOEXEC|TFD_NONBLOCK))==-1)
		goto err1;
	memset(&it,0,sizeof(it));

	if(!(h=crypto_thread_init(c->cln.cfg->global->glob)))goto err2;

	if(c->cln.cfg->proto==HELPER||c->cln.cfg->proto==PKTHELPER)
	{
		c->cln.host=c->cln.cfg->server;
		c->cln.port=c->cln.cfg->port;
		c->cln.proxy=0;
		c->cln.dst=NULL;
	}
	else if(c->cln.cfg->sockshost)
	{
		server=c->cln.cfg->sockshost;
		port=c->cln.cfg->socksport;
		c->cln.host=c->cln.cfg->server;
		c->cln.port=c->cln.cfg->port;
		c->cln.proxy=0;
	}
	else if(c->cln.cfg->proxyhost)
	{
		server=c->cln.cfg->proxyhost;
		port=c->cln.cfg->proxyport;
		c->cln.host=c->cln.cfg->server;
		c->cln.port=c->cln.cfg->port;
		c->cln.proxy=1;
	}
	else
	{
		server=c->cln.cfg->server;
		port=c->cln.cfg->port;
		c->cln.host=NULL;
		c->cln.port=0;
		c->cln.proxy=0;
	}

	p[0].fd=c->cln.cfg->global->cefd;
	p[0].events=POLLIN;
	p[1].fd=c->wfd;
	p[1].events=POLLIN;
	p[2].fd=tfd;
	p[2].events=POLLIN;

	while(1)
	{
		if(poll(p,3,-1)<=0)continue;
		if(p[0].revents&POLLIN)break;
		if(!(p[1].revents&POLLIN)&&!(p[2].revents&POLLIN))continue;
		if(read(p[1].fd,&dummy,sizeof(dummy))!=sizeof(dummy)&&
			read(p[2].fd,&dummy,sizeof(dummy))!=sizeof(dummy))
				continue;

		it.it_value.tv_sec=0;
		timerfd_settime(p[2].fd,0,&it,NULL);

		it.it_value.tv_sec=c->cln.cfg->delay;

		if(c->cln.cfg->proto==HELPER||c->cln.cfg->proto==PKTHELPER||
			(c->cln.dst=lookup(server,port,c->cln.cfg->dnstimeout,
				c->cln.cfg->prefv4,p[0].fd,&c->cln.dstlen)))
		{
			if(c->cln.cfg->rtp)
			{
				switch(c->cln.cfg->proto)
				{
				case UDP:
				case UDPLITE:
				case PKTHELPER:
					c->cln.rtp=rtp_init(h,1);
					break;
				default:c->cln.rtp=rtp_init(h,0);
				}
				if(!c->cln.rtp)
				{
					if(c->cln.dst)free(c->cln.dst);
					goto skip;
				}
			}
			else c->cln.rtp=NULL;
			if((conn=client_connect(h,&c->cln)))
			{
				switch(conn->proto)
				{
				case UDP:
				case UDPLITE:
				case PKTHELPER:
					conn->packet=1;
					break;
				default:conn->packet=0;
					break;
				}
				if(write(c->pfd,conn,sizeof(struct conn))!=
					sizeof(struct conn))
				{
					switch(conn->proto)
					{
					case TCP:
						tcpclose(conn->fd);
						break;
					case SCTP_OO:
						sctpooclose(conn->fd);
						break;
					case SCTP_OM:
						sctpomclose(conn->fd);
						break;
					case UDP:
					case UDPLITE:
						udpclose(conn->fd);
						break;
					case HELPER:
					case PKTHELPER:
						helperclose(conn->fd);
						break;
					default:close(conn->fd);
					}
					if(conn->dst)free(conn->dst);
					crypto_keysfree(h,conn->keys);
					crypto_freepair(h,conn->conn,3);
					if(conn->rtp)rtp_exit(conn->rtp);
				}
				else it.it_value.tv_sec=0;
				free(conn);
			}
			else
			{
				if(c->cln.rtp)rtp_exit(c->cln.rtp);
				if(c->cln.dst)free(c->cln.dst);
			}
		}

skip:		timerfd_settime(p[2].fd,0,&it,NULL);
	}

	crypto_thread_exit(h);
err2:	close(tfd);
err1:	pthread_exit(NULL);
}

COLD static int doexec(void *script,void *mode,void *dev)
{
	int status;
	pid_t pid;

	if(script)switch((pid=fork()))
	{
	case -1:return -1;
	case 0:	munlockall();
		execl(script,script,mode,dev,NULL);
		exit(255);
	default:if(waitpid(pid,&status,0)!=pid||status)return -1;
		break;
	}
	return 0;
}

COLD void *client_init(struct global *g,void *pass,int plen)
{
	int i;
	int j;
	int ncnt;
	int mask;
	void *h;
	void *key;
	struct clnctx *ctx;
	struct config *c;

	if(!(h=crypto_thread_init(g->glob)))ERR(err1);
	for(ncnt=0,c=g->cfg;c;c=c->next)if(c->mode==CLIENT&&c->connection)
		ncnt++;
	if(!(ctx=malloc(sizeof(struct clnctx)+ncnt*sizeof(struct connect))))
		ERR(err2);
	memset(ctx,0,sizeof(struct clnctx)+ncnt*sizeof(struct connect));
	ctx->total=ncnt;
	ctx->h=h;
	for(c=g->cfg,i=0;i<ncnt;i++,c=c->next)
	{
		while(c)if(c->mode!=CLIENT||!c->connection)c=c->next;
		else break;
		ctx->c[i].wfd=-1;
		ctx->c[i].pfd=-1;
		ctx->c[i].cln.cfg=c;
		ctx->c[i].cln.user=&ctx->c[i];
		ctx->c[i].cln.checkserver=checkserver;
		ctx->c[i].rt.c=c;
		ctx->c[i].rt.g=g;
		ctx->c[i].rt.pfd[0]=-1;
		ctx->c[i].rt.pfd[1]=-1;
		ctx->c[i].rt.user=&ctx->c[i];
		ctx->c[i].rt.notify=notify;
	}
	for(i=0;i<ncnt;i++)
	{
		mask=0x7f;
		if(!ctx->c[i].cln.cfg->brainpool)mask&=0x0f;
		if(!ctx->c[i].cln.cfg->secp)mask&=0x71;
		if(!ctx->c[i].cln.cfg->x25519)mask&=0x7e;
		if(ctx->c[i].cln.cfg->minbits>256)mask&=0x7c;
		if(ctx->c[i].cln.cfg->minbits>384)mask&=0x48;
		if(ctx->c[i].cln.cfg->minbits>512)mask&=0x08;
		if(ctx->c[i].cln.cfg->minbits>521)mask&=0x00;
		if(ctx->c[i].cln.cfg->maxbits)
		{
			if(ctx->c[i].cln.cfg->maxbits<521)mask&=0x77;
			if(ctx->c[i].cln.cfg->maxbits<512)mask&=0x37;
			if(ctx->c[i].cln.cfg->maxbits<384)mask&=0x13;
			if(ctx->c[i].cln.cfg->maxbits<256)mask&=0x00;
		}
		for(ctx->c[i].cln.curve=-1,j=0x40;ctx->c[i].cln.curve==-1&&j;
			j>>=1)if(mask&j)switch(j)
		{
		case 0x40:
			ctx->c[i].cln.curve=CRYPTO_BRAINPOOLP512R1;
			break;
		case 0x20:
			ctx->c[i].cln.curve=CRYPTO_BRAINPOOLP384R1;
			break;
		case 0x10:
			ctx->c[i].cln.curve=CRYPTO_BRAINPOOLP256R1;
			break;
		case 0x08:
			ctx->c[i].cln.curve=CRYPTO_SECP521R1;
			break;
		case 0x04:
			ctx->c[i].cln.curve=CRYPTO_SECP384R1;
			break;
		case 0x02:
			ctx->c[i].cln.curve=CRYPTO_SECP256R1;
			break;
		case 0x01:
			ctx->c[i].cln.curve=CRYPTO_X25519;
			break;
		}
		if(ctx->c[i].cln.curve==-1)ERR(err3);
		mask=0x3f;
		if(!ctx->c[i].cln.cfg->sigbrainpool)mask&=0x07;
		if(!ctx->c[i].cln.cfg->sigsecp)mask&=0x38;
		if(ctx->c[i].cln.cfg->sigminbits>256)mask&=0x36;
		if(ctx->c[i].cln.cfg->sigminbits>384)mask&=0x24;
		if(ctx->c[i].cln.cfg->sigminbits>512)mask&=0x04;
		if(ctx->c[i].cln.cfg->sigminbits>521)mask&=0x00;
		if(ctx->c[i].cln.cfg->sigmaxbits)
		{
			if(ctx->c[i].cln.cfg->sigmaxbits<521)mask&=0x3b;
			if(ctx->c[i].cln.cfg->sigmaxbits<512)mask&=0x1b;
			if(ctx->c[i].cln.cfg->sigmaxbits<384)mask&=0x09;
			if(ctx->c[i].cln.cfg->sigmaxbits<256)mask&=0x00;
		}
		for(ctx->c[i].cln.sigcurve=-1,j=0x20;
			ctx->c[i].cln.sigcurve==-1&&j;j>>=1)if(mask&j)switch(j)
		{
		case 0x20:
			ctx->c[i].cln.sigcurve=CRYPTO_BRAINPOOLP512R1;
			break;
		case 0x10:
			ctx->c[i].cln.sigcurve=CRYPTO_BRAINPOOLP384R1;
			break;
		case 0x08:
			ctx->c[i].cln.sigcurve=CRYPTO_BRAINPOOLP256R1;
			break;
		case 0x04:
			ctx->c[i].cln.sigcurve=CRYPTO_SECP521R1;
			break;
		case 0x02:
			ctx->c[i].cln.sigcurve=CRYPTO_SECP384R1;
			break;
		case 0x01:
			ctx->c[i].cln.sigcurve=CRYPTO_SECP256R1;
			break;
		}
		if(ctx->c[i].cln.sigcurve==-1)ERR(err3);
		if(!(key=crypto_genkey(h,ctx->c[i].cln.sigcurve)))ERR(err3);
		if(!(ctx->c[i].cln.cnf=crypto_config(h,key,
		    ctx->c[i].cln.cfg->setuplimit,ctx->c[i].cln.cfg->x25519,
		    ctx->c[i].cln.cfg->secp,ctx->c[i].cln.cfg->brainpool,
		    ctx->c[i].cln.cfg->minbits,ctx->c[i].cln.cfg->maxbits,
		    ctx->c[i].cln.cfg->sigsecp,ctx->c[i].cln.cfg->sigbrainpool,
		    ctx->c[i].cln.cfg->sigminbits,ctx->c[i].cln.cfg->sigmaxbits,
		    g->maxdata,g->datapad)))
		{
			crypto_free_key(h,key);
			ERR(err3);
		}
		if(!(ctx->c[i].cln.auth=loadkey(h,ctx->c[i].cln.cfg->keyfile,
			pass,plen)))ERR(err3);
		if(pipe2(ctx->c[i].rt.pfd,O_CLOEXEC|O_NONBLOCK|O_DIRECT))
			ERR(err3);
		if(!(ctx->c[i].rt.tun=opentun()))ERR(err3);
		ctx->c[i].pfd=ctx->c[i].rt.pfd[1];
		if((ctx->c[i].wfd=eventfd(0,EFD_CLOEXEC|EFD_NONBLOCK))==-1)
			ERR(err3);
		ctx->c[i].pub[0]=loadpub(h,ctx->c[i].cln.cfg->authfile[0],
			&ctx->c[i].publen[0]);
		if(ctx->c[i].cln.cfg->authfile[1])ctx->c[i].pub[1]=
			loadpub(h,ctx->c[i].cln.cfg->authfile[1],
				&ctx->c[i].publen[1]);
		if(!ctx->c[i].pub[0]&&!ctx->c[i].pub[1])ERR(err3);
	}
	return ctx;

err3:	for(i=0;i<ncnt;i++)
	{
		if(ctx->c[i].cln.cnf)crypto_free(h,ctx->c[i].cln.cnf);
		if(ctx->c[i].cln.auth)crypto_free_key(h,ctx->c[i].cln.auth);
		if(ctx->c[i].rt.pfd[0]!=-1)close(ctx->c[i].rt.pfd[0]);
		if(ctx->c[i].rt.pfd[1]!=-1)close(ctx->c[i].rt.pfd[1]);
		if(ctx->c[i].rt.tun)closetun(ctx->c[i].rt.tun);
		if(ctx->c[i].wfd!=-1)close(ctx->c[i].wfd);
		if(ctx->c[i].pub[0])free(ctx->c[i].pub[0]);
		if(ctx->c[i].pub[1])free(ctx->c[i].pub[1]);
	}
	free(ctx);
err2:	crypto_thread_exit(h);
err1:	return NULL;
}

COLD int client_start(void *data,int exec)
{
	int i;
	uint64_t dummy;
	struct clnctx *ctx=data;

	for(i=0;i<ctx->total;i++)
	{
		if(pthread_create(&ctx->c[i].th1,NULL,rxtx,&ctx->c[i].rt))
			goto err1;
		if(pthread_create(&ctx->c[i].th2,NULL,connector,&ctx->c[i]))
		{
			pthread_cancel(ctx->c[i].th1);
			pthread_join(ctx->c[i].th1,NULL);
			goto err1;
		}
		if(exec)if(doexec(ctx->c[i].rt.c->script,"up",
			ctx->c[i].rt.tun->dev))
		{
			pthread_cancel(ctx->c[i].th1);
			pthread_join(ctx->c[i].th1,NULL);
			pthread_cancel(ctx->c[i].th2);
			pthread_join(ctx->c[i].th2,NULL);
			goto err1;
		}
	}
	for(i=0;i<ctx->total;i++)
	{
		dummy=1;
		dummy=write(ctx->c[i].wfd,&dummy,sizeof(dummy));
	}
	return 0;

err1:	for(i--;i>=0;i--)
	{
		doexec(ctx->c[i].rt.c->script,"down",ctx->c[i].rt.tun->dev);
		pthread_cancel(ctx->c[i].th1);
		pthread_join(ctx->c[i].th1,NULL);
		pthread_cancel(ctx->c[i].th2);
		pthread_join(ctx->c[i].th2,NULL);
	}
	return -1;
}

COLD void client_stop(void *data,int exec)
{
	int i;
	struct clnctx *ctx=data;

	for(i=0;i<ctx->total;i++)
	{
		if(exec)doexec(ctx->c[i].rt.c->script,"down",
			ctx->c[i].rt.tun->dev);
		pthread_join(ctx->c[i].th1,NULL);
		pthread_join(ctx->c[i].th2,NULL);
	}
}

COLD void client_exit(void *data)
{
	int i;
	struct pollfd p;
	struct clnctx *ctx=data;
	struct conn c;

	for(i=0;i<ctx->total;i++)
	{
		crypto_free_key(ctx->h,ctx->c[i].cln.auth);
		crypto_free(ctx->h,ctx->c[i].cln.cnf);
                p.fd=ctx->c[i].rt.pfd[0];
		while(poll(&p,1,0)==1)if(!(p.revents&POLLIN))break;
		else if(read(p.fd,&c,sizeof(c))!=sizeof(c))break;
		else
		{
			switch(c.proto)
			{
			case TCP:
				tcpclose(c.fd);
				break;
			case SCTP_OO:
				sctpooclose(c.fd);
				break;
			case SCTP_OM:
				sctpomclose(c.fd);
				break;
			case UDP:
			case UDPLITE:
				udpclose(c.fd);
				break;
			case HELPER:
			case PKTHELPER:
				helperclose(c.fd);
				break;
			default:close(c.fd);
				break;
			}
			crypto_freepair(ctx->h,c.conn,3);
		}
		close(ctx->c[i].rt.pfd[0]);
		close(ctx->c[i].rt.pfd[1]);
		closetun(ctx->c[i].rt.tun);
		close(ctx->c[i].wfd);
		if(ctx->c[i].pub[0])free(ctx->c[i].pub[0]);
		if(ctx->c[i].pub[1])free(ctx->c[i].pub[1]);
	}
	crypto_thread_exit(ctx->h);
	free(ctx);
}

void client_keyupdate(void *data)
{
	int i;
	void *key;
	struct clnctx *ctx=data;

	for(i=0;i<ctx->total;i++)
	{
		if(!(key=crypto_genkey(ctx->h,ctx->c[i].cln.sigcurve)))continue;
		crypto_reconfig(ctx->h,ctx->c[i].cln.cnf,key);
	}
}
