/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#ifndef NO_ELLIGATOR
#include <elligator-2-curve25519.h>
#endif
#include <sys/times.h>
#include <sys/time.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <endian.h>
#include "usicrypt.h"
#include "misc.h"
#include "crypto.h"

#define CRYPTO_REQUEST	0x00
#define CRYPTO_RESPONSE	0x01
#define CRYPTO_AUTHREQ	0x02
#define CRYPTO_AUTHANON	0x03
#define CRYPTO_AUTHOK	0x04
#define CRYPTO_AUTHFAIL	0x05
#define CRYPTO_NOOP	0x06
#define CRYPTO_REKEYREQ	0x07
#define CRYPTO_REKEYACK	0x08
#define CRYPTO_FINMSG	0x09
#define CRYPTO_NEWKEY	0x0a
#define CRYPTO_REVOKE	0x0b

struct crypto_x25519
{
	void *key;
	void *pub;
	void *shared;
	unsigned char *der;
	int dlen;
	int slen;
	int lcllen;
	int rmtlen;
	unsigned char *lcl;
	unsigned char *rmt;
	unsigned char data[0];
};

struct crypto_keys
{
	int mode;
	void *key;
	void *pub;
	unsigned char data[0];
};

struct crypto_setup
{
	struct crypto_keys *ck;
	void *key;
	void *pub;
	void *ltk;
	void *sig;
	void *authsig;
	void *lclkey;
	void *rmtpub;
	int lcllen;
	int rmtlen;
	int authlen;
	int siglen;
	int authsiglen;
	int rekeylen;
	int curve;
	int anonymous;
	unsigned char *lcl;
	unsigned char *rmt;
	unsigned char *auth;
	unsigned char *rekey;
	unsigned char salt1[64];
	unsigned char salt2[64];
	unsigned char data[0];
};

struct crypto_nonce
{
	uint32_t msb;
	uint64_t lsb;
};

struct crypto_conn
{
	void *cfb8;
	void *rxgcm;
	void *txgcm;
	struct crypto_nonce *nonce;
	int maxmsg;
	int maxpad;
	int maxdata;
	int datapad;
	int packet;
	int minmsg;
	int small;
	int strip;
	int hdr;
	int off;
	int base;
	int alt;
};

struct crypto_conf
{
	unsigned int setup_max:16;
	unsigned int maxdata:16;
	unsigned int datapad:16;
	unsigned int enable_x25519:1;
	unsigned int enable_secp:1;
	unsigned int enable_brainpool:1;
	unsigned int enable_sig_secp:1;
	unsigned int enable_sig_brainpool:1;
	unsigned int minbits:10;
	unsigned int maxbits:10;
	unsigned int sig_minbits:10;
	unsigned int sig_maxbits:10;
	void *ltk;
};

struct crypto_rekeydata
{
	int state;
	int curve;
	int plen;
	int remote;
	void *key;
	void *pub;
	void *rmt;
	unsigned char salt1[64];
	unsigned char salt2[64];
	unsigned char salt[64];
};

struct crypto_pktkey
{
	int curve;
	void *key;
	unsigned char salt[32];
};

static inline void init_nonce(void *val,struct crypto_nonce *n,int small,
	void *bits)
{
	if(!small)
	{
		n->msb=be32toh(*((uint32_t *)val));
		n->lsb=be64toh(*((uint64_t *)(val+4)));
	}
	else
	{
		n->msb=0;
		n->lsb=be64toh(*((uint64_t *)val));
	}
	if(bits)*((uint64_t *)bits)=1;
}

HOT static inline void inc_set_nonce(void *dst,struct crypto_nonce *n,int small)
{
	if(UNLIKELY(!(++(n->lsb))))n->msb++;
	if(LIKELY(!small))
	{
		*((uint32_t *)dst)=htobe32(n->msb);
		*((uint64_t *)(dst+4))=htobe64(n->lsb);
	}
	else *((uint64_t *)dst)=htobe64(n->lsb);
}

HOT static inline int inc_cmp_nonce(void *ref,struct crypto_nonce *n,int small,
	long long *delta)
{
	if(UNLIKELY(!(++(n->lsb))))n->msb++;
	if(LIKELY(!small))
	{
		if(UNLIKELY(be32toh(*((uint32_t *)ref))!=n->msb))return -1;
		if(UNLIKELY(be64toh(*((uint64_t *)(ref+4)))!=n->lsb))return -1;
	}
	else if(UNLIKELY(be64toh(*((uint64_t *)ref))!=n->lsb))return -1;
	if(delta)*delta=1LL;
	return 0;
}

HOT static inline int check_set_nonce(void *ref,struct crypto_nonce *n,
	void *bits,int small,long long *delta)
{
	uint32_t msb;
	uint32_t hmem;
	uint64_t lsb;
	uint64_t lmem;
	uint64_t *prev=bits;

	if(LIKELY(!small))
	{
		hmem=be32toh(*((uint32_t *)ref));
		lmem=be64toh(*((uint64_t *)(ref+4)));
		msb=hmem-n->msb;
	}
	else
	{
		hmem=0;
		lmem=be64toh(*((uint64_t *)ref));
		msb=0;
	}
	if((lsb=lmem-n->lsb)>lmem)msb--;
	if(!msb&&lsb&&lsb<=0xffffffffffff)
	{
		n->lsb=lmem;
		n->msb=hmem;
		if(UNLIKELY(lsb>63))*prev=1;
		else
		{
			*prev<<=lsb;
			*prev|=1;
		}
	}
	else if(msb==-1&&lsb>-64&&LIKELY(!(*prev&(1ULL<<(-lsb)))))
		*prev|=1ULL<<(-lsb);
	else return -1;
	if(delta)*delta=(long long)lsb;
	return 0;
}

int crypto_random(void *h,void *data,int len)
{
	int r=-1;
	struct
	{
		int i;
		clock_t result;
		struct tms tms;
		struct timeval tv;
	} key;
	unsigned char in[32];
	unsigned char out[32];

	gettimeofday(&key.tv,NULL);
	key.result=times(&key.tms);

	for(key.i=0;key.i<len;key.i+=32)
	{
		if(usicrypt_random(h,in,32))goto err1;
		if(usicrypt_hmac(h,USICRYPT_SHA256,in,32,&key,sizeof(key),out))
			goto err1;
		memcpy(data+key.i,out,len-key.i>32?32:len-key.i);
	}
	r=0;
err1:	usicrypt_memclear(h,in,sizeof(in));
	usicrypt_memclear(h,out,sizeof(out));
	return r;
}

COLD void *crypto_init(void)
{
	void *h;

	if(!(h=usicrypt_global_init(NULL,NULL)))return NULL;
#ifndef NO_ELLIGATOR
	Elligator_2_Curve25519_initialize();
#endif
	return h;
}

COLD void crypto_exit(void *h)
{
	usicrypt_global_exit(h);
#ifndef NO_ELLIGATOR
	Elligator_2_Curve25519_destroy();
#endif
}

COLD void *crypto_thread_init(void *h)
{
	return usicrypt_thread_init(h);
}

COLD void crypto_thread_exit(void *h)
{
	usicrypt_thread_exit(h);	
}

void *crypto_genkey(void *h,int curve)
{
	switch(curve)
	{
	case CRYPTO_BRAINPOOLP512R1:
		return usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP512R1);
	case CRYPTO_BRAINPOOLP384R1:
		return usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP384R1);
	case CRYPTO_BRAINPOOLP256R1:
		return usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP256R1);
	case CRYPTO_SECP521R1:
		return usicrypt_ec_generate(h,USICRYPT_SECP521R1);
	case CRYPTO_SECP384R1:
		return usicrypt_ec_generate(h,USICRYPT_SECP384R1);
	case CRYPTO_SECP256R1:
		return usicrypt_ec_generate(h,USICRYPT_SECP256R1);
	default:return NULL;
	}
}

COLD struct crypto_conf *crypto_config(void *h,void *ltk,
	int setup_max,int enable_x25519,int enable_secp,int enable_brainpool,
	int minbits,int maxbits,int enable_sig_secp,int enable_sig_brainpool,
	int sig_minbits,int sig_maxbits,int maxdata,int datapad)
{
	struct crypto_conf *c;

	if(maxdata<0||maxdata>65535-32||datapad<0||datapad>65535-32)return NULL;
	if(maxdata&&datapad>maxdata)return NULL;
	if(setup_max<512||setup_max>65535-32)return NULL;
	if(minbits<0||minbits>1023)return NULL;
	if(maxbits<0||maxbits>1023)return NULL;
	if(sig_minbits<0||sig_minbits>1023)return NULL;
	if(sig_maxbits<0||sig_maxbits>1023)return NULL;
	if(maxbits)
	{
		if(maxbits<256)return NULL;
		if(minbits>maxbits)return NULL;
		if(maxbits>384&&setup_max<768)return NULL;
	}
	else if((enable_secp||enable_brainpool)&&setup_max<768)return NULL;
	if(sig_maxbits)
	{
		if(sig_maxbits<256)return NULL;
		if(sig_minbits>sig_maxbits)return NULL;
		if(sig_maxbits>384&&setup_max<768)return NULL;
	}
	else if(setup_max<768)return NULL;
	if(!enable_x25519&&!enable_secp&&!enable_brainpool)return NULL;
	if(!enable_sig_secp&&!enable_sig_brainpool)return NULL;
	if(enable_secp&&minbits>521)return NULL;
	if(enable_brainpool&&minbits>512)return NULL;
	if(enable_x25519&&minbits>256)return NULL;
	if(enable_sig_secp&&sig_minbits>521)return NULL;
	if(enable_sig_brainpool&&sig_minbits>512)return NULL;
	switch(usicrypt_ec_identifier(h,ltk))
	{
	case USICRYPT_BRAINPOOLP512R1:
		if(!enable_sig_brainpool||sig_minbits>512)return NULL;
		if(sig_maxbits&&sig_maxbits<512)return NULL;
		if(setup_max<768)return NULL;
		break;
	case USICRYPT_BRAINPOOLP384R1:
		if(!enable_sig_brainpool||sig_minbits>384)return NULL;
		if(sig_maxbits&&sig_maxbits<384)return NULL;
		break;
	case USICRYPT_BRAINPOOLP256R1:
		if(!enable_sig_brainpool||sig_minbits>256)return NULL;
		if(sig_maxbits&&sig_maxbits<256)return NULL;
		break;
	case USICRYPT_SECP521R1:
		if(!enable_sig_secp||sig_minbits>521)return NULL;
		if(sig_maxbits&&sig_maxbits<521)return NULL;
		if(setup_max<768)return NULL;
		break;
	case USICRYPT_SECP384R1:
		if(!enable_sig_secp||sig_minbits>384)return NULL;
		if(sig_maxbits&&sig_maxbits<384)return NULL;
		break;
	case USICRYPT_SECP256R1:
		if(!enable_sig_secp||sig_minbits>256)return NULL;
		if(sig_maxbits&&sig_maxbits<256)return NULL;
		break;
	default:return NULL;
	}
	if(!(c=malloc(sizeof(struct crypto_conf))))return NULL;
	memset(c,0,sizeof(struct crypto_conf));
	c->setup_max=setup_max;
	c->maxdata=maxdata;
	c->datapad=datapad;
	c->minbits=minbits;
	c->maxbits=maxbits;
	c->sig_minbits=sig_minbits;
	c->sig_maxbits=sig_maxbits;
	if(enable_x25519)c->enable_x25519=1;
	if(enable_secp)c->enable_secp=1;
	if(enable_brainpool)c->enable_brainpool=1;
	if(enable_sig_secp)c->enable_sig_secp=1;
	if(enable_sig_brainpool)c->enable_sig_brainpool=1;
	c->ltk=ltk;
	return c;
}

void crypto_reconfig(void *h,struct crypto_conf *c,void *ltk)
{
	usicrypt_ec_free(h,c->ltk);
	c->ltk=ltk;
}

COLD void crypto_free(void *h,struct crypto_conf *c)
{
	usicrypt_ec_free(h,c->ltk);
	free(c);
}

struct crypto_x25519 *crypto_initialrequest(void *h,struct crypto_conf *c,
	void **msg,int *mlen,int rtp)
{
	int i;
	int j;
	int l;
	struct crypto_x25519 *r;
#ifndef NO_POPCNT
	unsigned long long *ull;
#endif

	i=(c->setup_max<1288?1288:c->setup_max);
	if(!(r=malloc(sizeof(struct crypto_x25519)+(i<<1))))return NULL;
	r->lcl=r->data;
	r->rmt=r->data+i;
	r->pub=NULL;
	r->shared=NULL;

	l=0;
	while(1)
	{
		if(!(r->key=usicrypt_x25519_generate(h)))
		{
			if(++l>=100)
			{
				free(r);
				return NULL;
			}
			continue;
		}
		if(!(r->der=usicrypt_x25519_get_pub(h,r->key,&r->dlen)))
		{
			usicrypt_x25519_free(h,r->key);
			continue;
		}
#ifndef NO_ELLIGATOR
		if(!Elligator_2_Curve25519_encode(r->lcl,r->der+r->dlen-32,0))
		{
			usicrypt_x25519_free(h,r->key);
			free(r->der);
			continue;
		}
#else
		memcpy(r->lcl,r->der+r->dlen-32,32);
#endif
#ifndef NO_POPCNT
		ull=(unsigned long long *)(r->lcl);
		l=__builtin_popcountll(ull[0])+__builtin_popcountll(ull[1])+
			__builtin_popcountll(ull[2])+
			__builtin_popcountll(ull[3]);
		if(l>=96&&l<=160)break;
		usicrypt_x25519_free(h,r->key);
		free(r->der);
		l=0;
#else
		break;
#endif
	}

	l=0;
	while(1)
	{
		if(crypto_random(h,&r->lcllen,sizeof(r->lcllen)))
		{
			if(++l>=100)
			{
				usicrypt_x25519_free(h,r->key);
				free(r->der);
				free(r);
				return NULL;
			}
			continue;
		}
		r->lcllen&=0x7fffffff;
		r->lcllen%=c->setup_max-95;
		break;
	}

	r->lcllen+=96;

	if(rtp)if((r->lcllen=((r->lcllen+183)/184)*184)>1288)r->lcllen=1288;

	for(i=32;i<r->lcllen;i+=32)
	{
		j=(r->lcllen-i>32?32:r->lcllen-i);
#ifndef NO_POPCNT
		ull=(unsigned long long *)(r->lcl+i);
#endif
		l=0;
		while(1)
		{
			if(crypto_random(h,r->lcl+i,j))
			{
				if(++l>=100)
				{
					usicrypt_x25519_free(h,r->key);
					free(r->der);
					free(r);
					return NULL;
				}
				continue;
			}
#ifndef NO_POPCNT
			if(j<32)break;
			l=__builtin_popcountll(ull[0])+
				__builtin_popcountll(ull[1])+
				__builtin_popcountll(ull[2])+
				__builtin_popcountll(ull[3]);
			if(l>=96&&l<=160)break;
			l=0;
#else
			break;
#endif
		}
	}

	*msg=r->lcl;
	*mlen=r->lcllen;

	return r;
}

void crypto_initialfree(void *h,struct crypto_conf *c,struct crypto_x25519 *x)
{
	usicrypt_x25519_free(h,x->key);
	if(x->pub)usicrypt_x25519_free(h,x->pub);
	free(x->der);
	if(x->shared)
	{
		usicrypt_memclear(h,x->shared,x->slen);
		free(x->shared);
	}
	free(x);
}

struct crypto_x25519 *crypto_initialreply(void *h,struct crypto_conf *c,
	void *data,int len,void **msg,int *mlen,int rtp)
{
	int max;
#ifndef NO_ELLIGATOR
	bool v;
#endif
#ifndef NO_POPCNT
	int i;
	int l;
	unsigned long long *ull;
#endif
	struct crypto_x25519 *r;
	unsigned char bfr[32];

	if(!rtp)max=c->setup_max;
	else if((max=((c->setup_max+183)/184)*184)>1288)max=1288;
	if(len<96||len>max||(rtp&&len%184))goto err1;
#ifndef NO_POPCNT
	for(i=0;i<len;i+=32)
	{
		if(len-i<32)break;
		ull=(unsigned long long *)(data+i);
		l=__builtin_popcountll(ull[0])+__builtin_popcountll(ull[1])+
			__builtin_popcountll(ull[2])+
			__builtin_popcountll(ull[3]);
		if(l<96||l>160)goto err1;
	}
#endif
#ifndef NO_ELLIGATOR
	if(!Elligator_2_Curve25519_decode(bfr,&v,data)||v)goto err1;
#else
	memcpy(bfr,data,32);
#endif
	if(!(r=crypto_initialrequest(h,c,msg,mlen,rtp)))goto err2;
	memcpy(r->der+r->dlen-32,bfr,32);
	if(!(r->pub=usicrypt_x25519_set_pub(h,r->der,r->dlen)))goto err3;
	memcpy(r->rmt,data,len);
	r->rmtlen=len;
	usicrypt_memclear(h,bfr,sizeof(bfr));
	return r;

err3:	crypto_initialfree(h,c,r);
err2:	usicrypt_memclear(h,bfr,sizeof(bfr));
err1:	return NULL;
}

int crypto_initialresponse(void *h,struct crypto_conf *c,
	struct crypto_x25519 *x,void *data,int len,int rtp)
{
	int max;
#ifndef NO_ELLIGATOR
	bool v;
#endif
#ifndef NO_POPCNT
	int i;
	int l;
	unsigned long long *ull;
#endif

	if(!rtp)max=c->setup_max;
	else if((max=((c->setup_max+183)/184)*184)>1288)max=1288;
	if(len<96||len>max||(rtp&&len%184))return -1;
#ifndef NO_POPCNT
	for(i=0;i<len;i+=32)
	{
		if(len-i<32)break;
		ull=(unsigned long long *)(data+i);
		l=__builtin_popcountll(ull[0])+__builtin_popcountll(ull[1])+
			__builtin_popcountll(ull[2])+
			__builtin_popcountll(ull[3]);
		if(l<96||l>160)return -1;
	}
#endif
#ifndef NO_ELLIGATOR
	if(!Elligator_2_Curve25519_decode(x->der+x->dlen-32,&v,data)||v)
		return -1;
#else
	memcpy(x->der+x->dlen-32,data,32);
#endif
	if(!(x->pub=usicrypt_x25519_set_pub(h,x->der,x->dlen)))return -1;
	memcpy(x->rmt,data,len);
	x->rmtlen=len;
	return 0;
}

struct crypto_conn *crypto_initialderive(void *h,struct crypto_conf *cnf,
	struct crypto_x25519 *x,int packetmode,int mode,int rtp,int minmsg,
	int small,int strip,int alt)
{
	struct crypto_conn *c;
	int i;
	int nsz=12;
	int asz=16;
	unsigned char key[32];

	if(small)nsz=asz=8;
	if(!(x->shared=usicrypt_x25519_derive(h,x->key,x->pub,&x->slen)))
		goto err1;
	for(i=0;i<x->slen;i++)if(((unsigned char *)x->shared)[i])break;
	if(i==x->slen)goto err1;
	if(!(c=malloc(sizeof(struct crypto_conn))))goto err1;
	for(i=32;i<96;i++)x->rmt[i]^=x->lcl[i];
	if(usicrypt_hkdf(h,USICRYPT_SHA256,x->shared,x->slen,x->rmt+32,32,
		mode?"psvgcm":"pclgcm",7,key))goto err2;
	if(!(c->txgcm=usicrypt_aeadcipher_init(h,
		alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		key,256,nsz,asz)))goto err2;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,x->shared,x->slen,x->rmt+32,32,
		mode?"pclgcm":"psvgcm",7,key))goto err3;
	if(!(c->rxgcm=usicrypt_aeadcipher_init(h,
		alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		key,256,nsz,asz)))goto err3;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,x->shared,x->slen,x->rmt+64,32,
		"precfb8",8,key))goto err4;
	if(!(c->cfb8=usicrypt_blkcipher_init(h,
		alt?USICRYPT_CAMELLIA:USICRYPT_AES,USICRYPT_CFB8,
		key,256,NULL)))goto err4;
	if(!(c->nonce=malloc(2*sizeof(struct crypto_nonce)+sizeof(uint64_t))))
		goto err5;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,x->shared,x->slen,x->rmt+64,32,
		"psvnonce",9,key))goto err6;
	init_nonce(key,&c->nonce[mode?1:0],small,NULL);
	if(usicrypt_hkdf(h,USICRYPT_SHA256,x->shared,x->slen,x->rmt+64,32,
		"pclnonce",9,key))goto err7;
	init_nonce(key,&c->nonce[mode?0:1],small,&c->nonce[2]);
	usicrypt_memclear(h,key,sizeof(key));
	if(!rtp)c->maxmsg=cnf->setup_max;
	else if((c->maxmsg=((cnf->setup_max+183)/184)*184)>1288)c->maxmsg=1288;
	if((c->maxpad=c->maxmsg-nsz-asz-4)<0)c->maxpad=0;
	c->maxdata=cnf->maxdata;
	c->datapad=cnf->datapad;
	c->packet=packetmode;
	c->minmsg=minmsg;
	c->small=small;
	c->strip=strip;
	c->hdr=(c->small?20:32);
	c->off=(c->small?8:12);
	c->base=(c->strip?0:c->off);
	if(c->strip)c->hdr-=c->off;
	c->alt=alt;
	return c;

err7:	usicrypt_memclear(h,key,sizeof(key));
err6:	free(c->nonce);
err5:	usicrypt_blkcipher_exit(c->cfb8);
err4:	usicrypt_aeadcipher_exit(c->rxgcm);
err3:	usicrypt_aeadcipher_exit(c->txgcm);
err2:	free(c);
err1:	return NULL;
}

int crypto_hdrsize(void *h,struct crypto_conn *c)
{
	return c?c->hdr:32;
}

HOT void *crypto_encode(void *h,struct crypto_conn *c,int type,void *data,
	int len,int bits,int *rlen,int rtp)
{
	int pad=0;
	int maxpad;
	unsigned int val;
	unsigned char *msg;
	unsigned char *wrk;
	unsigned char iv[16];
	unsigned char aad[14];

	if(UNLIKELY(len<=0)||UNLIKELY(len>65535-c->hdr))goto err1;
	if(c->packet&&UNLIKELY(len>16383-c->hdr))goto err1;
	if(UNLIKELY(rtp))if(len>1288-c->hdr)goto err1;

	if(UNLIKELY(!type))
	{
		if(c->maxmsg)
		{
			if(len>c->maxmsg-c->hdr)goto err1;
			maxpad=c->maxmsg-len-c->hdr;
			if(maxpad>c->maxpad)maxpad=c->maxpad;
		}
		else maxpad=c->maxpad;

		if(len+maxpad>65535-c->hdr)maxpad=65535-c->hdr-len;
		if(c->packet&&len+maxpad>16383-c->hdr)maxpad=16383-c->hdr-len;
		if(maxpad>4095)maxpad=4095;

		if(maxpad>0)
		{
			if(crypto_random(h,&val,sizeof(val)))goto err1;
			pad=val%(maxpad+1);
		}
	}
	else if(c->datapad)
	{
		if(c->maxdata)
		{
			if(len<c->maxdata)
			{
				maxpad=c->maxdata-len;
				if(maxpad>c->datapad)maxpad=c->datapad;
			}
			else maxpad=0;
		}
		else maxpad=c->datapad;

		if(len+maxpad>65535-c->hdr)maxpad=65535-c->hdr-len;
		if(c->packet&&len+maxpad>16383-c->hdr)maxpad=16383-c->hdr-len;
		if(maxpad>4095)maxpad=4095;

		if(maxpad>0)
		{
			if(crypto_random(h,&val,sizeof(val)))goto err1;
			pad=val%(maxpad+1);
		}
	}

	if(UNLIKELY(c->minmsg))
		if(len+c->hdr+pad<c->minmsg)pad=c->minmsg-c->hdr-len;

	if(UNLIKELY(rtp))
	{
		if(len+pad>=1288-c->hdr)pad=1288-c->hdr-len;
		else pad=(((len+pad+c->hdr+183)/184)*184)-len-c->hdr;
	}

	*rlen=len+c->hdr+pad;

	if(UNLIKELY(!(msg=malloc(*rlen))))goto err1;
	wrk=(c->strip?aad:msg);
	if(pad)memset(msg+len+c->hdr,val>>24,pad);

	inc_set_nonce(wrk,&c->nonce[1],c->small);
	if(c->packet)*((unsigned short *)(msg+c->base))=
		htobe16((len+c->hdr+pad)|((bits&3)<<14));
	else *((unsigned short *)(msg+c->base))=htobe16((len+c->hdr+pad));
	*((unsigned short *)(msg+c->hdr-2))=htobe16((pad<<4)|(type&0xf));
	memcpy(msg+c->hdr,data,len);
	if(c->strip)*((unsigned short *)(aad+c->off))=
		*((unsigned short *)msg);
	if(UNLIKELY(usicrypt_aeadcipher_encrypt(c->txgcm,wrk,msg+c->hdr-2,
		len+2+pad,wrk,c->base+2,msg+c->hdr-2,msg+c->base+2)))goto err3;
	if(c->small)
	{
		memcpy(iv,msg+c->base+2,8);
		memcpy(iv+8,msg+c->base+2,8);
		usicrypt_blkcipher_reset(c->cfb8,iv);
	}
	else usicrypt_blkcipher_reset(c->cfb8,msg+c->base+2);
	if(UNLIKELY(usicrypt_blkcipher_encrypt(c->cfb8,msg,c->base+2,msg)))
		goto err2;
	return msg;

err3:	usicrypt_memclear(h,msg+c->hdr,len);
err2:	free(msg);
err1:	return NULL;
}

HOT int crypto_unmask(struct crypto_conn *c,void *data,int len,int *rval,
	int *bits)
{
	unsigned char iv[16];

	if(UNLIKELY(len<c->hdr-2))return -1;
	if(c->small)
	{
		memcpy(iv,data+c->base+2,8);
		memcpy(iv+8,data+c->base+2,8);
		usicrypt_blkcipher_reset(c->cfb8,iv);
	}
	else usicrypt_blkcipher_reset(c->cfb8,data+c->base+2);
	if(UNLIKELY(usicrypt_blkcipher_decrypt(c->cfb8,data,c->base+2,data)))
			return -1;
	*rval=be16toh(*((unsigned short *)(data+c->base)));
	if(c->packet)
	{
		*bits=*rval>>14;
		*rval&=0x3fff;
	}
	else *bits=0;
	if(UNLIKELY(*rval<c->hdr))return -1;
	return 0;
}

HOT void *crypto_decode(void *h,struct crypto_conn *c,int *type,void *data,
	int len,int *rlen,long long *delta)
{
	int val;
	unsigned char *msg=NULL;
	unsigned char *wrk;
	unsigned char aad[14];

	if(UNLIKELY(len<c->hdr))goto err1;
	if(c->strip)
	{
		if(delta)*delta=1LL;
		wrk=aad;
		inc_set_nonce(aad,&c->nonce[0],c->small);
		*((unsigned short *)(aad+c->off))=*((unsigned short *)data);
	}
	else wrk=data;
	if(UNLIKELY(usicrypt_aeadcipher_decrypt(c->rxgcm,wrk,data+c->hdr-2,
		len-c->hdr+2,wrk,c->base+2,data+c->hdr-2,data+c->base+2)))
			goto err2;
	val=be16toh(*((unsigned short *)(data+c->hdr-2)));
	if(!c->strip)
	{
		if(c->packet)
		{
			if(UNLIKELY(check_set_nonce(data,&c->nonce[0],
				&c->nonce[2],c->small,delta)))goto err2;
		}
		else if(UNLIKELY(inc_cmp_nonce(data,&c->nonce[0],c->small,
			delta)))goto err2;
	}
	if(UNLIKELY(!(*type=val&0xf)))if(c->maxmsg)if(UNLIKELY(len>c->maxmsg))
		goto err2;
	val>>=4;
	if(UNLIKELY((*rlen=len-c->hdr-val)<=0))goto err2;
	if(UNLIKELY(!(msg=malloc(*rlen))))goto err2;
	memcpy(msg,data+c->hdr,*rlen);
	len=*rlen+c->hdr;
err2:	usicrypt_memclear(h,data+c->hdr-2,len-c->hdr+2);
err1:	return msg;
}

struct crypto_setup *crypto_realrequest(void *h,struct crypto_conf *c,int curve,
	void **msg,int *mlen)
{
	int len;
	unsigned char *ptr;
	struct crypto_setup *b;

	if(!(b=malloc(sizeof(struct crypto_setup)+(3*c->setup_max))))goto err1;
	if(!(b->ck=malloc(sizeof(struct crypto_keys)+c->setup_max)))goto err2;
	b->lcl=b->data;
	b->rmt=b->data+c->setup_max;
	b->auth=b->data+(c->setup_max<<1);
	b->lclkey=NULL;
	b->rmtpub=NULL;
	b->lcl[0]=CRYPTO_REQUEST;
	if(crypto_random(h,b->lcl+1,64))goto err3;
	if(curve==CRYPTO_X25519)
	{
		if(!c->enable_x25519||c->minbits>256)goto err3;
		if(c->maxbits&&c->maxbits<256)goto err3;
		b->curve=USICRYPT_X25519;
		if(!(b->key=usicrypt_x25519_generate(h)))goto err3;
		if(!(ptr=usicrypt_x25519_get_pub(h,b->key,&len)))goto err4;
	}
	else
	{
		switch(curve)
		{
		case CRYPTO_BRAINPOOLP512R1:
			if(!c->enable_brainpool||c->minbits>512)goto err3;
			if(c->maxbits&&c->maxbits<512)goto err3;
			b->curve=USICRYPT_BRAINPOOLP512R1;
			break;
		case CRYPTO_BRAINPOOLP384R1:
			if(!c->enable_brainpool||c->minbits>384)goto err3;
			if(c->maxbits&&c->maxbits<384)goto err3;
			b->curve=USICRYPT_BRAINPOOLP384R1;
			break;
		case CRYPTO_BRAINPOOLP256R1:
			if(!c->enable_brainpool||c->minbits>256)goto err3;
			if(c->maxbits&&c->maxbits<256)goto err3;
			b->curve=USICRYPT_BRAINPOOLP256R1;
			break;
		case CRYPTO_SECP521R1:
			if(!c->enable_secp||c->minbits>521)goto err3;
			if(c->maxbits&&c->maxbits<521)goto err3;
			b->curve=USICRYPT_SECP521R1;
			break;
		case CRYPTO_SECP384R1:
			if(!c->enable_secp||c->minbits>384)goto err3;
			if(c->maxbits&&c->maxbits<384)goto err3;
			b->curve=USICRYPT_SECP384R1;
			break;
		case CRYPTO_SECP256R1:
			if(!c->enable_secp||c->minbits>256)goto err3;
			if(c->maxbits&&c->maxbits<256)goto err3;
			b->curve=USICRYPT_SECP256R1;
			break;
		default:goto err3;
		}
		if(!(b->key=usicrypt_ec_generate(h,b->curve)))goto err3;
		if(!(ptr=usicrypt_ec_get_pub(h,b->key,&len)))goto err4;
	}
	b->lcl[65]=(unsigned char)len;
	memcpy(b->lcl+66,ptr,len);
	free(ptr);
	b->lcllen=66+len;
	if(!(ptr=usicrypt_ec_get_pub(h,c->ltk,&len)))goto err4;
	b->lcl[b->lcllen]=(unsigned char)len;
	memcpy(b->lcl+b->lcllen+1,ptr,len);
	free(ptr);
	b->lcllen+=len+1;
	if(!(ptr=usicrypt_ec_sign(h,USICRYPT_SHA384,c->ltk,b->lcl,b->lcllen,
		&len)))goto err3;
	b->lcl[b->lcllen]=(unsigned char)len;
	memcpy(b->lcl+b->lcllen+1,ptr,len);
	free(ptr);
	b->lcllen+=len+1;
	b->pub=NULL;
	b->ltk=NULL;
	*msg=b->lcl;
	*mlen=b->lcllen;
	return b;

err4:	if(curve==CRYPTO_X25519)usicrypt_x25519_free(h,b->key);
	else usicrypt_ec_free(h,b->key);
err3:	free(b->ck);
err2:	free(b);
err1:	return NULL;
}

struct crypto_setup *crypto_realreply(void *h,struct crypto_conf *c,
	void *data,int len,void **msg,int *mlen)
{
	int l1;
	int l2;
	int l3;
	unsigned char *ptr;
	struct crypto_setup *b;

	if(len>c->setup_max||len<67)goto err1;
	if(!(b=malloc(sizeof(struct crypto_setup)+(3*c->setup_max))))goto err1;
	if(!(b->ck=malloc(sizeof(struct crypto_keys)+c->setup_max)))goto err2;
	b->lcl=b->data;
	b->rmt=b->data+c->setup_max;
	b->auth=b->data+(c->setup_max<<1);
	b->lclkey=NULL;
	b->rmtpub=NULL;
	memcpy(b->rmt,data,len);
	b->rmtlen=len;

	if(b->rmt[0]!=CRYPTO_REQUEST)goto err3;
	l1=b->rmt[65];
	if(!l1||b->rmtlen<=67+l1)goto err3;
	l2=b->rmt[66+l1];
	if(!l2||b->rmtlen<=67+l1+l2)goto err3;
	l3=b->rmt[67+l1+l2];
	if(!l3||b->rmtlen!=68+l1+l2+l3)goto err3;
	switch((b->curve=usicrypt_pub_type_from_p8(h,b->rmt+66,l1)))
	{
	case USICRYPT_BRAINPOOLP512R1:
		if(!c->enable_brainpool||c->minbits>512)goto err3;
		if(c->maxbits&&c->maxbits<512)goto err3;
		goto ec;
	case USICRYPT_BRAINPOOLP384R1:
		if(!c->enable_brainpool||c->minbits>384)goto err3;
		if(c->maxbits&&c->maxbits<384)goto err3;
		goto ec;
	case USICRYPT_BRAINPOOLP256R1:
		if(!c->enable_brainpool||c->minbits>256)goto err3;
		if(c->maxbits&&c->maxbits<256)goto err3;
		goto ec;
	case USICRYPT_SECP521R1:
		if(!c->enable_secp||c->minbits>521)goto err3;
		if(c->maxbits&&c->maxbits<521)goto err3;
		goto ec;
	case USICRYPT_SECP384R1:
		if(!c->enable_secp||c->minbits>384)goto err3;
		if(c->maxbits&&c->maxbits<384)goto err3;
		goto ec;
	case USICRYPT_SECP256R1:
		if(!c->enable_secp||c->minbits>256)goto err3;
		if(c->maxbits&&c->maxbits<256)goto err3;
ec:		if(!(b->pub=usicrypt_ec_set_pub(h,b->rmt+66,l1)))goto err3;
		break;
	case USICRYPT_X25519:
		if(!c->enable_x25519||c->minbits>256)goto err3;
		if(c->maxbits&&c->maxbits<256)goto err3;
		if(!(b->pub=usicrypt_x25519_set_pub(h,b->rmt+66,l1)))goto err3;
		break;
	default:goto err3;
	}
	switch(usicrypt_pub_type_from_p8(h,b->rmt+67+l1,l2))
	{
	case USICRYPT_BRAINPOOLP512R1:
		if(!c->enable_sig_brainpool||c->sig_minbits>512)goto err4;
		if(c->sig_maxbits&&c->sig_maxbits<512)goto err4;
		if(c->setup_max<768)goto err4;
		break;
	case USICRYPT_BRAINPOOLP384R1:
		if(!c->enable_sig_brainpool||c->sig_minbits>384)goto err4;
		if(c->sig_maxbits&&c->sig_maxbits<384)goto err4;
		break;
	case USICRYPT_BRAINPOOLP256R1:
		if(!c->enable_sig_brainpool||c->sig_minbits>256)goto err4;
		if(c->sig_maxbits&&c->sig_maxbits<256)goto err4;
		break;
	case USICRYPT_SECP521R1:
		if(!c->enable_sig_secp||c->sig_minbits>521)goto err4;
		if(c->sig_maxbits&&c->sig_maxbits<521)goto err4;
		if(c->setup_max<768)goto err4;
		break;
	case USICRYPT_SECP384R1:
		if(!c->enable_sig_secp||c->sig_minbits>384)goto err4;
		if(c->sig_maxbits&&c->sig_maxbits<384)goto err4;
		break;
	case USICRYPT_SECP256R1:
		if(!c->enable_sig_secp||c->sig_minbits>256)goto err4;
		if(c->sig_maxbits&&c->sig_maxbits<256)goto err4;
		break;
	default:goto err4;
	}
	if(!(b->ltk=usicrypt_ec_set_pub(h,b->rmt+67+l1,l2)))goto err4;
	if(usicrypt_ec_verify(h,USICRYPT_SHA384,b->ltk,b->rmt,67+l1+l2,
		b->rmt+68+l1+l2,l3))goto err5;
	b->siglen=l2;
	b->sig=b->rmt+67+l1;

	b->lcl[0]=CRYPTO_RESPONSE;
	if(crypto_random(h,b->lcl+1,64))goto err5;
	if(b->curve==USICRYPT_X25519)
	{
		if(!(b->key=usicrypt_x25519_generate(h)))goto err5;
		if(!(ptr=usicrypt_x25519_get_pub(h,b->key,&l1)))goto err6;
	}
	else
	{
		if(!(b->key=usicrypt_ec_generate(h,b->curve)))goto err5;
		if(!(ptr=usicrypt_ec_get_pub(h,b->key,&l1)))goto err6;
	}
	b->lcl[65]=(unsigned char)l1;
	memcpy(b->lcl+66,ptr,l1);
	free(ptr);
	b->lcllen=66+l1;
	if(!(ptr=usicrypt_ec_get_pub(h,c->ltk,&l1)))goto err6;
	b->lcl[b->lcllen]=(unsigned char)l1;
	memcpy(b->lcl+b->lcllen+1,ptr,l1);
	free(ptr);
	b->lcllen+=l1+1;
	if(!(ptr=usicrypt_ec_sign(h,USICRYPT_SHA384,c->ltk,b->lcl,b->lcllen,
		&l1)))goto err6;
	b->lcl[b->lcllen]=(unsigned char)l1;
	memcpy(b->lcl+b->lcllen+1,ptr,l1);
	free(ptr);
	b->lcllen+=l1+1;
	memcpy(b->salt1,b->lcl+1,32);
	memcpy(b->salt2,b->lcl+33,32);
	memcpy(b->salt1+32,b->rmt+1,32);
	memcpy(b->salt2+32,b->rmt+33,32);
	b->ck->mode=1;
	b->ck->key=c->ltk;
	b->ck->pub=b->ltk;
	*msg=b->lcl;
	*mlen=b->lcllen;
	return b;

err6:	if(b->curve==USICRYPT_X25519)usicrypt_x25519_free(h,b->key);
	else usicrypt_ec_free(h,b->key);
err5:	usicrypt_ec_free(h,b->ltk);
err4:	if(b->curve==USICRYPT_X25519)usicrypt_x25519_free(h,b->pub);
	else usicrypt_ec_free(h,b->pub);
err3:	free(b->ck);
err2:	free(b);
err1:	return NULL;
}

int crypto_realresponse(void *h,struct crypto_conf *c,struct crypto_setup *b,
	void *data,int len,void **sig,int *siglen)
{
	int l1;
	int l2;
	int l3;

	if(len<67||len>c->setup_max)return -1;
	memcpy(b->rmt,data,len);
	b->rmtlen=len;

	if(b->rmt[0]!=CRYPTO_RESPONSE)return -1;
	l1=b->rmt[65];
	if(!l1||b->rmtlen<=67+l1)return -1;
	l2=b->rmt[66+l1];
	if(!l2||b->rmtlen<=67+l1+l2)return -1;
	l3=b->rmt[67+l1+l2];
	if(!l3||b->rmtlen!=68+l1+l2+l3)return -1;
	if(usicrypt_pub_type_from_p8(h,b->rmt+66,l1)!=b->curve)return -1;
	switch(usicrypt_pub_type_from_p8(h,b->rmt+67+l1,l2))
	{
	case USICRYPT_BRAINPOOLP512R1:
		if(!c->enable_sig_brainpool||c->sig_minbits>512)return -1;
		if(c->sig_maxbits&&c->sig_maxbits<512)return -1;
		if(c->setup_max<768)return -1;
		break;
	case USICRYPT_BRAINPOOLP384R1:
		if(!c->enable_sig_brainpool||c->sig_minbits>384)return -1;
		if(c->sig_maxbits&&c->sig_maxbits<384)return -1;
		break;
	case USICRYPT_BRAINPOOLP256R1:
		if(!c->enable_sig_brainpool||c->sig_minbits>256)return -1;
		if(c->sig_maxbits&&c->sig_maxbits<256)return -1;
		break;
	case USICRYPT_SECP521R1:
		if(!c->enable_sig_secp||c->sig_minbits>521)return -1;
		if(c->sig_maxbits&&c->sig_maxbits<521)return -1;
		if(c->setup_max<768)return -1;
		break;
	case USICRYPT_SECP384R1:
		if(!c->enable_sig_secp||c->sig_minbits>384)return -1;
		if(c->sig_maxbits&&c->sig_maxbits<384)return -1;
		break;
	case USICRYPT_SECP256R1:
		if(!c->enable_sig_secp||c->sig_minbits>256)return -1;
		if(c->sig_maxbits&&c->sig_maxbits<256)return -1;
		break;
	default:return -1;
	}
	if(b->curve==USICRYPT_X25519)
	{
		if(!(b->pub=usicrypt_x25519_set_pub(h,b->rmt+66,l1)))return -1;
	}
	else if(!(b->pub=usicrypt_ec_set_pub(h,b->rmt+66,l1)))return -1;
	if(!(b->ltk=usicrypt_ec_set_pub(h,b->rmt+67+l1,l2)))return -1;
	if(usicrypt_ec_verify(h,USICRYPT_SHA384,b->ltk,b->rmt,67+l1+l2,
		b->rmt+68+l1+l2,l3))return -1;
	b->siglen=l2;
	b->sig=b->rmt+67+l1;
	memcpy(b->salt1,b->rmt+1,32);
	memcpy(b->salt2,b->rmt+33,32);
	memcpy(b->salt1+32,b->lcl+1,32);
	memcpy(b->salt2+32,b->lcl+33,32);
	b->ck->mode=0;
	b->ck->key=c->ltk;
	b->ck->pub=b->ltk;
	*sig=b->sig;
	*siglen=b->siglen;
	return 0;
}

struct crypto_conn *crypto_derivekey(void *h,struct crypto_conf *cnf,
	struct crypto_setup *b,int packetmode,int rtp,int minmsg,int small,
	int strip,int alt)
{
	int i;
	int slen;
	int nsz=12;
	int asz=16;
	struct crypto_conn *c;
	unsigned char *shared;
	unsigned char key[32];

	if(small)nsz=asz=8;
	if(!(c=malloc(sizeof(struct crypto_conn))))goto err1;
	if(b->curve==USICRYPT_X25519)
	{
		if(!(shared=usicrypt_x25519_derive(h,b->key,b->pub,&slen)))
			goto err2;
	}
	else if(!(shared=usicrypt_ec_derive(h,b->key,b->pub,&slen)))goto err2;
	for(i=0;i<slen;i++)if(shared[i])break;
	if(i==slen)goto err3;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt1,64,
		b->ck->mode?"svrgcm":"clngcm",7,key))goto err3;
	if(!(c->txgcm=usicrypt_aeadcipher_init(h,
		alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		key,256,nsz,asz)))goto err3;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt1,64,
		b->ck->mode?"clngcm":"svrgcm",7,key))goto err4;
	if(!(c->rxgcm=usicrypt_aeadcipher_init(h,
		alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		key,256,nsz,asz)))goto err4;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt2,64,
		"cfb8",5,key))goto err5;
	if(!(c->cfb8=usicrypt_blkcipher_init(h,
		alt?USICRYPT_CAMELLIA:USICRYPT_AES,USICRYPT_CFB8,
		key,256,NULL)))goto err5;
	if(!(c->nonce=malloc(2*sizeof(struct crypto_nonce)+sizeof(uint64_t))))
		goto err6;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt2,64,
		"svrnonce",9,key))goto err7;
	init_nonce(key,&c->nonce[b->ck->mode?1:0],small,NULL);
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt2,64,
		"clnnonce",9,key))goto err8;
	init_nonce(key,&c->nonce[b->ck->mode?0:1],small,&c->nonce[2]);
	usicrypt_memclear(h,key,sizeof(key));
	usicrypt_memclear(h,shared,slen);
	free(shared);
	if(!rtp)c->maxmsg=cnf->setup_max;
	else if((c->maxmsg=((cnf->setup_max+183)/184)*184)>1288)c->maxmsg=1288;
	if((c->maxpad=c->maxmsg-nsz-asz-4)<0)c->maxpad=0;
	c->maxdata=cnf->maxdata;
	c->datapad=cnf->datapad;
	c->packet=packetmode;
	c->minmsg=minmsg;
	c->small=small;
	c->strip=strip;
	c->hdr=(c->small?20:32);
	c->off=(c->small?8:12);
	c->base=(c->strip?0:c->off);
	if(c->strip)c->hdr-=c->off;
	c->alt=alt;
	return c;

err8:	usicrypt_memclear(h,key,sizeof(key));
err7:	free(c->nonce);
err6:	usicrypt_blkcipher_exit(c->cfb8);
err5:	usicrypt_aeadcipher_exit(c->rxgcm);
err4:	usicrypt_aeadcipher_exit(c->txgcm);
err3:	usicrypt_memclear(h,shared,slen);
	free(shared);
err2:	free(c);
err1:	return NULL;
}

int crypto_rekeysend(void *h,struct crypto_keys *keys,
	struct crypto_rekeydata **ptr,int curve,void **msg,int *mlen)
{
	int slen;
	void *sig;
	struct crypto_rekeydata *r;

	if(!*ptr)
	{
		if(!(r=malloc(sizeof(struct crypto_rekeydata))))goto err1;
		*ptr=r;
		memset(r,0,sizeof(struct crypto_rekeydata));
	}
	else r=*ptr;

	if(!(r->state&0x01))switch(curve)
	{
	case CRYPTO_BRAINPOOLP512R1:
		r->key=usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP512R1);
		r->curve=USICRYPT_BRAINPOOLP512R1;
		break;
	case CRYPTO_BRAINPOOLP384R1:
		r->key=usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP384R1);
		r->curve=USICRYPT_BRAINPOOLP384R1;
		break;
	case CRYPTO_BRAINPOOLP256R1:
		r->key=usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP256R1);
		r->curve=USICRYPT_BRAINPOOLP256R1;
		break;
	case CRYPTO_SECP521R1:
		r->key=usicrypt_ec_generate(h,USICRYPT_SECP521R1);
		r->curve=USICRYPT_SECP521R1;
		break;
	case CRYPTO_SECP384R1:
		r->key=usicrypt_ec_generate(h,USICRYPT_SECP384R1);
		r->curve=USICRYPT_SECP384R1;
		break;
	case CRYPTO_SECP256R1:
		r->key=usicrypt_ec_generate(h,USICRYPT_SECP256R1);
		r->curve=USICRYPT_SECP256R1;
		break;
	case CRYPTO_X25519:
		r->key=usicrypt_x25519_generate(h);
		r->curve=USICRYPT_X25519;
		break;
	default:r->key=NULL;
		break;
	}
	if(!(r->state&0x01))
	{
		if(!r->key)goto err1;
		if(r->state&0x10)if(r->curve!=r->remote)
		{
			if(r->remote==USICRYPT_X25519)
				usicrypt_x25519_free(h,r->key);
			else usicrypt_ec_free(h,r->key);
			r->key=NULL;
			goto err1;
		}
		r->state|=0x01;
	}
	if(!(r->state&0x02))
	{
		if(crypto_random(h,r->salt1,sizeof(r->salt1)))goto err1;
		else r->state|=0x02;
	}
	if(!(r->state&0x04))
	{
		if(r->curve==USICRYPT_X25519)
		{
			if(!(r->pub=usicrypt_x25519_get_pub(h,r->key,&r->plen)))
				goto err1;
		}
		else if(!(r->pub=usicrypt_ec_get_pub(h,r->key,&r->plen)))
			goto err1;
		r->state|=0x04;
	}
	keys->data[0]=CRYPTO_REKEYREQ;
	memcpy(keys->data+1,r->salt1,sizeof(r->salt1));
	keys->data[sizeof(r->salt1)+1]=(unsigned char)r->plen;
	memcpy(keys->data+sizeof(r->salt1)+2,r->pub,r->plen);
	*mlen=r->plen+sizeof(r->salt1)+2;
	if(!(sig=usicrypt_ec_sign(h,USICRYPT_SHA384,keys->key,keys->data,*mlen,
		&slen)))goto err2;
	keys->data[*mlen]=(unsigned char)slen;
	memcpy(keys->data+*mlen+1,sig,slen);
	*mlen+=slen+1;
	if(!(*msg=malloc(*mlen)))goto err3;
	memcpy(*msg,keys->data,*mlen);
	usicrypt_memclear(h,keys->data,*mlen);
	usicrypt_memclear(h,sig,slen);
	free(sig);
	r->state|=0x08;
	if((r->state&0x1f)==0x1f)return 1;
	return 0;

err3:	usicrypt_memclear(h,keys->data,*mlen);
	usicrypt_memclear(h,sig,slen);
	free(sig);
err2:	if(r->curve==USICRYPT_X25519)usicrypt_x25519_free(h,r->pub);
	else usicrypt_ec_free(h,r->pub);
	r->pub=NULL;
err1:	return -1;
}

int crypto_rekeyrecv(void *h,struct crypto_keys *keys,
	struct crypto_rekeydata **ptr,int curve,void *data,int len,void *mem)
{
	int l1;
	int l2;
	int remote;
	struct crypto_rekeydata *r;
	void *rmt;

	if(len==64&&*((unsigned char *)data)==CRYPTO_NOOP)return 3;
	if(len>0&&*((unsigned char *)data)==CRYPTO_NEWKEY)return 4;
	if(len>0&&*((unsigned char *)data)==CRYPTO_REVOKE)return 5;

	if(len>0&&*((unsigned char *)data)==CRYPTO_REKEYACK)
	{
		if(len<=66)goto err1;
		l1=((unsigned char *)data)[sizeof(r->salt)+1];
		if(!l1||len!=l1+sizeof(r->salt)+2)goto err1;
		if(usicrypt_ec_verify(h,USICRYPT_SHA384,keys->pub,data,
			sizeof(r->salt)+1,data+sizeof(r->salt)+2,l1))goto err1;
		if(memcmp(mem?mem:r->salt1,data+1,sizeof(r->salt1)))goto err1;
		return 2;
	}

	if(!*ptr)
	{
		if(!(r=malloc(sizeof(struct crypto_rekeydata))))goto err1;
		*ptr=r;
		memset(r,0,sizeof(struct crypto_rekeydata));
	}
	else r=*ptr;

	if(len<=sizeof(r->salt2)+2)goto err1;
	if(!(l1=((unsigned char *)data)[sizeof(r->salt2)+1]))goto err1;
	if(len<=sizeof(r->salt2)+2+l1)goto err1;
	if(!(l2=((unsigned char *)data)[sizeof(r->salt2)+2+l1]))goto err1;
	if(len!=sizeof(r->salt2)+3+l1+l2)goto err1;
	if(*((unsigned char *)data)!=CRYPTO_REKEYREQ)goto err1;
	if(usicrypt_ec_verify(h,USICRYPT_SHA384,keys->pub,data,
		sizeof(r->salt2)+2+l1,data+sizeof(r->salt2)+3+l1,l2))goto err1;

	switch((remote=usicrypt_pub_type_from_p8(h,data+sizeof(r->salt2)+2,
		len-sizeof(r->salt2)-2)))
	{
	case USICRYPT_BRAINPOOLP512R1:
		if(curve!=CRYPTO_BRAINPOOLP512R1)goto err1;
		break;
	case USICRYPT_BRAINPOOLP384R1:
		if(curve!=CRYPTO_BRAINPOOLP384R1)goto err1;
		break;
	case USICRYPT_BRAINPOOLP256R1:
		if(curve!=CRYPTO_BRAINPOOLP256R1)goto err1;
		break;
	case USICRYPT_SECP521R1:
		if(curve!=CRYPTO_SECP521R1)goto err1;
		break;
	case USICRYPT_SECP384R1:
		if(curve!=CRYPTO_SECP384R1)goto err1;
		break;
	case USICRYPT_SECP256R1:
		if(curve!=CRYPTO_SECP256R1)goto err1;
		break;
	case USICRYPT_X25519:
		if(curve!=CRYPTO_X25519)goto err1;
		break;
	default:goto err1;
	}
	if(remote==USICRYPT_X25519)
	{
		if(!(rmt=usicrypt_x25519_set_pub(h,data+sizeof(r->salt2)+2,
			len-sizeof(r->salt2)-2)))goto err1;
	}
	else
	{
		if(!(rmt=usicrypt_ec_set_pub(h,data+sizeof(r->salt2)+2,
			len-sizeof(r->salt2)-2)))goto err1;
	}
	if(r->rmt)
	{
		if(r->remote==USICRYPT_X25519)usicrypt_x25519_free(h,r->rmt);
		else usicrypt_ec_free(h,r->rmt);
	}
	r->rmt=rmt;
	r->remote=remote;
	memcpy(r->salt2,data+1,sizeof(r->salt2));
	if(r->state&0x01)if(r->curve!=r->remote)return -1;
	r->state|=0x10;
	if((r->state&0x1f)==0x1f)return 1;
	return 0;

err1:	return -1;
}

struct crypto_conn *crypto_rekeyderive(void *h,struct crypto_keys *keys,
	struct crypto_rekeydata **ptr,struct crypto_conn *conn,
	void **msg,int *mlen)
{
	int slen;
	int len;
	int nsz=12;
	int asz=16;
	void *shared;
	void *cfb8;
	void *rxgcm;
	void *txgcm;
	void *sig;
	struct crypto_rekeydata *r;
	struct crypto_conn *c;
	unsigned char key[32];

	if(conn->small)nsz=asz=8;

	if(!(r=*ptr)||(r->state&0x1f)!=0x1f)goto err1;

	if(r->curve==USICRYPT_X25519)
	{
		if(!(shared=usicrypt_x25519_derive(h,r->key,r->rmt,&slen)))
			goto err1;
	}
	else
	{
		if(!(shared=usicrypt_ec_derive(h,r->key,r->rmt,&slen)))
			goto err1;
	}
	for(len=0;len<slen;len++)if(((unsigned char *)shared)[len])break;
	if(len==slen)goto err2;
	if(keys->mode)
	{
		memcpy(r->salt,r->salt2,sizeof(r->salt2)>>1);
		memcpy(r->salt+(sizeof(r->salt2)>>1),r->salt1,
			sizeof(r->salt1)>>1);
	}
	else
	{
		memcpy(r->salt,r->salt1,sizeof(r->salt1)>>1);
		memcpy(r->salt+(sizeof(r->salt1)>>1),r->salt2,
			sizeof(r->salt2)>>1);
	}
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,r->salt,
		sizeof(r->salt)>>1,keys->mode?"svrgcm":"clngcm",7,key))
		goto err2;
	if(!(txgcm=usicrypt_aeadcipher_init(h,
		conn->alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		key,256,nsz,asz)))goto err2;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,r->salt,
		sizeof(r->salt)>>1,keys->mode?"clngcm":"svrgcm",7,key))
		goto err3;
	if(!(rxgcm=usicrypt_aeadcipher_init(h,
		conn->alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		key,256,nsz,asz)))goto err3;
	if(keys->mode)
	{
		memcpy(r->salt,r->salt2+(sizeof(r->salt2)>>1),
			sizeof(r->salt2)>>1);
		memcpy(r->salt+(sizeof(r->salt2)>>1),
			r->salt1+(sizeof(r->salt1)>>1),sizeof(r->salt1)>>1);
	}
	else
	{
		memcpy(r->salt,r->salt1+(sizeof(r->salt1)>>1),
			sizeof(r->salt1)>>1);
		memcpy(r->salt+(sizeof(r->salt1)>>1),
			r->salt2+(sizeof(r->salt2)>>1),sizeof(r->salt2)>>1);
	}
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,
		r->salt+(sizeof(r->salt)>>1),sizeof(r->salt)>>1,"cfb8",5,key))
		goto err4;
	if(!(cfb8=usicrypt_blkcipher_init(h,
		conn->alt?USICRYPT_CAMELLIA:USICRYPT_AES,USICRYPT_CFB8,
		key,256,NULL)))goto err4;

	if(!(c=malloc(sizeof(struct crypto_conn))))goto err5;
	c->cfb8=cfb8;
	c->nonce=conn->nonce;
	c->txgcm=txgcm;
	c->rxgcm=rxgcm;
	c->maxmsg=conn->maxmsg;
	c->maxpad=conn->maxpad;
	c->maxdata=conn->maxdata;
	c->datapad=conn->datapad;
	c->packet=conn->packet;
	c->minmsg=conn->minmsg;
	c->small=conn->small;
	c->strip=conn->strip;
	c->hdr=conn->hdr;
	c->off=conn->off;
	c->base=conn->base;
	c->alt=conn->alt;
	usicrypt_memclear(h,shared,slen);
	free(shared);

	keys->data[0]=CRYPTO_REKEYACK;
	memcpy(keys->data+1,r->salt2,sizeof(r->salt2));
	if(!(sig=usicrypt_ec_sign(h,USICRYPT_SHA384,keys->key,keys->data,
		sizeof(r->salt2)+1,&len)))goto err6;
	keys->data[sizeof(r->salt2)+1]=(unsigned char)len;
	*mlen=sizeof(r->salt2)+2+len;
	if(!(*msg=malloc(*mlen)))goto err7;
	memcpy(*msg,keys->data,sizeof(r->salt2)+2);
	memcpy(*msg+sizeof(r->salt2)+2,sig,len);
	usicrypt_memclear(h,sig,len);
	free(sig);

	r->state|=0x20;
	return c;

err7:	usicrypt_memclear(h,sig,len);
	free(sig);
err6:	free(c);
err5:	usicrypt_blkcipher_exit(cfb8);
err4:	usicrypt_aeadcipher_exit(rxgcm);
err3:	usicrypt_aeadcipher_exit(txgcm);
err2:	usicrypt_memclear(h,shared,slen);
	free(shared);
err1:	return NULL;
}

void crypto_rekeyfree(void *h,struct crypto_rekeydata **ptr, void *mem)
{
	struct crypto_rekeydata *r;

	if(!(r=*ptr))return;
	if(r->key)
	{
		if(r->curve==USICRYPT_X25519)usicrypt_x25519_free(h,r->key);
		else usicrypt_ec_free(h,r->key);
	}
	if(r->rmt)
	{
		if(r->remote==USICRYPT_X25519)usicrypt_x25519_free(h,r->rmt);
		else usicrypt_ec_free(h,r->rmt);
	}
	if(r->pub)free(r->pub);
	if(mem)memcpy(mem,r->salt1,sizeof(r->salt1));
	usicrypt_memclear(h,r->salt1,sizeof(r->salt1));
	usicrypt_memclear(h,r->salt2,sizeof(r->salt2));
	usicrypt_memclear(h,r->salt,sizeof(r->salt));
	free(r);
	*ptr=NULL;
}

void crypto_freeconn(void *h,struct crypto_conn *c)
{
	usicrypt_aeadcipher_exit(c->rxgcm);
	usicrypt_aeadcipher_exit(c->txgcm);
	usicrypt_blkcipher_exit(c->cfb8);
	usicrypt_memclear(h,&c->nonce,2*sizeof(struct crypto_nonce));
	free(c->nonce);
	free(c);
}

void crypto_freepair(void *h,struct crypto_conn **c,int which)
{
	if(which&2)
	{
		if(c[1])
		{
			if(c[1]!=c[0])
			{
				if(!c[0]||(c[0]&&c[0]->cfb8!=c[1]->cfb8))
					usicrypt_blkcipher_exit(c[1]->cfb8);
				if(!c[0]||(c[0]&&c[0]->nonce!=c[1]->nonce))
				{
					usicrypt_memclear(h,&c[1]->nonce,
						2*sizeof(struct crypto_nonce));
					free(c[1]->nonce);
				}
				usicrypt_aeadcipher_exit(c[1]->rxgcm);
				usicrypt_aeadcipher_exit(c[1]->txgcm);
				free(c[1]);
			}
			c[1]=NULL;
		}
	}

	if(which&1)
	{
		if(c[0])
		{
			if(c[0]!=c[1])
			{
				if(!c[1]||(c[1]&&c[1]->cfb8!=c[0]->cfb8))
					usicrypt_blkcipher_exit(c[0]->cfb8);
				if(!c[1]||(c[1]&&c[1]->nonce!=c[0]->nonce))
				{
					usicrypt_memclear(h,&c[0]->nonce,
						2*sizeof(struct crypto_nonce));
					free(c[0]->nonce);
				}
				usicrypt_aeadcipher_exit(c[0]->rxgcm);
				usicrypt_aeadcipher_exit(c[0]->txgcm);
				free(c[0]);
			}
			c[0]=NULL;
		}
	}
}

int crypto_authrequest(void *h,struct crypto_conf *c,struct crypto_setup *b,
	void *auth,void **msg,int *mlen,int rekey)
{
	int l1;
	int l2;
	unsigned char *ptr;

	if(auth)switch(usicrypt_ec_identifier(h,auth))
	{
	case USICRYPT_BRAINPOOLP512R1:
	case USICRYPT_BRAINPOOLP384R1:
	case USICRYPT_BRAINPOOLP256R1:
	case USICRYPT_SECP521R1:
	case USICRYPT_SECP384R1:
	case USICRYPT_SECP256R1:
		break;
	default:return -1;
	}
	b->auth[0]=CRYPTO_AUTHREQ;
	memcpy(b->auth+1,b->rmt+1,64);
	if(!rekey)l1=0;
	else if(b->curve==USICRYPT_X25519)
	{
		if(!(b->lclkey=usicrypt_x25519_generate(h)))return -1;
		if(!(ptr=usicrypt_x25519_get_pub(h,b->lclkey,&l1)))return -1;
	}
	else
	{
		if(!(b->lclkey=usicrypt_ec_generate(h,b->curve)))return -1;
		if(!(ptr=usicrypt_ec_get_pub(h,b->lclkey,&l1)))return -1;
	}
	if((b->auth[65]=l1))
	{
		memcpy(b->auth+66,ptr,l1);
		free(ptr);
	}
	if(!(ptr=usicrypt_ec_get_pub(h,auth?auth:c->ltk,&l2)))return -1;
	b->auth[66+l1]=(unsigned char)l2;
	memcpy(b->auth+67+l1,ptr,l2);
	free(ptr);
	b->authlen=67+l1+l2;
	if(!(ptr=usicrypt_ec_sign(h,USICRYPT_SHA384,auth?auth:c->ltk,b->auth,
		b->authlen,&l2)))return -1;
	memcpy(b->auth+b->authlen,ptr,l2);
	free(ptr);
	b->authlen+=l2;
	*msg=b->auth;
	*mlen=b->authlen;
	return 0;
}

int crypto_authparse(void *h,struct crypto_conf *c,struct crypto_setup *b,
	void *data,int len)
{
	if(len<67||len>c->setup_max)return -1;
	memcpy(b->auth,data,len);
	if(b->auth[0]!=CRYPTO_AUTHREQ&&b->auth[0]!=CRYPTO_AUTHANON)return -1;
	b->authlen=len;
	if(memcmp(b->auth+1,b->lcl+1,64))return -1;
	b->rekeylen=b->auth[65];
	b->rekey=b->auth+66;
	if(b->rekeylen)
	{
		if(usicrypt_pub_type_from_p8(h,b->rekey,b->rekeylen)!=b->curve)
			return -1;
		if(b->curve==USICRYPT_X25519)
		{
			if(!(b->rmtpub=usicrypt_x25519_set_pub(h,b->rekey,
				b->rekeylen)))return -1;
		}
		else if(!(b->rmtpub=usicrypt_ec_set_pub(h,b->rekey,
			b->rekeylen)))return -1;
	}
	if(len<67+b->rekeylen)return -1;
	b->authsiglen=b->auth[66+b->rekeylen];
	if(!b->authsiglen||b->authlen<=67+b->rekeylen+b->authsiglen)return -1;
	b->authsig=b->auth+67+b->rekeylen;
	switch(usicrypt_pub_type_from_p8(h,b->authsig,b->authsiglen))
	{
	case USICRYPT_BRAINPOOLP512R1:
	case USICRYPT_BRAINPOOLP384R1:
	case USICRYPT_BRAINPOOLP256R1:
	case USICRYPT_SECP521R1:
	case USICRYPT_SECP384R1:
	case USICRYPT_SECP256R1:
		break;
	default:return -1;
	}
	if(b->auth[0]==CRYPTO_AUTHANON)
	{
		b->anonymous=1;
		if(b->siglen!=b->authsiglen||
			memcmp(b->sig,b->authsig,b->siglen))return -1;
	}
	else b->anonymous=0;
	return 0;
}

int crypto_wantsrekey(void *h,struct crypto_conf *c,struct crypto_setup *b)
{
	if(!b->rekeylen)return -1;
	return 0;
}

int crypto_anonymousauth(void *h,struct crypto_conf *c,struct crypto_setup *b)
{
	if(!b->anonymous)return -1;
	return 0;
}

void *crypto_getauthdata(void *h,struct crypto_conf *c,struct crypto_setup *b,
	int *len)
{
	*len=b->authsiglen;
	return b->authsig;
}

int crypto_authverify(void *h,struct crypto_conf *c,struct crypto_setup *b)
{
	void *pub;

	if(!(pub=usicrypt_ec_set_pub(h,b->authsig,b->authsiglen)))goto err1;
	if(usicrypt_ec_verify(h,USICRYPT_SHA384,pub,b->auth,
		67+b->rekeylen+b->authsiglen,
		b->auth+67+b->rekeylen+b->authsiglen,
		b->authlen-67-b->rekeylen-b->authsiglen))goto err2;
	usicrypt_ec_free(h,pub);
	return 0;

err2:	usicrypt_ec_free(h,pub);
err1:	return -1;
}

int crypto_authreply(void *h,struct crypto_conf *c,struct crypto_setup *b,
	int ok,void **msg,int *mlen)
{
	int l1;
	int l2;
	unsigned char *ptr;

	memcpy(b->auth+1,b->rmt+1,64);
	b->auth[65]=0;
	if(ok)
	{
		b->auth[0]=CRYPTO_AUTHOK;
		if(b->rekeylen)
		{
			if(b->curve==USICRYPT_X25519)
			{
				if(!(b->lclkey=usicrypt_x25519_generate(h)))
					return -1;
				if(!(ptr=usicrypt_x25519_get_pub(h,b->lclkey,
					&l1)))return -1;
			}
			else
			{
				if(!(b->lclkey=usicrypt_ec_generate(h,
					b->curve)))return -1;
				if(!(ptr=usicrypt_ec_get_pub(h,b->lclkey,&l1)))
					return -1;
			}
		}
		else l1=0;
	}
	else
	{
		b->auth[0]=CRYPTO_AUTHFAIL;
		l1=0;
	}
	if((b->auth[65]=(unsigned char)l1))
	{
		memcpy(b->auth+66,ptr,l1);
		free(ptr);
	}
	if(!(ptr=usicrypt_ec_sign(h,USICRYPT_SHA384,c->ltk,b->auth,66+l1,&l2)))
		return -1;
	b->auth[66+l1]=(unsigned char)l2;
	memcpy(b->auth+67+l1,ptr,l2);
	free(ptr);
	b->authlen=67+l1+l2;
	*msg=b->auth;
	*mlen=b->authlen;
	return 0;
}

int crypto_authresponse(void *h,struct crypto_conf *c,struct crypto_setup *b,
	void *data,int len)
{
	int l1;
	int l2;

	if(len<=67||len>c->setup_max)return -1;
	memcpy(b->auth,data,len);
	b->authlen=len;
	if(memcmp(b->auth+1,b->lcl+1,64))return -1;
	if((l1=b->auth[65]))
	{
		if(b->authlen<=67+l1)return -1;
		if(usicrypt_pub_type_from_p8(h,b->auth+66,l1)!=b->curve)
			return -1;
		if(b->curve==USICRYPT_X25519)
		{
			if(!(b->rmtpub=usicrypt_x25519_set_pub(h,b->auth+66,
				l1)))return -1;
		}
		else if(!(b->rmtpub=usicrypt_ec_set_pub(h,b->auth+66,l1)))
			return -1;
	}
	l2=b->auth[66+l1];
	if(b->authlen!=67+l1+l2)return -1;
	if(usicrypt_ec_verify(h,USICRYPT_SHA384,b->ltk,b->auth,66+l1,
		b->auth+67+l1,l2))return -1;
	if(b->auth[0]!=CRYPTO_AUTHOK)return 1;
	return 0;
}

int crypto_getcurve(void *h,struct crypto_conf *c,struct crypto_setup *b)
{
	switch(b->curve)
	{
	case USICRYPT_BRAINPOOLP512R1:
		return CRYPTO_BRAINPOOLP512R1;
	case USICRYPT_BRAINPOOLP384R1:
		return CRYPTO_BRAINPOOLP384R1;
	case USICRYPT_BRAINPOOLP256R1:
		return CRYPTO_BRAINPOOLP256R1;
	case USICRYPT_SECP521R1:
		return CRYPTO_SECP521R1;
	case USICRYPT_SECP384R1:
		return CRYPTO_SECP384R1;
	case USICRYPT_SECP256R1:
		return CRYPTO_SECP256R1;
	case USICRYPT_X25519:
		return CRYPTO_X25519;
	default:return -1;
	}
}

struct crypto_conn *crypto_authderive(void *h,struct crypto_conf *cnf,
	struct crypto_setup *b,int packetmode,int rtp,int minmsg,int small,
	int strip,int alt)
{
	int i;
	int slen;
	int nsz=12;
	int asz=16;
	struct crypto_conn *c;
	unsigned char *shared;
	unsigned char key[32];

	if(!b->lclkey||!b->rmtpub)goto err1;
	if(small)nsz=asz=8;
	if(!(c=malloc(sizeof(struct crypto_conn))))goto err1;
	if(b->curve==USICRYPT_X25519)
	{
		if(!(shared=usicrypt_x25519_derive(h,b->lclkey,b->rmtpub,
			&slen)))goto err2;
	}
	else if(!(shared=usicrypt_ec_derive(h,b->lclkey,b->rmtpub,&slen)))
		goto err2;
	for(i=0;i<slen;i++)if(shared[i])break;
	if(i==slen)goto err3;
	memcpy(key,b->salt1,32);
	memcpy(b->salt1,b->salt1+32,32);
	memcpy(b->salt1+32,key,32);
	memcpy(key,b->salt2,32);
	memcpy(b->salt2,b->salt2+32,32);
	memcpy(b->salt2+32,key,32);
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt1,64,
		b->ck->mode?"svrgcm":"clngcm",7,key))goto err3;
	if(!(c->txgcm=usicrypt_aeadcipher_init(h,
		alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		key,256,nsz,asz)))goto err3;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt1,64,
		b->ck->mode?"clngcm":"svrgcm",7,key))goto err4;
	if(!(c->rxgcm=usicrypt_aeadcipher_init(h,
		alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		key,256,nsz,asz)))goto err4;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt2,64,
		"cfb8",5,key))goto err5;
	if(!(c->cfb8=usicrypt_blkcipher_init(h,
		alt?USICRYPT_CAMELLIA:USICRYPT_AES,USICRYPT_CFB8,
		key,256,NULL)))goto err5;
	if(!(c->nonce=malloc(2*sizeof(struct crypto_nonce)+sizeof(uint64_t))))
		goto err6;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt2,64,
		"svrnonce",9,key))goto err7;
	init_nonce(key,&c->nonce[b->ck->mode?1:0],small,NULL);
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,b->salt2,64,
		"clnnonce",9,key))goto err8;
	init_nonce(key,&c->nonce[b->ck->mode?0:1],small,&c->nonce[2]);
	usicrypt_memclear(h,key,sizeof(key));
	usicrypt_memclear(h,shared,slen);
	free(shared);
	if(!rtp)c->maxmsg=cnf->setup_max;
	else if((c->maxmsg=((cnf->setup_max+183)/184)*184)>1288)c->maxmsg=1288;
	if((c->maxpad=c->maxmsg-nsz-asz-4)<0)c->maxpad=0;
	c->maxdata=cnf->maxdata;
	c->datapad=cnf->datapad;
	c->packet=packetmode;
	c->minmsg=minmsg;
	c->small=small;
	c->strip=strip;
	c->hdr=(c->small?20:32);
	c->off=(c->small?8:12);
	c->base=(c->strip?0:c->off);
	if(c->strip)c->hdr-=c->off;
	c->alt=alt;
	return c;

err8:	usicrypt_memclear(h,key,sizeof(key));
err7:	free(c->nonce);
err6:	usicrypt_blkcipher_exit(c->cfb8);
err5:	usicrypt_aeadcipher_exit(c->rxgcm);
err4:	usicrypt_aeadcipher_exit(c->txgcm);
err3:	usicrypt_memclear(h,shared,slen);
	free(shared);
err2:	free(c);
err1:	return NULL;
}

void crypto_setupfree(void *h,struct crypto_conf *c,struct crypto_setup *b,
	struct crypto_keys **k)
{
	if(b->curve==USICRYPT_X25519)
	{
		if(b->key)usicrypt_x25519_free(h,b->key);
		if(b->pub)usicrypt_x25519_free(h,b->pub);
		if(b->lclkey)usicrypt_x25519_free(h,b->lclkey);
		if(b->rmtpub)usicrypt_x25519_free(h,b->rmtpub);
	}
	else
	{
		if(b->key)usicrypt_ec_free(h,b->key);
		if(b->pub)usicrypt_ec_free(h,b->pub);
		if(b->lclkey)usicrypt_ec_free(h,b->lclkey);
		if(b->rmtpub)usicrypt_ec_free(h,b->rmtpub);
	}
	if(k)*k=b->ck;
	else
	{
		free(b->ck);
		if(b->ltk)usicrypt_ec_free(h,b->ltk);
	}
	usicrypt_memclear(h,b,sizeof(struct crypto_setup)+(3*c->setup_max));
	free(b);
}

void crypto_keysfree(void *h,struct crypto_keys *k)
{
	if(k->pub)usicrypt_ec_free(h,k->pub);
	free(k);
}

HOT void crypto_memclear(void *h,void *data,int len)
{
	usicrypt_memclear(h,data,len);
}

COLD void *crypto_read_pem_key(void *h,void *data,int len,void *passwd,int plen)
{
	int derlen;
	int declen;
	int eclen;
	void *der;
	void *dec;
	void *ec;
	void *key=NULL;

	if(!(der=usicrypt_pem_to_p8(h,data,len,&derlen)))goto err1;
	if(passwd&&plen)
	{
		if(!(dec=usicrypt_decrypt_p8(h,passwd,plen,der,derlen,&declen)))
			goto err2;
		usicrypt_memclear(h,der,derlen);
		free(der);
		der=dec;
		derlen=declen;
	}

	switch(usicrypt_key_type_from_p8(h,der,derlen))
	{
	case USICRYPT_BRAINPOOLP512R1:
	case USICRYPT_BRAINPOOLP384R1:
	case USICRYPT_BRAINPOOLP256R1:
	case USICRYPT_SECP521R1:
	case USICRYPT_SECP384R1:
	case USICRYPT_SECP256R1:
		break;
	default:goto err2;
	}

	if(!(ec=usicrypt_p8_to_ec_key(h,der,derlen,&eclen)))goto err2;
	key=usicrypt_ec_set_key(h,ec,eclen);
	free(ec);

err2:	usicrypt_memclear(h,der,derlen);
	free(der);
err1:	return key;
}

COLD void *crypto_make_pem_key(void *h,void *key,int *len,void *passwd,int plen)
{
	int derlen;
	int p8len;
	int enclen;
	void *der;
	void *p8;
	void *enc;
	void *pem=NULL;

	if(!(der=usicrypt_ec_get_key(h,key,&derlen)))goto err1;
	if(!(p8=usicrypt_ec_key_to_p8(h,der,derlen,&p8len)))goto err2;
	if(passwd&&plen)
	{
		if(!(enc=usicrypt_encrypt_p8(h,passwd,plen,p8,p8len,
			USICRYPT_AES,USICRYPT_CFB,256,USICRYPT_SHA384,
			100000,&enclen)))goto err3;
		usicrypt_memclear(h,p8,p8len);
		free(p8);
		p8=enc;
		p8len=enclen;
	}
	pem=usicrypt_p8_to_pem(h,p8,p8len,len);

err3:	usicrypt_memclear(h,p8,p8len);
	free(p8);
err2:	usicrypt_memclear(h,der,derlen);
	free(der);
err1:	return pem;
}

void *crypto_read_pem_pub(void *h,void *data,int len,int *publen)
{
	void *pub;

	if(!(pub=usicrypt_pem_to_p8(h,data,len,publen)))goto err1;

	switch(usicrypt_pub_type_from_p8(h,pub,*publen))
	{
	case USICRYPT_BRAINPOOLP512R1:
	case USICRYPT_BRAINPOOLP384R1:
	case USICRYPT_BRAINPOOLP256R1:
	case USICRYPT_SECP521R1:
	case USICRYPT_SECP384R1:
	case USICRYPT_SECP256R1:
		break;
	default:goto err2;
	}

	return pub;

err2:	free(pub);
err1:	return NULL;
}

void *crypto_write_pem_pub(void *h,void *data,int len,int *pemlen)
{
	void *pem;

	switch(usicrypt_pub_type_from_p8(h,data,len))
	{
	case USICRYPT_BRAINPOOLP512R1:
	case USICRYPT_BRAINPOOLP384R1:
	case USICRYPT_BRAINPOOLP256R1:
	case USICRYPT_SECP521R1:
	case USICRYPT_SECP384R1:
	case USICRYPT_SECP256R1:
		break;
	default:goto err1;
	}

	if(!(pem=usicrypt_p8_to_pem(h,data,len,pemlen)))goto err1;

	return pem;

err1:	return NULL;
}

COLD void *crypto_make_pem_pub(void *h,void *key,int *len)
{
	int p8len;
	void *p8;
	void *pem=NULL;

	if(!(p8=usicrypt_ec_get_pub(h,key,&p8len)))goto err1;
	pem=usicrypt_p8_to_pem(h,p8,p8len,len);
	free(p8);

err1:	return pem;
}

COLD void crypto_free_key(void *h,void *key)
{
	usicrypt_ec_free(h,key);
}

int crypto_portrequest(void *h,void *data,int *len,int min)
{
	if(crypto_random(h,len,sizeof(*len)))return -1;
	*len&=0x1ff;
	*len+=1;
	if(*len<min)*len+=min;
	return crypto_random(h,data,*len);
}

int crypto_portreply(void *h,void *data,int *len,void *port,int min,int alt)
{
	void *cfb8;
	unsigned char key[32];
	unsigned char iv[16];

	if(*len<24||*len>512)goto err1;
	memcpy(key,data,16);
	memcpy(iv,data+16,8);
	usicrypt_memclear(h,data,*len);
	if(crypto_random(h,len,sizeof(*len)))goto err2;
	*len&=0x1ff;
	*len+=1;
	if(*len<min)*len+=min;
	if(crypto_random(h,data,*len))goto err2;
	memcpy(key+16,data,16);
	memcpy(iv+8,data+16,8);
	if(!(cfb8=usicrypt_blkcipher_init(h,
		alt?USICRYPT_CAMELLIA:USICRYPT_AES,USICRYPT_CFB8,key,256,
		iv)))goto err3;
	if(usicrypt_blkcipher_encrypt(cfb8,port,2,data+24))goto err4;
	usicrypt_blkcipher_exit(cfb8);
	usicrypt_memclear(h,key,sizeof(key));
	usicrypt_memclear(h,iv,sizeof(iv));
	return 0;

err4:	usicrypt_blkcipher_exit(cfb8);
err3:	usicrypt_memclear(h,data,*len);
err2:	usicrypt_memclear(h,key,sizeof(key));
	usicrypt_memclear(h,iv,sizeof(iv));
err1:	return -1;
}

int crypto_portresponse(void *h,void *snd,int slen,void *rcv,int rlen,
	void *port,int alt)
{
	int r=-1;
	void *cfb8;
	unsigned char key[32];
	unsigned char iv[16];

	if(slen<24||slen>512||rlen<26||rlen>512)goto err1;
	memcpy(key,snd,16);
	memcpy(key+16,rcv,16);
	memcpy(iv,snd+16,8);
	memcpy(iv+8,rcv+16,8);
	if(!(cfb8=usicrypt_blkcipher_init(h,
		alt?USICRYPT_CAMELLIA:USICRYPT_AES,USICRYPT_CFB8,key,256,
		iv)))goto err2;
	if(usicrypt_blkcipher_decrypt(cfb8,rcv+24,2,port))goto err3;
	r=0;
err3:	usicrypt_blkcipher_exit(cfb8);
err2:	usicrypt_memclear(h,key,sizeof(key));
	usicrypt_memclear(h,iv,sizeof(iv));
	usicrypt_memclear(h,snd,slen);
	usicrypt_memclear(h,rcv,rlen);
err1:	return r;
}

int crypto_noopmsg(void *h,void **data,int *len)
{
	if(!(*data=malloc(64)))return -1;
	((unsigned char *)(*data))[0]=CRYPTO_NOOP;
	if(crypto_random(h,*data+1,64))
	{
		free(*data);
		return -1;
	}
	*len=64;
	return 0;
}

int crypto_pktfinmsg(void *h,struct crypto_keys *keys,void *data,int *len)
{
	int slen;
	void *sig;

	((unsigned char *)data)[0]=CRYPTO_FINMSG;
	if(crypto_random(h,data+1,64))return -1;
	if(!(sig=usicrypt_ec_sign(h,USICRYPT_SHA384,keys->key,data,65,&slen)))
		return -1;
	((unsigned char *)data)[65]=(unsigned char)slen;
	memcpy(data+66,sig,slen);
	usicrypt_memclear(h,sig,slen);
	free(sig);
	*len=slen+66;
	return 0;
}

int crypto_pktackmsg(void *h,struct crypto_keys *keys,void **data,int *len,
	void *pub)
{
	int slen;
	struct crypto_pktkey *p=pub;
	void *sig;

	keys->data[0]=CRYPTO_REKEYACK;
	memcpy(keys->data+1,p->salt,sizeof(p->salt));
	if(!(sig=usicrypt_ec_sign(h,USICRYPT_SHA384,keys->key,keys->data,
		sizeof(p->salt)+1,&slen)))goto err1;
	keys->data[sizeof(p->salt)+1]=(unsigned char)slen;
	memcpy(keys->data+sizeof(p->salt)+2,sig,slen);
	usicrypt_memclear(h,sig,slen);
	free(sig);
	*len=sizeof(p->salt)+2+slen;
	if(!(*data=malloc(*len)))goto err2;
	memcpy(*data,keys->data,*len);
	usicrypt_memclear(h,keys->data,*len);
	return 0;

err2:	usicrypt_memclear(h,keys->data,*len);
err1:	return -1;
}

int crypto_pkttype(void *h,struct crypto_keys *keys,void *data,int len,
	void *key)
{
	int l;
	struct crypto_pktkey *k=key;


	if(len>0)switch(((unsigned char *)data)[0])
	{
	case CRYPTO_REKEYREQ:
		return 1;

	case CRYPTO_REKEYACK:
		if(!k||len<=sizeof(k->salt)+2)break;
		l=((unsigned char *)data)[sizeof(k->salt)+1];
		if(len!=sizeof(k->salt)+2+l)break;
		if(usicrypt_ec_verify(h,USICRYPT_SHA384,keys->pub,data,
			sizeof(k->salt)+1,data+sizeof(k->salt)+2,l))break;
		if(memcmp(data+1,k->salt,sizeof(k->salt)))break;
		return 2;

	case CRYPTO_FINMSG:
		if(len<=66)break;
		l=((unsigned char *)data)[65];
		if(len!=66+l)break;
		if(usicrypt_ec_verify(h,USICRYPT_SHA384,keys->pub,data,65,
			data+66,l))break;
		return 0;

	case CRYPTO_NOOP:
		if(len!=64)break;
		return 3;

	case CRYPTO_NEWKEY:
		return 4;

	case CRYPTO_REVOKE:
		return 5;
	}
	return -1;
}

int crypto_pktrekeygen(void *h,void **key,int curve)
{
	struct crypto_pktkey *c=*key;

	if(c)
	{
		switch(c->curve)
		{
			case CRYPTO_X25519:
				usicrypt_x25519_free(h,c->key);
				break;
			default:usicrypt_ec_free(h,c->key);
				break;
		}
	}
	else if(!(*key=c=malloc(sizeof(struct crypto_pktkey))))goto err1;

	if(crypto_random(h,c->salt,sizeof(c->salt)))goto err2;

	c->curve=curve;

	switch(curve)
	{
	case CRYPTO_BRAINPOOLP512R1:
		c->key=usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP512R1);
		break;
	case CRYPTO_BRAINPOOLP384R1:
		c->key=usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP384R1);
		break;
	case CRYPTO_BRAINPOOLP256R1:
		c->key=usicrypt_ec_generate(h,USICRYPT_BRAINPOOLP256R1);
		break;
	case CRYPTO_SECP521R1:
		c->key=usicrypt_ec_generate(h,USICRYPT_SECP521R1);
		break;
	case CRYPTO_SECP384R1:
		c->key=usicrypt_ec_generate(h,USICRYPT_SECP384R1);
		break;
	case CRYPTO_SECP256R1:
		c->key=usicrypt_ec_generate(h,USICRYPT_SECP256R1);
		break;
	case CRYPTO_X25519:
		c->key=usicrypt_x25519_generate(h);
		break;
	default:c->key=NULL;
		break;
	}
	if(!c->key)goto err3;
	*key=c;
	return 0;

err3:	usicrypt_memclear(h,c->salt,sizeof(c->salt));
err2:	free(c);
err1:	*key=NULL;
	return -1;
}

void crypto_pktrekeyfree(void *h,void **key)
{
	struct crypto_pktkey *c=*key;

	if(!c)return;
	if(c->key)switch(c->curve)
	{
		case CRYPTO_X25519:
			usicrypt_x25519_free(h,c->key);
			break;
		default:usicrypt_ec_free(h,c->key);
			break;
	}
	usicrypt_memclear(h,c->salt,sizeof(c->salt));
	free(c);
	*key=NULL;
}

int crypto_pktrekeytxmsg(void *h,struct crypto_keys *keys,void *key,
	void **msg,int *mlen)
{
	int plen;
	int slen;
	struct crypto_pktkey *c=key;
	void *pub;
	void *sig;

	if(!c||!c->key)goto err1;
	switch(c->curve)
	{
		case CRYPTO_X25519:
			if(!(pub=usicrypt_x25519_get_pub(h,c->key,&plen)))
				goto err1;
			break;
		default:if(!(pub=usicrypt_ec_get_pub(h,c->key,&plen)))goto err1;
			break;
	}
	keys->data[0]=CRYPTO_REKEYREQ;
	memcpy(keys->data+1,c->salt,sizeof(c->salt));
	keys->data[sizeof(c->salt)+1]=(unsigned char)plen;
	memcpy(keys->data+sizeof(c->salt)+2,pub,plen);
	*mlen=plen+sizeof(c->salt)+2;
	if(!(sig=usicrypt_ec_sign(h,USICRYPT_SHA384,keys->key,keys->data,*mlen,
		&slen)))goto err2;
	keys->data[*mlen]=(unsigned char)slen;
	memcpy(keys->data+*mlen+1,sig,slen);
	*mlen+=slen+1;
	usicrypt_memclear(h,sig,slen);
	free(sig);
	if(!(*msg=malloc(*mlen)))goto err2;
	memcpy(*msg,keys->data,*mlen);
	usicrypt_memclear(h,keys->data,*mlen);
	usicrypt_memclear(h,pub,plen);
	free(pub);
	return 0;

err2:	usicrypt_memclear(h,keys->data,*mlen);
	usicrypt_memclear(h,pub,plen);
	free(pub);
err1:	return -1;
}

int crypto_pktrekeyrxmsg(void *h,struct crypto_keys *keys,void **key,
	void *msg,int mlen)
{
	int l1;
	int l2;
	struct crypto_pktkey *c=*key;

	if(c)
	{
		if(c->key)switch(c->curve)
		{
			case CRYPTO_X25519:
				usicrypt_x25519_free(h,c->key);
				break;
			default:usicrypt_ec_free(h,c->key);
				break;
		}
		c->key=NULL;
	}
	else if(!(*key=c=malloc(sizeof(struct crypto_pktkey))))goto err1;

	if(mlen<=sizeof(c->salt)+2)goto err2;
	if(!(l1=((unsigned char *)msg)[sizeof(c->salt)+1]))goto err2;
	if(mlen<=sizeof(c->salt)+2+l1)goto err2;
	if(!(l2=((unsigned char *)msg)[sizeof(c->salt)+2+l1]))goto err2;
	if(mlen!=sizeof(c->salt)+3+l1+l2)goto err2;
	if(*((unsigned char *)msg)!=CRYPTO_REKEYREQ)goto err2;
	if(usicrypt_ec_verify(h,USICRYPT_SHA384,keys->pub,msg,
		sizeof(c->salt)+2+l1,msg+sizeof(c->salt)+3+l1,l2))goto err2;

	c->key=NULL;
	memcpy(c->salt,msg+1,sizeof(c->salt));

	switch(usicrypt_pub_type_from_p8(h,msg+sizeof(c->salt)+2,
		mlen-sizeof(c->salt)-2))
	{
	case USICRYPT_BRAINPOOLP512R1:
		c->curve=CRYPTO_BRAINPOOLP512R1;
		break;
	case USICRYPT_BRAINPOOLP384R1:
		c->curve=CRYPTO_BRAINPOOLP384R1;
		break;
	case USICRYPT_BRAINPOOLP256R1:
		c->curve=CRYPTO_BRAINPOOLP256R1;
		break;
	case USICRYPT_SECP521R1:
		c->curve=CRYPTO_SECP521R1;
		break;
	case USICRYPT_SECP384R1:
		c->curve=CRYPTO_SECP384R1;
		break;
	case USICRYPT_SECP256R1:
		c->curve=CRYPTO_SECP256R1;
		break;
	case USICRYPT_X25519:
		c->curve=CRYPTO_X25519;
		break;
	default:goto err3;
	}

	switch(c->curve)
	{
	case CRYPTO_X25519:
		if(!(c->key=usicrypt_x25519_set_pub(h,msg+sizeof(c->salt)+2,
			mlen-sizeof(c->salt)-2)))goto err3;
		break;
	default:if(!(c->key=usicrypt_ec_set_pub(h,msg+sizeof(c->salt)+2,
			mlen-sizeof(c->salt)-2)))goto err3;
		break;
	}

	return 0;

err3:	usicrypt_memclear(h,c->salt,sizeof(c->salt));
err2:	free(c);
err1:	*key=NULL;
	return -1;
}

struct crypto_conn *crypto_pktrekeyderive(void *h,struct crypto_keys *keys,
	void **key,void **pub,struct crypto_conn *conn)
{
	int i;
	int slen;
	int nsz=12;
	int asz=16;
	struct crypto_pktkey *k;
	struct crypto_pktkey *p;
	void *shared;
	void *rxgcm;
	void *txgcm;
	struct crypto_conn *c=NULL;
	unsigned char kmem[32];
	unsigned char salt[64];

	if(conn->small)nsz=asz=8;

	if(!key||!pub)goto err1;
	k=*key;
	p=*pub;
	if(!k||!p||!k->key||!p->key||k->curve!=p->curve)goto err1;

	if(k->curve==USICRYPT_X25519)
	{
		if(!(shared=usicrypt_x25519_derive(h,k->key,p->key,&slen)))
			goto err1;
	}
	else
	{
		if(!(shared=usicrypt_ec_derive(h,k->key,p->key,&slen)))
			goto err1;
	}
	for(i=0;i<slen;i++)if(((unsigned char *)shared)[i])break;
	if(i==slen)goto err2;

	if(keys->mode)
	{
		memcpy(salt,p->salt,sizeof(p->salt)>>1);
		memcpy(salt+(sizeof(k->salt)>>1),k->salt,sizeof(k->salt)>>1);
	}
	else
	{
		memcpy(salt,k->salt,sizeof(k->salt)>>1);
		memcpy(salt+(sizeof(p->salt)>>1),p->salt,sizeof(p->salt)>>1);
	}

	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,salt,sizeof(salt),
		keys->mode?"svrgcm":"clngcm",7,kmem))goto err2;
	if(!(txgcm=usicrypt_aeadcipher_init(h,
		conn->alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		kmem,256,nsz,asz)))goto err2;
	if(usicrypt_hkdf(h,USICRYPT_SHA256,shared,slen,salt,sizeof(salt),
		keys->mode?"clngcm":"svrgcm",7,kmem))goto err3;
	if(!(rxgcm=usicrypt_aeadcipher_init(h,
		conn->alt?USICRYPT_CHACHA20_POLY1305:USICRYPT_AES_GCM,
		kmem,256,nsz,asz)))goto err3;
	if(!(c=malloc(sizeof(struct crypto_conn))))goto err4;
	c->cfb8=conn->cfb8;
	c->nonce=conn->nonce;
	c->txgcm=txgcm;
	c->rxgcm=rxgcm;
	c->maxmsg=conn->maxmsg;
	c->maxpad=conn->maxpad;
	c->maxdata=conn->maxdata;
	c->datapad=conn->datapad;
	c->packet=conn->packet;
	c->minmsg=conn->minmsg;
	c->small=conn->small;
	c->strip=conn->strip;
	c->hdr=conn->hdr;
	c->off=conn->off;
	c->base=conn->base;
	c->alt=conn->alt;
	crypto_pktrekeyfree(h,key);
	crypto_pktrekeyfree(h,pub);
	goto err2;

err4:	usicrypt_aeadcipher_exit(rxgcm);
err3:	usicrypt_aeadcipher_exit(txgcm);
err2:	usicrypt_memclear(h,salt,sizeof(salt));
	usicrypt_memclear(h,shared,slen);
	free(shared);
err1:	return c;
}

void *crypto_chkgetpub(void *h,struct crypto_keys *keys,void *msg,int mlen,
	int *plen)
{
	int l1;
	int l2;
	void *res=NULL;

	if(mlen<=2)goto err1;
	if(!(l1=((unsigned char *)msg)[1]))goto err1;
	if(mlen<=2+l1)goto err1;
	if(!(l2=((unsigned char *)msg)[2+l1]))goto err1;
	if(mlen!=3+l1+l2)goto err1;
	if(*((unsigned char *)msg)!=CRYPTO_NEWKEY&&
		*((unsigned char *)msg)!=CRYPTO_REVOKE)goto err1;
	if(usicrypt_ec_verify(h,USICRYPT_SHA384,keys->pub,msg,2+l1,msg+3+l1,l2))
		goto err1;

	switch(usicrypt_pub_type_from_p8(h,msg+2,l1))
	{
	case USICRYPT_BRAINPOOLP512R1:
	case USICRYPT_BRAINPOOLP384R1:
	case USICRYPT_BRAINPOOLP256R1:
	case USICRYPT_SECP521R1:
	case USICRYPT_SECP384R1:
	case USICRYPT_SECP256R1:
		break;
	default:goto err1;
	}

	if(!(res=malloc(l1)))goto err1;
	memcpy(res,msg+2,l1);
	*plen=l1;

err1:	return res;
}

void *crypto_newrvkpub(void *h,struct crypto_keys *keys,void *pub,int plen,
	int mode,int *mlen)
{
	int slen;
	void *sig;
	unsigned char *res=NULL;

	switch(usicrypt_pub_type_from_p8(h,pub,plen))
	{
	case USICRYPT_BRAINPOOLP512R1:
	case USICRYPT_BRAINPOOLP384R1:
	case USICRYPT_BRAINPOOLP256R1:
	case USICRYPT_SECP521R1:
	case USICRYPT_SECP384R1:
	case USICRYPT_SECP256R1:
		break;
	default:goto err1;
	}

	keys->data[0]=(mode?CRYPTO_NEWKEY:CRYPTO_REVOKE);
	keys->data[1]=(unsigned char)plen;
	memcpy(keys->data+2,pub,plen);
	if(!(sig=usicrypt_ec_sign(h,USICRYPT_SHA384,keys->key,keys->data,plen+2,
		&slen)))goto err1;

	*mlen=plen+slen+3;
	if(!(res=malloc(*mlen)))goto err2;
	memcpy(res,keys->data,plen+2);
	res[plen+2]=(unsigned char)slen;
	memcpy(res+plen+3,sig,slen);

err2:	usicrypt_memclear(h,sig,slen);
	free(sig);
err1:	return res;
}

void *crypto_getpub(void *h,struct crypto_keys *keys,int *plen)
{
	return usicrypt_ec_get_pub(h,keys->pub,plen);
}
