/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#ifndef CRYPTO_H_INCLUDED
#define CRYPTO_H_INCLUDED

#define CRYPTO_BRAINPOOLP512R1	0
#define CRYPTO_BRAINPOOLP384R1	1
#define CRYPTO_BRAINPOOLP256R1	2
#define CRYPTO_SECP521R1	3
#define CRYPTO_SECP384R1	4
#define CRYPTO_SECP256R1	5
#define CRYPTO_X25519		6

struct crypto_conf;
struct crypto_x25519;
struct crypto_keys;
struct crypto_setup;
struct crypto_conn;
struct crypto_rekeydata;

extern void *crypto_init(void);
extern void crypto_exit(void *h);

extern void *crypto_thread_init(void *h);
extern void crypto_thread_exit(void *h);

extern void *crypto_genkey(void *h,int curve);

extern struct crypto_conf *crypto_config(void *h,void *ltk,int setup_max,
	int enable_x25519,int enable_secp,int enable_brainpool,int minbits,
	int maxbits,int enable_sig_secp,int enable_sig_brainpool,
	int sig_minbits,int sig_maxbits,int maxdata,int datapad);
extern void crypto_reconfig(void *h,struct crypto_conf *c,void *ltk);

extern void crypto_free(void *h,struct crypto_conf *c);

extern struct crypto_x25519 *crypto_initialrequest(void *h,
	struct crypto_conf *c,void **msg,int *mlen,int rtp);
extern struct crypto_x25519 *crypto_initialreply(void *h,
	struct crypto_conf *c,void *data,int len,void **msg,int *mlen,int rtp);
extern int crypto_initialresponse(void *h,struct crypto_conf *c,
	struct crypto_x25519 *x,void *data,int len,int rtp);
extern struct crypto_conn *crypto_initialderive(void *h,struct crypto_conf *c,
	struct crypto_x25519 *x,int packetmode,int mode,int rtp,int minmsg,
	int small,int strip,int alt);

extern void crypto_initialfree(void *h,struct crypto_conf *c,
	struct crypto_x25519 *x);

extern int crypto_hdrsize(void *h,struct crypto_conn *c);

extern void *crypto_encode(void *h,struct crypto_conn *c,int type,void *data,
	int len,int bits,int *rlen,int rtp);
extern int crypto_unmask(struct crypto_conn *c,void *data,int len,int *rval,
	int *bits);
extern void *crypto_decode(void *h,struct crypto_conn *c,int *type,void *data,
	int len,int *rlen,long long *delta);

extern struct crypto_setup *crypto_realrequest(void *h,struct crypto_conf *c,
	int curve,void **msg,int *mlen);
extern struct crypto_setup *crypto_realreply(void *h,struct crypto_conf *c,
	void *data,int len,void **msg,int *mlen);
extern int crypto_realresponse(void *h,struct crypto_conf *c,
	struct crypto_setup *b,void *data,int len,void **sig,int *siglen);
extern struct crypto_conn *crypto_derivekey(void *h,struct crypto_conf *c,
	struct crypto_setup *b,int packetmode,int rtp,int minmsg,int small,
	int strip,int alt);

extern int crypto_rekeysend(void *h,struct crypto_keys *keys,
	struct crypto_rekeydata **ptr,int curve,void **msg,int *mlen);
extern int crypto_rekeyrecv(void *h,struct crypto_keys *keys,
	struct crypto_rekeydata **ptr,int curve,void *data,int len,void *mem);
extern struct crypto_conn *crypto_rekeyderive(void *h,
	struct crypto_keys *keys,struct crypto_rekeydata **ptr,
	struct crypto_conn *conn,void **msg,int *mlen);
extern void crypto_rekeyfree(void *h,struct crypto_rekeydata **ptr,void *mem);

extern void crypto_freeconn(void *h,struct crypto_conn *c);
extern void crypto_freepair(void *h,struct crypto_conn **c,int which);

extern int crypto_authrequest(void *h,struct crypto_conf *c,
	struct crypto_setup *b,void *auth,void **msg,int *mlen,int rekey);
extern int crypto_authparse(void *h,struct crypto_conf *c,
	struct crypto_setup *b,void *data,int len);
extern int crypto_wantsrekey(void *h,struct crypto_conf *c,
	struct crypto_setup *b);
extern int crypto_anonymousauth(void *h,struct crypto_conf *c,
	struct crypto_setup *b);
extern void *crypto_getauthdata(void *h,struct crypto_conf *c,
	struct crypto_setup *b,int *len);
extern int crypto_authverify(void *h,struct crypto_conf *c,
	struct crypto_setup *b);
extern int crypto_authreply(void *h,struct crypto_conf *c,
	struct crypto_setup *b,int ok,void **msg,int *mlen);
extern int crypto_authresponse(void *h,struct crypto_conf *c,
	struct crypto_setup *b,void *data,int len);
extern struct crypto_conn *crypto_authderive(void *h,struct crypto_conf *c,
	struct crypto_setup *b,int packetmode,int rtp,int minmsg,int small,
	int strip,int alt);

extern int crypto_getcurve(void *h,struct crypto_conf *c,
	struct crypto_setup *b);

extern void crypto_setupfree(void *h,struct crypto_conf *c,
	struct crypto_setup *b,struct crypto_keys **k);
extern void crypto_keysfree(void *h,struct crypto_keys *k);

extern void crypto_memclear(void *h,void *dasta,int len);

extern void *crypto_read_pem_key(void *h,void *data,int len,void *passwd,
	int plen);
extern void *crypto_make_pem_key(void *h,void *key,int *len,void *passwd,
	int plen);
extern void *crypto_read_pem_pub(void *h,void *data,int len,int *publen);
extern void *crypto_write_pem_pub(void *h,void *data,int len,int *pemlen);
extern void *crypto_make_pem_pub(void *h,void *key,int *len);

extern void crypto_free_key(void *h,void *key);

extern int crypto_random(void *h,void *data,int len);

extern int crypto_portrequest(void *h,void *data,int *len,int min);
extern int crypto_portreply(void *h,void *data,int *len,void *port,int min,
	int alt);
extern int crypto_portresponse(void *h,void *snd,int slen,void *rcv,int rlen,
	void *port,int alt);

extern int crypto_noopmsg(void *h,void **data,int *len);

extern int crypto_pktfinmsg(void *h,struct crypto_keys *keys,void *data,
	int *len);
extern int crypto_pktackmsg(void *h,struct crypto_keys *keys,void **data,
	int *len,void *pub);
extern int crypto_pkttype(void *h,struct crypto_keys *keys,void *data,int len,
	void *key);
extern int crypto_pktrekeygen(void *h,void **key,int curve);
extern void crypto_pktrekeyfree(void *h,void **key);
extern int crypto_pktrekeytxmsg(void *h,struct crypto_keys *keys,void *key,
		void **msg,int *mlen);
extern int crypto_pktrekeyrxmsg(void *h,struct crypto_keys *keys,void **key,
	void *msg,int mlen);
extern struct crypto_conn *crypto_pktrekeyderive(void *h,
	struct crypto_keys *keys,void **key,void **pub,
	struct crypto_conn *conn);

extern void *crypto_chkgetpub(void *h,struct crypto_keys *keys,void *msg,
	int mlen,int *plen);
extern void *crypto_newrvkpub(void *h,struct crypto_keys *keys,void *pub,
	int plen,int mode,int *mlen);
extern void *crypto_getpub(void *h,struct crypto_keys *keys,int *plen);

#endif
