/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/timerfd.h>
#include <sys/eventfd.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>
#include <poll.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "dns.h"

static void notifier(union sigval arg)
{
	int *nfd=(int *)arg.sival_ptr;
	uint64_t dummy=1;
	dummy=write(*nfd,&dummy,sizeof(dummy));
}

void *lookup(char *name,int port,int timeout,int prefer_v4,int efd,int *len)
{
	int nfd;
	int tfd;
	union addr
	{
		struct sockaddr_in a4;
		struct sockaddr_in6 a6;
	} *a;
	struct addrinfo *rp;
	struct gaicb *gcptr;
	struct gaicb gc;
	struct addrinfo hints;
	struct sigevent sp;
	struct itimerspec it;
	struct pollfd p[3];

	if(!name||port<=0||port>65535||timeout<=0||timeout>60)goto err1;

	if(!(a=malloc(sizeof(union addr))))goto err1;
	memset(a,0,sizeof(union addr));

	if(inet_pton(AF_INET,name,&a->a4.sin_addr)==1)
	{
		a->a4.sin_family=AF_INET;
		a->a4.sin_port=htons(port);
		*len=sizeof(struct sockaddr_in);
		return a;
	}

	if(inet_pton(AF_INET6,name,&a->a6.sin6_addr)==1)
	{
		a->a6.sin6_family=AF_INET6;
		a->a6.sin6_port=htons(port);
		*len=sizeof(struct sockaddr_in6);
		return a;
	}

	memset(&hints,0,sizeof(hints));
	hints.ai_family=AF_UNSPEC;

	memset(&gc,0,sizeof(gc));
	gc.ar_name=name;
	gc.ar_request=&hints;
	gcptr=&gc;

	memset(&sp,0,sizeof(sp));
	sp.sigev_notify=SIGEV_THREAD;
	sp.sigev_notify_function=notifier;
	sp.sigev_value.sival_ptr=&nfd;

	memset(&it,0,sizeof(it));
	it.it_value.tv_sec=timeout;

	if((nfd=eventfd(0,EFD_CLOEXEC|EFD_NONBLOCK))==-1)goto err2;
	if((tfd=timerfd_create(CLOCK_MONOTONIC,TFD_CLOEXEC|TFD_NONBLOCK))==-1)
		goto err3;
	if(timerfd_settime(tfd,0,&it,NULL))goto err4;

	p[0].fd=nfd;
	p[0].events=POLLIN;
	p[1].fd=tfd;
	p[1].events=POLLIN;
	p[2].fd=efd;
	p[2].events=POLLIN;

	if(getaddrinfo_a(GAI_NOWAIT,&gcptr,1,&sp))goto err4;

	while(1)if(poll(p,3,-1)>=1)break;

	if(p[0].revents&POLLIN)
	{
		if(gai_error(&gc))goto err5;
	}
	else if(p[1].revents&POLLIN)goto err5;
	else if(p[2].revents&POLLIN)goto err5;

	for(rp=gc.ar_result;rp;rp=rp->ai_next)switch(rp->ai_family)
	{
	case AF_INET6:
		if(a->a6.sin6_family==AF_INET6)break;
		if(a->a4.sin_family==AF_INET&&prefer_v4)break;
		a->a6.sin6_family=AF_INET6;
		a->a6.sin6_addr=((struct sockaddr_in6 *)rp->ai_addr)->sin6_addr;
		a->a6.sin6_port=htons(port);
		break;

	case AF_INET:
		if(a->a4.sin_family==AF_INET)break;
		if(a->a6.sin6_family==AF_INET6&&!prefer_v4)break;
		a->a4.sin_family=AF_INET;
		a->a4.sin_addr=((struct sockaddr_in *)rp->ai_addr)->sin_addr;
		a->a4.sin_port=htons(port);
		break;
	}

	if(gc.ar_result)freeaddrinfo(gc.ar_result);
	if(a->a4.sin_family==AF_INET)*len=sizeof(struct sockaddr_in);
	else if(a->a6.sin6_family==AF_INET6)*len=sizeof(struct sockaddr_in6);
	else goto err4;
	close(tfd);
	close(nfd);

	return a;

err5:	gai_cancel(&gc);
	if(gc.ar_result)freeaddrinfo(gc.ar_result);
err4:	close(tfd);
err3:	close(nfd);
err2:	free(a);
err1:	return NULL;
}
