/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/eventfd.h>
#include <sys/signalfd.h>
#include <sys/wait.h>
#include <sys/capability.h>
#include <sys/prctl.h>
#include <sys/ptrace.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/utsname.h>
#include <pwd.h>
#include <grp.h>
#include <sched.h>
#include <signal.h>
#include <poll.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include "misc.h"
#include "crypto.h"
#include "parse.h"
#include "sctp_om.h"
#include "client.h"
#include "server.h"
#include "keygen.h"

#ifndef MCL_ONFAULT
#define MCL_ONFAULT 0
#endif

COLD static int setrsrc(int mode)
{
	struct rlimit lim;

	if(prctl(PR_SET_DUMPABLE,0,0,0,0))return -1;
	lim.rlim_cur=lim.rlim_max=0;
	if(setrlimit(RLIMIT_CORE,&lim))return -1;
	if(mode)
	{
		lim.rlim_cur=lim.rlim_max=RLIM_INFINITY;
		if(setrlimit(RLIMIT_CPU,&lim))return -1;
		if(setrlimit(RLIMIT_MEMLOCK,&lim))return -1;
		if(setrlimit(RLIMIT_NPROC,&lim))return -1;
		lim.rlim_cur=lim.rlim_max=0xfffff;
		if(setrlimit(RLIMIT_NOFILE,&lim))return -1;
	}
#ifndef DEBUG
	if(prctl(PR_TASK_PERF_EVENTS_DISABLE,0,0,0,0))return -1;
#endif
	return 0;
}

COLD static int setrt(void)
{
	struct sched_param sched;

	if(mlockall(MCL_CURRENT|MCL_FUTURE|MCL_ONFAULT))return -1;
	memset(&sched,0,sizeof(sched));
	sched.sched_priority=1;
	if(sched_setscheduler(0,SCHED_RR|SCHED_RESET_ON_FORK,&sched))return -1;
	return 0;
}

COLD static int setprivs(char *user,int sysadm,int caps,int ugid,uid_t ruid,
	uid_t suid)
{
#ifdef PR_CAP_AMBIENT
	int flag=0;
	int major;
	int minor;
	char *ptr;
	struct utsname u;
#endif
	struct passwd *pwd;
	struct __user_cap_header_struct h;
	struct __user_cap_data_struct c[2];

#ifdef PR_CAP_AMBIENT
	if(uname(&u))return -1;
	if(!(ptr=strtok(u.release,".")))return -1;
	major=atoi(ptr);
	if(!(ptr=strtok(NULL,".-")))return -1;
	minor=atoi(ptr);
	if(major>4||(major==4&&minor>=3))flag=1;
#endif
	if(!(pwd=getpwnam(user)))return -1;
	if(ruid||suid)if(pwd->pw_uid!=ruid||pwd->pw_uid!=suid)return -1;
	memset(&h,0,sizeof(h));
	h.version=_LINUX_CAPABILITY_VERSION;
	h.pid=0;
	memset(&c,0,sizeof(c));
	c[CAP_TO_INDEX(CAP_NET_ADMIN)].effective|=CAP_TO_MASK(CAP_NET_ADMIN);
	c[CAP_TO_INDEX(CAP_NET_BIND_SERVICE)].effective|=
		CAP_TO_MASK(CAP_NET_BIND_SERVICE);
	c[CAP_TO_INDEX(CAP_NET_RAW)].effective|=CAP_TO_MASK(CAP_NET_RAW);
	c[CAP_TO_INDEX(CAP_NET_BROADCAST)].effective|=
		CAP_TO_MASK(CAP_NET_BROADCAST);
	c[CAP_TO_INDEX(CAP_SETUID)].effective|=CAP_TO_MASK(CAP_SETUID);
	c[CAP_TO_INDEX(CAP_SETGID)].effective|=CAP_TO_MASK(CAP_SETGID);
	c[CAP_TO_INDEX(CAP_SYS_ADMIN)].effective|=CAP_TO_MASK(CAP_SYS_ADMIN);
	c[CAP_TO_INDEX(CAP_IPC_LOCK)].effective|=CAP_TO_MASK(CAP_IPC_LOCK);
	c[CAP_TO_INDEX(CAP_SYS_NICE)].effective|=CAP_TO_MASK(CAP_SYS_NICE);
	c[CAP_TO_INDEX(CAP_SYS_RESOURCE)].effective|=
		CAP_TO_MASK(CAP_SYS_RESOURCE);
	c[CAP_TO_INDEX(CAP_SYS_TIME)].effective|=CAP_TO_MASK(CAP_SYS_TIME);
	c[CAP_TO_INDEX(CAP_KILL)].effective|=CAP_TO_MASK(CAP_KILL);
	c[CAP_TO_INDEX(CAP_SETPCAP)].effective|=CAP_TO_MASK(CAP_SETPCAP);
	c[0].permitted=c[0].effective;
	c[1].permitted=c[1].effective;
	c[0].inheritable=c[0].effective;
	c[1].inheritable=c[1].effective;
	if(capset(&h,c))return -1;
	if(initgroups(user,pwd->pw_gid))return -1;
	if(prctl(PR_SET_KEEPCAPS,1,0,0,0))return -1;
	if(setresgid(pwd->pw_gid,pwd->pw_gid,pwd->pw_gid))return -1;
	if(setresuid(pwd->pw_uid,pwd->pw_uid,pwd->pw_uid))return -1;
#ifdef PR_CAP_AMBIENT
	if(flag)
	{
		if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,CAP_NET_ADMIN,0,0))
			return -1;
		if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,
			CAP_NET_BIND_SERVICE,0,0))return -1;
		if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,CAP_NET_RAW,0,0))
			return -1;
		if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,CAP_NET_BROADCAST,
			0,0))return -1;
		if(ugid)
		{
			if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,CAP_SETUID,
				0,0))return -1;
			if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,CAP_SETGID,
				0,0))return -1;
		}
		if(sysadm)
		{
			if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,
				CAP_SYS_ADMIN,0,0))return -1;
			if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,
				CAP_IPC_LOCK,0,0))return -1;
			if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,
				CAP_SYS_NICE,0,0))return -1;
			if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,
				CAP_SYS_RESOURCE,0,0))return -1;
			if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,
				CAP_SYS_TIME,0,0))return -1;
			if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,CAP_KILL,
				0,0))return -1;
		}
		if(caps)if(prctl(PR_CAP_AMBIENT,PR_CAP_AMBIENT_RAISE,
				CAP_SETPCAP,0,0))return -1;
	}
#endif
	memset(&h,0,sizeof(h));
	h.version=_LINUX_CAPABILITY_VERSION;
	h.pid=0;
	memset(&c,0,sizeof(c));
	c[CAP_TO_INDEX(CAP_NET_ADMIN)].permitted|=CAP_TO_MASK(CAP_NET_ADMIN);
	c[CAP_TO_INDEX(CAP_NET_BIND_SERVICE)].permitted|=
		CAP_TO_MASK(CAP_NET_BIND_SERVICE);
	c[CAP_TO_INDEX(CAP_NET_RAW)].permitted|=CAP_TO_MASK(CAP_NET_RAW);
	c[CAP_TO_INDEX(CAP_NET_BROADCAST)].permitted|=
		CAP_TO_MASK(CAP_NET_BROADCAST);
	if(ugid)
	{
		c[CAP_TO_INDEX(CAP_SETUID)].permitted|=CAP_TO_MASK(CAP_SETUID);
		c[CAP_TO_INDEX(CAP_SETGID)].permitted|=CAP_TO_MASK(CAP_SETGID);
	}
	if(sysadm)
	{
		c[CAP_TO_INDEX(CAP_SYS_ADMIN)].permitted|=
			CAP_TO_MASK(CAP_SYS_ADMIN);
		c[CAP_TO_INDEX(CAP_IPC_LOCK)].permitted|=
			CAP_TO_MASK(CAP_IPC_LOCK);
		c[CAP_TO_INDEX(CAP_SYS_NICE)].permitted|=
			CAP_TO_MASK(CAP_SYS_NICE);
		c[CAP_TO_INDEX(CAP_SYS_RESOURCE)].permitted|=
			CAP_TO_MASK(CAP_SYS_RESOURCE);
		c[CAP_TO_INDEX(CAP_SYS_TIME)].permitted|=
			CAP_TO_MASK(CAP_SYS_TIME);
		c[CAP_TO_INDEX(CAP_KILL)].permitted|=CAP_TO_MASK(CAP_KILL);
	}
	if(caps)c[CAP_TO_INDEX(CAP_SETPCAP)].permitted|=
		CAP_TO_MASK(CAP_SETPCAP);
	c[0].inheritable=c[0].permitted;
	c[1].inheritable=c[1].permitted;
	if(capset(&h,c))return -1;
	if(prctl(PR_SET_KEEPCAPS,0,0,0,0))return -1;
	return 0;
}

static void *loadpass(void *glob,int mode,int *plen)
{
	void *h;
	void *r=NULL;
	char *pass;
	char *pass1=NULL;
	char *pass2=NULL;
	char *ptr;
	char bfr[256];

	if(!(h=crypto_thread_init(glob)))return NULL;

	switch(mode)
	{
	case 0:	if((pass=getenv("VPN_PASSWD")))
		{
			pass1=strdup(pass);
			for(ptr=pass;*ptr;ptr++)*ptr=' ';
			unsetenv("VPN_PASSWD");
			if(!pass1||!*pass1)goto fail;
			break;
		}
		if(!fgets(bfr,sizeof(bfr),stdin))goto fail;
		for(ptr=bfr;*ptr;ptr++)if(*ptr=='\r'||*ptr=='\n')
		{
			*ptr=0;
			break;
		}
		pass1=strdup(bfr);
		if(!pass1||!*pass1)goto fail;
		break;

	case 1:	pass=getpass("passphrase: ");
		pass1=strdup(pass);
		crypto_memclear(h,pass,strlen(pass));
		if(!pass1||!*pass1)goto fail;
		break;

	case 2:	pass=getpass("passphrase: ");
		pass1=strdup(pass);
		crypto_memclear(h,pass,strlen(pass));
		pass=getpass("passphrase (again): ");
		pass2=strdup(pass);
		crypto_memclear(h,pass,strlen(pass));
		if(!pass1||!*pass1||!pass2||!*pass2)goto fail;
		if(strcmp(pass1,pass2))goto fail;
		break;
	}

	*plen=strlen(pass1);
	r=pass1;

fail:	if(!r&&pass1)
	{
		crypto_memclear(h,pass1,strlen(pass1));
		free(pass1);
	}

	if(pass2)
	{
		crypto_memclear(h,pass2,strlen(pass2));
		free(pass2);
	}

	crypto_memclear(h,bfr,sizeof(bfr));

	crypto_thread_exit(h);

	return r;
}

static void clearpass(void *glob,char *pass,int plen)
{
	void *h;

	if(!pass)return;
	if(!(h=crypto_thread_init(glob)))while(plen--)*pass++=0;
	else
	{
		crypto_memclear(h,pass,plen);
		crypto_thread_exit(h);
	}
}

static int writepid(char *fn)
{
	FILE *fp;

	if(!(fp=fopen(fn,"we")))return -1;
	fprintf(fp,"%d\n",getpid());
	fclose(fp);
	return 0;
}

int main(int argc,char *argv[])
{
	int c;
	int sfd;
	int cefd;
	int sefd;
	int status;
	int res=1;
	int fg=1;
	int rt=0;
	int enc=0;
	int enable=1;
	int plen;
	uid_t ruid;
	uid_t euid;
	uid_t suid;
	uint64_t dummy;
	char *pass=NULL;
	char *config="/etc/vpnd2.conf";
	char *which=NULL;
	void *glob;
	void *ctx1=NULL;
	void *ctx2=NULL;
	struct pollfd p;
	struct signalfd_siginfo sig;
	struct global g;
	sigset_t msk;

#ifndef DEBUG
	if(ptrace(PTRACE_TRACEME,0,0,0)||getenv("LD_PRELOAD")||
		getenv("LD_LIBRARY_PATH"))
	{
		fprintf(stderr,"refusing insecure processing\n");
		goto err1;
	}
#endif

	while((c=getopt(argc,argv,"c:k:K:dpP"))!=-1)switch(c)
	{
	case 'p':
		enc=1;
		break;

	case 'P':
		enc=2;
		break;

	case 'c':
		config=optarg;
		break;

	case 'K':
		rt=1;
	case 'k':
		which=optarg;
		break;

	case 'd':
		fg=0;
		break;

	default:fprintf(stderr,"unknown option\n");
		goto err1;
	}

	if(!(glob=crypto_init()))
	{
		fprintf(stderr,"crypto init failed\n");
		goto err1;
	}

	switch(enc)
	{
	case 1:	if(!(pass=loadpass(glob,0,&plen)))
		{
			fprintf(stderr,"no passphrase\n");
			goto err2;
		}
		break;

	case 2:	if(!(pass=loadpass(glob,which?2:1,&plen)))
		{
			fprintf(stderr,"no passphrase or mismatch\n");
			goto err2;
		}
		break;
	}

	if(which)
	{
		if(optind!=argc-1)
			fprintf(stderr,"missing or illegal key pathname\n");
		else if(setrsrc(0))
			fprintf(stderr,"resource setup failed\n");
		else if(rt&&setrt())
			fprintf(stderr,"realtime setup failed\n");
		else if(setegid(getgid())||seteuid(getuid()))
			fprintf(stderr,"privilege setup failed\n");
		else if(generate(argv[optind],which,glob,pass,plen))
			fprintf(stderr,"creating keypair failed\n");
		else res=0;
		goto err2;
	}
	else if(optind!=argc)
	{
		fprintf(stderr,"excess parameter(s)\n");
		goto err2;
	}

	if(getresuid(&ruid,&euid,&suid))
	{
		perror("getresuid");
		goto err1;
	}

	sigfillset(&msk);
	sigprocmask(SIG_BLOCK,&msk,NULL);
	sigemptyset(&msk);
	sigaddset(&msk,SIGINT);
	sigaddset(&msk,SIGHUP);
	sigaddset(&msk,SIGQUIT);
	sigaddset(&msk,SIGTERM);
	sigaddset(&msk,SIGCHLD);
	sigaddset(&msk,SIGUSR1);
	sigaddset(&msk,SIGUSR2);

	if((sfd=signalfd(-1,&msk,SFD_NONBLOCK|SFD_CLOEXEC))==-1)goto err2;
	if((cefd=eventfd(0,EFD_CLOEXEC|EFD_NONBLOCK))==-1)goto err3;
	if((sefd=eventfd(0,EFD_CLOEXEC|EFD_NONBLOCK))==-1)goto err4;

	p.fd=sfd;
	p.events=POLLIN;

	if(setrsrc(1))
	{
		fprintf(stderr,"resource setup failed\n");
		goto err5;
	}
	if(parseconfig(config,&g,glob,cefd,sefd))goto err5;
	if(ruid||suid)if(g.caps||g.ugid)
	{
		fprintf(stderr,"root only configuration forbidden\n");
		goto err6;
	}
	sctpomlockselect();
	if(g.isclient)if(!(ctx1=client_init(&g,pass,plen)))
	{
		fprintf(stderr,"client setup failed\n");
		goto err6;
	}
	if(g.isserver)if(!(ctx2=server_init(&g,pass,plen)))
	{
		fprintf(stderr,"server setup failed\n");
		goto err7;
	}
	if(pass)
	{
		clearpass(glob,pass,plen);
		pass=NULL;
		enable=0;
	}
	else if(!g.newpub||!g.revoke)enable=0;
	if(g.realtime)if(setrt())
	{
		fprintf(stderr,"realtime setup failed\n");
		goto err8;
	}
	if(g.user)if(setprivs(g.user,g.sysadm,g.caps,g.ugid,ruid,suid))
	{
		fprintf(stderr,"privilege setup failed\n");
		goto err8;
	}
	if(g.isserver)if(server_unix_init(ctx2))
	{
		fprintf(stderr,"server setup failed\n");
		goto err8;
	}
	if(!fg)if(daemon(0,0))
	{
		fprintf(stderr,"becoming daemon failed\n");
		goto err8;
	}
	if(ruid||suid)if(prctl(PR_SET_NO_NEW_PRIVS,0,0,0,0))
	{
		fprintf(stderr,"privilege locking failed\n");
		goto err8;
	}
	if(ctx1)if(client_start(ctx1,1))
	{
		fprintf(stderr,"client start failed\n");
		goto err8;
	}
	if(ctx2)if(server_start(ctx2,1))
	{
		fprintf(stderr,"server start failed\n");
		goto err10;
	}

	if(g.pidfile)if(writepid(g.pidfile))
	{
		fprintf(stderr,"pid file write failed\n");
		goto err11;
	}

#ifdef DEBUG
	prctl(PR_SET_NAME,"main",0,0,0,0);
#endif

	res=0;

	while(1)
	{
		if(poll(&p,1,-1)!=1)continue;
		if(read(sfd,&sig,sizeof(sig))!=sizeof(sig))continue;
		if(sig.ssi_signo==SIGCHLD)while(waitpid(0,&status,WNOHANG)>0);
		else if(sig.ssi_signo==SIGUSR1)
		{
			if(enable&&ctx2)
			{
				dummy=1;
				dummy=write(sefd,&dummy,sizeof(dummy));
				server_stop(ctx2,0);
				server_keyupdate(ctx2);
				dummy=read(sefd,&dummy,sizeof(dummy));
				if(server_start(ctx2,0))goto err10;
			}
		}
		else if(sig.ssi_signo==SIGUSR2)
		{
			if(ctx1)
			{
				dummy=1;
				dummy=write(cefd,&dummy,sizeof(dummy));
				client_stop(ctx1,0);
				client_keyupdate(ctx1);
				dummy=read(cefd,&dummy,sizeof(dummy));
				if(client_start(ctx1,0))
				{
					dummy=1;
					dummy=write(sefd,&dummy,sizeof(dummy));
					if(ctx2)server_stop(ctx2,1);
					goto err9;
				}
			}
		}
		else break;
	}

	if(g.pidfile)unlink(g.pidfile);

err11:	dummy=1;
	dummy=write(sefd,&dummy,sizeof(dummy));
	if(ctx2)server_stop(ctx2,1);

err10:	dummy=1;
	dummy=write(cefd,&dummy,sizeof(dummy));
	if(ctx1)client_stop(ctx1,1);

err9:	while(poll(&p,1,0)==1)if(read(sfd,&sig,sizeof(sig))==sizeof(sig))
		if(sig.ssi_signo==SIGCHLD)while(waitpid(0,&status,WNOHANG)>0);

err8:	if(ctx2)server_exit(ctx2);
err7:	if(ctx1)client_exit(ctx1);
err6:	freeconfig(&g);
err5:	close(sefd);
err4:	close(cefd);
err3:	close(sfd);
err2:	if(pass)clearpass(glob,pass,plen);
	crypto_exit(glob);

err1:	return res;
}
