/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/stat.h>
#include <stddef.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "misc.h"
#include "proto.h"
#include "parse.h"

#define STRING		0
#define FILENAME	1
#define FILENAME2	2
#define NUMBER		3
#define BOOLEAN		4
#define PROTOCOL	5
#define LISTEN		6

#define MINBIT		256
#define MAXBIT		521

struct parsecfg
{
	const char *name;
	const int type;
	const int min;
	const int max;
	const int offset;
};


static const struct parsecfg globcfg[]=
{
	{"setuptimeout",NUMBER,1,10,offsetof(struct global,setuptimeout)},
	{"listen",LISTEN,0,0,offsetof(struct global,sockets)},
	{"key",FILENAME,0,0,offsetof(struct global,keyfile)},
	{"newpub",FILENAME,1,0,offsetof(struct global,newpub)},
	{"revoke",FILENAME,1,0,offsetof(struct global,revoke)},
	{"setuplimit",NUMBER,512,4095,offsetof(struct global,setuplimit)},
	{"pendinglimit",NUMBER,1,262144,offsetof(struct global,pendinglimit)},
	{"keepalive",NUMBER,15,120,offsetof(struct global,keepalive)},
	{"keysecp",BOOLEAN,0,0,offsetof(struct global,sigsecp)},
	{"keybrainpool",BOOLEAN,0,0,offsetof(struct global,sigbrainpool)},
	{"keybitsmin",NUMBER,MINBIT,MAXBIT,offsetof(struct global,sigminbits)},
	{"keybitsmax",NUMBER,MINBIT,MAXBIT,offsetof(struct global,sigmaxbits)},
	{"x25519",BOOLEAN,0,0,offsetof(struct global,x25519)},
	{"secp",BOOLEAN,0,0,offsetof(struct global,secp)},
	{"brainpool",BOOLEAN,0,0,offsetof(struct global,brainpool)},
	{"bitsmin",NUMBER,MINBIT,MAXBIT,offsetof(struct global,minbits)},
	{"bitsmax",NUMBER,MINBIT,MAXBIT,offsetof(struct global,maxbits)},
	{"dosysadm",BOOLEAN,0,0,offsetof(struct global,sysadm)},
	{"chgcaps",BOOLEAN,0,0,offsetof(struct global,caps)},
	{"setugid",BOOLEAN,0,0,offsetof(struct global,ugid)},
	{"user",STRING,0,0,offsetof(struct global,user)},
	{"realtime",BOOLEAN,0,0,offsetof(struct global,realtime)},
	{"padlimit",NUMBER,0,4095,offsetof(struct global,maxdata)},
	{"padding",NUMBER,0,4095,offsetof(struct global,datapad)},
	{"sendbuf",NUMBER,1024,524288,offsetof(struct global,sendbuf)},
	{"maxseg",NUMBER,512,65535,offsetof(struct global,maxseg)},
	{"tcpdefer",NUMBER,1,60,offsetof(struct global,tcpdefer)},
	{"tcpacktimeout",NUMBER,1,120,offsetof(struct global,tcpuser)},
	{"fragments",BOOLEAN,0,0,offsetof(struct global,frag)},
	{"minport",NUMBER,1024,65535,offsetof(struct global,portmin)},
	{"maxport",NUMBER,1024,65535,offsetof(struct global,portmax)},
	{"altcrypto",BOOLEAN,0,0,offsetof(struct global,alt)},
	{"pidfile",FILENAME,1,0,offsetof(struct global,pidfile)},
	{NULL,0,0,0,0}
};

static const struct parsecfg srvcln[]=
{
	{"server",STRING,0,0,offsetof(struct config,server)},
	{"port",NUMBER,1,65535,offsetof(struct config,port)},
	{"proto",PROTOCOL,0,0,offsetof(struct config,proto)},
	{"dnstimeout",NUMBER,1,60,offsetof(struct config,dnstimeout)},
	{"connecttimeout",NUMBER,1,60,offsetof(struct config,conntimeout)},
	{"retrydelay",NUMBER,1,60,offsetof(struct config,delay)},
	{"key",FILENAME,0,0,offsetof(struct config,keyfile)},
	{"keysecp",BOOLEAN,0,0,offsetof(struct config,sigsecp)},
	{"keybrainpool",BOOLEAN,0,0,offsetof(struct config,sigbrainpool)},
	{"keybitsmin",NUMBER,MINBIT,MAXBIT,offsetof(struct config,sigminbits)},
	{"keybitsmax",NUMBER,MINBIT,MAXBIT,offsetof(struct config,sigmaxbits)},
	{"x25519",BOOLEAN,0,0,offsetof(struct config,x25519)},
	{"secp",BOOLEAN,0,0,offsetof(struct config,secp)},
	{"brainpool",BOOLEAN,0,0,offsetof(struct config,brainpool)},
	{"bitsmin",NUMBER,MINBIT,MAXBIT,offsetof(struct config,minbits)},
	{"bitsmax",NUMBER,MINBIT,MAXBIT,offsetof(struct config,maxbits)},
	{"setuplimit",NUMBER,512,4095,offsetof(struct config,setuplimit)},
	{"script",FILENAME,0,1,offsetof(struct config,script)},
	{"auth",FILENAME2,0,0,offsetof(struct config,authfile)},
	{"rohc",BOOLEAN,0,0,offsetof(struct config,rohc)},
	{"rohccid",NUMBER,16,16383,offsetof(struct config,rohccid)},
	{"preferv4",BOOLEAN,0,0,offsetof(struct config,prefv4)},
	{"rekey",NUMBER,60,86400,offsetof(struct config,rekey)},
	{"sendbuf",NUMBER,1024,524288,offsetof(struct config,sendbuf)},
	{"maxseg",NUMBER,512,65535,offsetof(struct config,maxseg)},
	{"tcpskipack",BOOLEAN,0,0,offsetof(struct config,tcpdefer)},
	{"fragments",BOOLEAN,0,0,offsetof(struct config,frag)},
	{"small",BOOLEAN,0,0,offsetof(struct config,small)},
	{"strip",BOOLEAN,0,0,offsetof(struct config,strip)},
	{"paranoia",BOOLEAN,0,0,offsetof(struct config,paranoia)},
	{"noop",BOOLEAN,0,0,offsetof(struct config,noop)},
	{"sockshost",STRING,0,0,offsetof(struct config,sockshost)},
	{"socksport",NUMBER,1,65535,offsetof(struct config,socksport)},
	{"proxyhost",STRING,0,0,offsetof(struct config,proxyhost)},
	{"proxyport",NUMBER,1,65535,offsetof(struct config,proxyport)},
	{"helpercmd",FILENAME,0,1,offsetof(struct config,helpercmd)},
	{"helperconf",FILENAME,1,1,offsetof(struct config,helpercnf)},
	{"rekeyfail",BOOLEAN,0,0,offsetof(struct config,rekeyfail)},
	{"connection",BOOLEAN,0,0,offsetof(struct config,connection)},
	{"sctpsrc",BOOLEAN,0,0,offsetof(struct config,srcaddr)},
	{NULL,0,0,0,0}
};

COLD static int parseline(char *line,int num,const struct parsecfg *items,
	void *base)
{
	int i;
	int proto;
	int small;
	int strip;
	unsigned long number;
	char *mem;
	char *item;
	char *value;
	char *addr;
	char *ptr;
	char *xtra;
	struct listen *l;
	struct stat stb;

	item=strtok_r(line," \t\r\n",&mem);
	value=strtok_r(NULL," \t\r\n",&mem);

	if(!item||!value||!*item||!*value)goto err1;
	for(i=0;items[i].name;i++)if(!strcmp(item,items[i].name))break;
	if(!items[i].name)goto err1;

	switch(items[i].type)
	{
	case STRING:
		if(!(ptr=strdup(value)))goto err2;
		*((char **)(base+items[i].offset))=ptr;
		break;

	case FILENAME:
		if(items[i].min&&!items[i].max)
		{
			if(!(ptr=strrchr(value,'/')))goto err1;
			*ptr=0;
			if(*value!='/')goto err1;
			if(stat(value,&stb)||!S_ISDIR(stb.st_mode))goto err1;
			if(access(value,R_OK))goto err1;
			if(access(value,W_OK))goto err1;
			if(access(value,X_OK))goto err1;
			*ptr='/';
			if(!stat(value,&stb)&&!S_ISREG(stb.st_mode))goto err1;
		}
		else if(!items[i].min||!items[i].max)
		{
			if(stat(value,&stb)||!S_ISREG(stb.st_mode))goto err1;
			if(access(value,R_OK))goto err1;
			if(items[i].max&&access(value,X_OK))goto err1;
		}
		if(!(ptr=strdup(value)))goto err2;
		*((char **)(base+items[i].offset))=ptr;
		break;

	case FILENAME2:
		xtra=strtok_r(NULL," \t\r\n",&mem);
		if(xtra)
		{
			if(*xtra!='/'||*value!='/')goto err1;
			if(stat(value,&stb))
			{
				if(stat(xtra,&stb))goto err1;
				if(!S_ISREG(stb.st_mode))goto err1;
				if(access(xtra,R_OK))goto err1;
				if(access(xtra,W_OK))goto err1;
				ptr=strrchr(value,'/');
				*ptr=0;
				if(!*value)goto err1;
				if(stat(value,&stb))goto err1;
				if(!S_ISDIR(stb.st_mode))goto err1;
				if(access(value,R_OK))goto err1;
				if(access(value,W_OK))goto err1;
				if(access(value,X_OK))goto err1;
				*ptr='/';
			}
			else
			{
				if(!S_ISREG(stb.st_mode))goto err1;
				if(access(value,R_OK))goto err1;
				if(access(value,W_OK))goto err1;
			}
			if(stat(xtra,&stb))
			{
				if(stat(value,&stb))goto err1;
				if(!S_ISREG(stb.st_mode))goto err1;
				if(access(value,R_OK))goto err1;
				if(access(value,W_OK))goto err1;
				ptr=strrchr(xtra,'/');
				*ptr=0;
				if(!*xtra)goto err1;
				if(stat(xtra,&stb))goto err1;
				if(!S_ISDIR(stb.st_mode))goto err1;
				if(access(xtra,R_OK))goto err1;
				if(access(xtra,W_OK))goto err1;
				if(access(xtra,X_OK))goto err1;
				*ptr='/';
			}
			else
			{
				if(!S_ISREG(stb.st_mode))goto err1;
				if(access(xtra,R_OK))goto err1;
				if(access(xtra,W_OK))goto err1;
			}
		}
		else
		{
			if(stat(value,&stb)||!S_ISREG(stb.st_mode))goto err1;
			if(access(value,R_OK))goto err1;
		}
		if(!(ptr=strdup(value)))goto err2;
		((char **)(base+items[i].offset))[0]=ptr;
		if(!xtra)ptr=NULL;
		else if(!(ptr=strdup(xtra)))goto err2;
		((char **)(base+items[i].offset))[1]=ptr;
		break;

	case NUMBER:
		if((number=strtoul(value,&mem,10))==ULONG_MAX||*mem)goto err1;
		if(number<items[i].min||number>items[i].max)goto err1;
		*((unsigned int *)(base+items[i].offset))=(unsigned int)number;
		break;

	case BOOLEAN:
		if(!strcmp(value,"enable"))
			*((unsigned int *)(base+items[i].offset))=1;
		else if(!strcmp(value,"disable"))
			*((unsigned int *)(base+items[i].offset))=0;
		else goto err1;
		break;

	case PROTOCOL:
		if(!strcmp(value,"tcp"))
			*((unsigned int *)(base+items[i].offset))=TCP;
		else if(!strcmp(value,"tcp-socks5"))
			*((unsigned int *)(base+items[i].offset))=SOCKS5;
		else if(!strcmp(value,"tcp-proxy"))
			*((unsigned int *)(base+items[i].offset))=PROXY;
		else if(!strcmp(value,"sctp-oo"))
			*((unsigned int *)(base+items[i].offset))=SCTP_OO;
		else if(!strcmp(value,"sctp-om"))
			*((unsigned int *)(base+items[i].offset))=SCTP_OM;
		else if(!strcmp(value,"udp"))
			*((unsigned int *)(base+items[i].offset))=UDP;
		else if(!strcmp(value,"udplite"))
			*((unsigned int *)(base+items[i].offset))=UDPLITE;
		else if(!strcmp(value,"tcp-rtp"))
			*((unsigned int *)(base+items[i].offset))=TCP_RTP;
		else if(!strcmp(value,"sctp-oo-rtp"))
			*((unsigned int *)(base+items[i].offset))=SCTP_OO_RTP;
		else if(!strcmp(value,"sctp-om-rtp"))
			*((unsigned int *)(base+items[i].offset))=SCTP_OM_RTP;
		else if(!strcmp(value,"udp-rtp"))
			*((unsigned int *)(base+items[i].offset))=UDP_RTP;
		else if(!strcmp(value,"udplite-rtp"))
			*((unsigned int *)(base+items[i].offset))=UDPLITE_RTP;
		else if(!strcmp(value,"helper"))
			*((unsigned int *)(base+items[i].offset))=HELPER;
		else if(!strcmp(value,"packet-helper"))
			*((unsigned int *)(base+items[i].offset))=PKTHELPER;
		else goto err1;
		break;

	case LISTEN:
		if(!(addr=strtok_r(NULL," \t\r\n",&mem))||!*addr)goto err1;
		if(!(ptr=strtok_r(NULL," \t\r\n",&mem))||!*ptr)goto err1;
		xtra=strtok_r(NULL," \t\r\n",&mem);
		if(!strcmp(value,"tcp"))proto=TCP;
		else if(!strcmp(value,"sctp-oo"))proto=SCTP_OO;
		else if(!strcmp(value,"sctp-om"))proto=SCTP_OM;
		else if(!strcmp(value,"udp"))proto=UDP;
		else if(!strcmp(value,"udplite"))proto=UDPLITE;
		else if(!strcmp(value,"tcp-rtp"))proto=TCP_RTP;
		else if(!strcmp(value,"sctp-oo-rtp"))proto=SCTP_OO_RTP;
		else if(!strcmp(value,"sctp-om-rtp"))proto=SCTP_OM_RTP;
		else if(!strcmp(value,"udp-rtp"))proto=UDP_RTP;
		else if(!strcmp(value,"udplite-rtp"))proto=UDPLITE_RTP;
		else if(!strcmp(value,"unix"))proto=UNIX;
		else if(!strcmp(value,"packet-unix"))proto=PKTUNIX;
		else goto err1;
		small=strip=0;
		if(xtra)
		{
			if(!strcmp(xtra,"small"))small=1;
			else if(!strcmp(xtra,"strip"))strip=1;
			else if(!strcmp(xtra,"small+strip"))small=strip=1;
			else goto err1;
			if(strip)switch(proto)
			{
			case TCP:
			case SCTP_OO:
			case SCTP_OM:
			case TCP_RTP:
			case SCTP_OO_RTP:
			case SCTP_OM_RTP:
			case UNIX:
				break;
			default:goto err1;
			}
		}
		if(proto!=UNIX&&proto!=PKTUNIX)
		{
			if((number=strtoul(ptr,&mem,10))==ULONG_MAX||*mem)
				goto err1;
			if(number<1||number>65535)goto err1;
		}
		else if(strcmp(addr,"*"))goto err1;
		if(!(l=malloc(sizeof(struct listen))))goto err2;
		if(proto!=UNIX&&proto!=PKTUNIX)
		{
			l->port=(unsigned int)number;
			l->fn=NULL;
		}
		else
		{
			l->port=0;
			if(!(l->fn=strdup(ptr)))goto err3;
		}
		if(!strcmp(addr,"*"))l->addr=NULL;
		else if(!(l->addr=strdup(addr)))
		{
			if(l->fn)free(l->fn);
			goto err3;
		}
		l->proto=(unsigned int)(proto&0x17f);
		l->rtp=((proto&0x80)?1:0);
		l->small=small;
		l->strip=strip;
		l->next=*((struct listen **)(base+items[i].offset));
		*((struct listen **)(base+items[i].offset))=l;
		break;

	default:goto err1;
	}

	return 0;

err3:	free(l);
err2:	fprintf(stderr,"internal error\n");
	return -1;

err1:	fprintf(stderr,"configuration syntax error line %d\n",num);
	return -1;
}

COLD static int verify(struct config *c)
{
	if(c->mode==CLIENT)
	{
		if(c->proto!=HELPER&&c->proto!=PKTHELPER)
		{
			if(!c->server||!c->port||c->proto==-1)return -1;
			if(!c->dnstimeout||!c->conntimeout)return -1;
		}
		else if(!c->helpercmd||!c->conntimeout)return -1;
		if(c->minbits>c->maxbits||c->sigminbits>c->sigmaxbits)return -1;
		if(!c->x25519&&!c->secp&&!c->brainpool)return -1;
		if(!c->sigsecp&&!c->sigbrainpool)return -1;
		if(c->setuplimit<512||!c->delay)return -1;
		if(!c->keyfile)return -1;
		if(c->small&&(!c->rekey||c->rekey>3600))return -1;
		switch(c->proto)
		{
		case SOCKS5:
			if(!c->sockshost||c->socksport<1||c->socksport>65535)
				return -1;
			c->proto=TCP;
			c->proxyhost=NULL;
			c->proxyport=0;
			break;
		case PROXY:
			if(!c->proxyhost||c->proxyport<1||c->proxyport>65535)
				return -1;
			c->proto=TCP;
			c->sockshost=NULL;
			c->socksport=0;
			break;
		default:c->proxyhost=NULL;
			c->proxyport=0;
			c->sockshost=NULL;
			c->socksport=0;
			break;
		}
	}
	else
	{
		if(c->server||c->port||c->proto!=-1||c->rtp)return -1;
		if(c->dnstimeout||c->conntimeout)return -1;
		if(c->minbits||c->maxbits||c->sigminbits||c->sigmaxbits)
			return -1;
		if(c->x25519||c->secp||c->brainpool)return -1;
		if(c->sigsecp||c->sigbrainpool)return -1;
		if(c->setuplimit||c->delay)return -1;
		if(c->keyfile||c->sendbuf||c->maxseg)return -1;
		if(c->tcpdefer||c->frag||c->small)return -1;
		if(c->paranoia||c->noop)return -1;
		if(c->sockshost||c->socksport)return -1;
		if(c->proxyhost||c->proxyport)return -1;
		if(c->helpercmd||c->helpercnf)return -1;
		if(c->authfile[1]||c->strip||c->srcaddr)return -1;
	}
	if(!c->script||!c->authfile[0])return -1;
	return 0;
}

COLD int parseconfig(char *fn,struct global *g,void *gctx,int cefd,int sefd)
{
	int mode=0;
	int num=0;
	int small=0;
	char *ptr;
	FILE *fp;
	struct config *cfg;
	struct listen *l;
	char line[8192];

	memset(g,0,sizeof(struct global));

	if(!(fp=fopen(fn,"re")))
	{
		fprintf(stderr,"can't open %s\n",fn);
		goto err1;
	}

	while(fgets(line,sizeof(line),fp))
	{
		num++;
		for(ptr=line;*ptr;ptr++)if(*ptr!=' '||*ptr!='\t')break;
		if(!*ptr||*ptr=='#'||*ptr=='\r'||*ptr=='\n')continue;
		if(*ptr=='[')
		{
			if(!strncmp(ptr,"[global]",8))mode=1;
			else if(!strncmp(ptr,"[client]",8))mode=2;
			else if(!strncmp(ptr,"[server]",8))mode=3;
			else goto err3;
			if(mode>1)
			{
				if(!(cfg=malloc(sizeof(struct config))))
				{
					fprintf(stderr,"internal error\n");
					goto err2;
				}
				memset(cfg,0,sizeof(struct config));
				cfg->global=g;
				cfg->proto=-1;
				cfg->next=g->cfg;
				g->cfg=cfg;
			}
			switch(mode)
			{
			case 2:	cfg->mode=CLIENT;
				break;
			case 3:	cfg->mode=SERVER;
				break;
			}
		}
		else switch(mode)
		{
		case 0:	goto err3;
		case 1:	if(parseline(ptr,num,globcfg,g))goto err2;
			break;
		default:if(parseline(ptr,num,srvcln,cfg))goto err2;
			break;
		}
	}

	fclose(fp);

	if(!g->setuptimeout||!g->cfg)goto err4;
	for(cfg=g->cfg;cfg;cfg=cfg->next)
	{
		if(cfg->connection)
		{
			if(cfg->mode==SERVER)g->isserver=1;
			else g->isclient=1;
		}
		if(cfg->proto!=-1)
		{
			if(cfg->strip)switch(cfg->proto)
			{
			case TCP:
			case SCTP_OO:
			case SCTP_OM:
			case TCP_RTP:
			case SCTP_OO_RTP:
			case SCTP_OM_RTP:
			case SOCKS5:
			case PROXY:
			case HELPER:
				if(cfg->mode==CLIENT)break;
			default:goto err4;
			}
			cfg->rtp=((cfg->proto&0x80)?1:0);
			cfg->proto&=0x17f;
		}
		else cfg->rtp=0;
		if(!cfg->rohccid)cfg->rohccid=256;
		if(verify(cfg))goto err4;
	}
	if(g->minbits>g->maxbits||g->sigminbits>g->sigmaxbits)goto err4;
	if(g->portmin||g->portmax)if(!g->portmin||!g->portmax||
		g->portmin>g->portmax||g->portmax-g->portmin<16)goto err4;
	if(g->isserver)
	{
		if(!g->x25519&&!g->secp&&!g->brainpool)goto err4;
		if(!g->sigsecp&&!g->sigbrainpool)goto err4;
		if(g->setuplimit<512||!g->sockets||!g->keyfile)goto err4;
	}

	for(l=g->sockets;l;l=l->next)switch(l->proto)
	{
	case UDP:
	case UDPLITE:
		if(!g->keepalive)goto err4;
	default:if(l->small)small=1;
		break;
	}

	if(small)for(cfg=g->cfg;cfg;cfg=cfg->next)
		if(cfg->mode==SERVER)if(!cfg->rekey||cfg->rekey>3600)
	{
err4:		fprintf(stderr,"configuration syntax error\n");
		return -1;
	}

	g->glob=gctx;
	g->cefd=cefd;
	g->sefd=sefd;

	return 0;

err3:	fprintf(stderr,"configuration syntax error line %d\n",num);
err2:	fclose(fp);
err1:	return -1;
}

COLD void freeconfig(struct global *g)
{
	struct listen *l;
	struct config *c;

	while(g->sockets)
	{
		l=g->sockets;
		g->sockets=l->next;
		if(l->fn)free(l->fn);
		if(l->addr)free(l->addr);
		free(l);
	}

	while(g->cfg)
	{
		c=g->cfg;
		g->cfg=c->next;
		if(c->keyfile)free(c->keyfile);
		if(c->script)free(c->script);
		if(c->authfile[0])free(c->authfile[0]);
		if(c->authfile[1])free(c->authfile[1]);
		if(c->server)free(c->server);
		free(c);
	}

	if(g->keyfile)free(g->keyfile);
	if(g->user)free(g->user);
}
