/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "misc.h"
#include "crypto.h"
#include "rtp.h"

struct rtp
{
	int offset;
	unsigned int ssrc;
	unsigned int hdr;
	unsigned int oecnt;
	unsigned int pusicnt;
	unsigned short seq;
	unsigned char marker;
	unsigned char cc;
};

#pragma pack(push,1)

union rtphdr
{
	unsigned char hdr1[12];
	unsigned short hdr2[6];
	unsigned int hdr4[3];
};

#pragma pack(pop)

HOT static inline unsigned int khz90(void)
{
	unsigned long tme;
	struct timespec ts;

	if(UNLIKELY(clock_gettime(CLOCK_MONOTONIC_RAW,&ts)))return 0;
	tme=((unsigned long)ts.tv_sec)*10000UL+
		(unsigned long)(ts.tv_nsec/100000);
	return (unsigned int)(tme*9UL)+(unsigned int)(ts.tv_nsec%100000/11112);
}

HOT int rtp_pack(void *ctx,void *data,int len,void **out,int *olen)
{
	struct rtp *r=ctx;
	unsigned char e=0;
	unsigned int pusi=0;
	union
	{
		unsigned char *ptr;
		unsigned int *ival;
		union rtphdr *hdr;
	} u;

	*olen=0;

	if(UNLIKELY(!len))*olen=12+r->offset;
	else switch(len)
	{
	case 1288:
		*olen+=188;
		e++;
	case 1104:
		*olen+=188;
		e++;
	case 920:
		*olen+=188;
		e++;
	case 736:
		*olen+=188;
		e++;
	case 552:
		*olen+=188;
		e++;
	case 368:
		*olen+=188;
		e++;
	case 184:
		*olen+=200+r->offset;
		e++;
	}

	if(UNLIKELY(!*olen)||UNLIKELY(!(u.ptr=*out=malloc(*olen))))return -1;
	if(r->offset)
	{
		*u.ptr++=((*olen-r->offset)>>8)&0xff;
		*u.ptr++=(*olen-r->offset)&0xff;
	}

	u.hdr->hdr1[0]=0x80;
	u.hdr->hdr1[1]=33|r->marker;
	u.hdr->hdr2[1]=htons(r->seq++);
	u.hdr->hdr4[1]=htonl(khz90());
	u.hdr->hdr4[2]=r->ssrc;

	if(UNLIKELY((r->oecnt+=e)>=20000))
	{
		r->oecnt-=20000;
		r->hdr^=0x40;
	}

	if(UNLIKELY((r->pusicnt+=e)>=1000))
	{
		r->pusicnt-=1000;
		pusi=0x400000;
	}

	r->marker=0;
	r->cc+=e;

	switch(len)
	{
	case 1288:
		u.ival[285]=htonl(r->hdr|((r->cc+--e)&0xf));
		memcpy(u.ptr+1144,data+1104,184);
	case 1104:
		u.ival[238]=htonl(r->hdr|((r->cc+--e)&0xf));
		memcpy(u.ptr+956,data+920,184);
	case 920:
		u.ival[191]=htonl(r->hdr|((r->cc+--e)&0xf));
		memcpy(u.ptr+768,data+736,184);
	case 736:
		u.ival[144]=htonl(r->hdr|((r->cc+--e)&0xf));
		memcpy(u.ptr+580,data+552,184);
	case 552:
		u.ival[97]=htonl(r->hdr|((r->cc+--e)&0xf));
		memcpy(u.ptr+392,data+368,184);
	case 368:
		u.ival[50]=htonl(r->hdr|((r->cc+--e)&0xf));
		memcpy(u.ptr+204,data+184,184);
	case 184:
		u.ival[3]=htonl(r->hdr|((r->cc+--e)&0xf)|pusi);
		memcpy(u.ptr+16,data,184);
	}

	r->cc+=e;

	return 0;
}

HOT int rtp_unpack(void *ctx,void *data,int len,void **out,int *olen)
{
	if(UNLIKELY(len<12))goto err1;
	if(UNLIKELY(((unsigned char *)data)[0]!=0x80))goto err1;
	if(UNLIKELY((((unsigned char *)data)[1]&0x7f)!=33))goto err1;

	if(UNLIKELY(len==12))return -2;

	*olen=0;

	switch(len)
	{
	case 1328:
		*olen+=184;
	case 1140:
		*olen+=184;
	case 952:
		*olen+=184;
	case 764:
		*olen+=184;
	case 576:
		*olen+=184;
	case 388:
		*olen+=184;
	case 200:
		*olen+=184;
	}

	if(UNLIKELY(!*olen)||UNLIKELY(!(*out=malloc(*olen))))goto err1;

	switch(len)
	{
	case 1328:
		if(UNLIKELY(((unsigned char *)data)[1140]!=0x47))goto err2;
		memcpy(*out+1104,data+1144,184);
	case 1140:
		if(UNLIKELY(((unsigned char *)data)[952]!=0x47))goto err2;
		memcpy(*out+920,data+956,184);
	case 952:
		if(UNLIKELY(((unsigned char *)data)[764]!=0x47))goto err2;
		memcpy(*out+736,data+768,184);
	case 764:
		if(UNLIKELY(((unsigned char *)data)[576]!=0x47))goto err2;
		memcpy(*out+552,data+580,184);
	case 576:
		if(UNLIKELY(((unsigned char *)data)[388]!=0x47))goto err2;
		memcpy(*out+368,data+392,184);
	case 388:
		if(UNLIKELY(((unsigned char *)data)[200]!=0x47))goto err2;
		memcpy(*out+184,data+204,184);
	case 200:
		if(UNLIKELY(((unsigned char *)data)[12]!=0x47))goto err2;
		memcpy(*out,data+16,184);
	}

	return 0;

err2:	free(*out);
err1:	return -1;
}

void *rtp_init(void *h,int packet)
{
	struct rtp *r;

	if(!(r=malloc(sizeof(struct rtp))))goto err1;
	memset(r,0,sizeof(struct rtp));
	if(crypto_random(h,&r->ssrc,sizeof(r->ssrc)))goto err2;
	if(crypto_random(h,&r->hdr,sizeof(r->hdr)))goto err2;
	r->hdr&=0x000fff00;
	r->hdr|=0x47010090;
	r->offset=(packet?0:2);
	r->marker=0x80;
	return r;

err2:	free(r);
err1:	return NULL;
}

void rtp_exit(void *ctx)
{
	free(ctx);
}
