/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <pthread.h>
#include <sys/timerfd.h>
#include <sys/mman.h>
#include <poll.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#ifndef NO_ROHC
#include <arpa/inet.h>
#include <rohc/rohc_comp.h>
#include <rohc/rohc_decomp.h>
#endif
#ifdef DEBUG
#include <sys/prctl.h>
#endif
#ifdef ROHCDEBUG
#include <stdarg.h>
#endif
#if defined(DEBUG) || defined(ROHCDEBUG)
#include <stdio.h>
#endif
#include "misc.h"
#include "crypto.h"
#include "proto.h"
#include "parse.h"
#include "tcp.h"
#include "sctp_oo.h"
#include "sctp_om.h"
#include "udp.h"
#include "rtp.h"
#include "unix.h"
#include "helper.h"
#include "tun.h"
#include "keyio.h"
#include "rxtx.h"

#ifdef DEBUG
#define ERR(a) do							    \
{									    \
	fprintf(stderr,"Connection abort: %s line %d\n",__FILE__,__LINE__); \
	goto a;								    \
} while(0)
#define ERROR \
	fprintf(stderr,"Connection abort: %s line %d\n",__FILE__,__LINE__);
#define NOTIFY(a) fprintf(stderr,"Notice: %s\n",a);
#else
#define ERR(a) goto a
#define ERROR
#define NOTIFY(a)
#endif

#pragma pack(push,1)

struct rtp
{
	uint8_t verpxcc;	/* rtp version/p/x/csrc count   */
	uint8_t mpt;		/* rtp marker bit/payload type  */
	uint16_t seq;		/* rtp sequence number          */
	uint32_t timestamp;	/* rtp timestamp                */
	uint32_t ssrc;		/* rtp synchronization source   */
	uint32_t csrc[15];	/* rtp contributing source      */
};

#pragma pack(pop)

struct localdata
{
	int rx;
	int tx;
	int rxbits;
	int txbits;
	int state;
	int keycnt;
	int noopcnt;
	int nettot;
	int netpos;
	int tuntot;
	int tunpos;
	int minhdr;
	int minread;
	int limit;
	int ping;
	int pong;
	int failcnt;
	struct itimerspec it;
	void *tunmsg;
	void *netmsg;
	void *ctrl;
	void *ack;
#ifndef NO_ROHC
	void *fbk;
#endif
	void *noop;
	void *newpub;
	void *revoke;
	void *pktkey;
	void *pktpub;
	struct crypto_rekeydata *rkdta;
	unsigned char saltmem[64];
};

#ifndef NO_ROHC

struct rohcdata
{
	int x;
	int fbklen;
	int state;
	struct rohc_comp *comp;
	struct rohc_decomp *decomp;
	struct rohc_buf rohctunin;
	struct rohc_buf rohctunout;
	struct rohc_buf rohcnetin;
	struct rohc_buf rohcnetout;
	struct rohc_buf rohcfbkin;
	struct rohc_buf rohcfbkout;
	unsigned char rohcnetbuf[65535];
	unsigned char rohctunbuf[65535-32];
};

static int rohcrand(const struct rohc_comp *const comp,void *h)
{
	int num;

	crypto_random(h,&num,sizeof(num));
	return num;
}

static void rohctrace(void *const priv_ctxt,const rohc_trace_level_t level,
	const rohc_trace_entity_t entity,const int profile,
	const char *const format,...)
{
#ifdef ROHCDEBUG
	char *lvl;
	char *ent;
	char *prf;
	va_list ap;
	char bfr[8192];

	switch(level)
	{
	case ROHC_TRACE_DEBUG:
		lvl="DEBUG";
		break;
	case ROHC_TRACE_INFO:
		lvl="INFO";
		break;
	case ROHC_TRACE_WARNING:
		lvl="WARN";
		break;
	case ROHC_TRACE_ERROR:
		lvl="ERROR";
		break;
	case ROHC_TRACE_LEVEL_MAX:
		lvl="MAX";
		break;
	default:lvl="UNKNOWN";
		break;
	}

	switch(entity)
	{
	case ROHC_TRACE_COMP:
		ent="COMP";
		break;
	case ROHC_TRACE_DECOMP:
		ent="DECOMP";
		break;
	case ROHC_TRACE_ENTITY_MAX:
		ent="MAX";
		break;
	default:ent="UNKNOWN";
		break;
	}

	switch(profile)
	{
	case ROHC_PROFILE_UNCOMPRESSED:
		prf="UNCOMPRESSED";
		break;
	case ROHC_PROFILE_RTP:
		prf="RTP";
		break;
	case ROHC_PROFILE_UDP:
		prf="UDP";
		break;
	case ROHC_PROFILE_ESP:
		prf="ESP";
		break;
	case ROHC_PROFILE_IP:
		prf="IP";
		break;
	case ROHC_PROFILE_RTP_LLA:
		prf="RTP_LLA";
		break;
	case ROHC_PROFILE_TCP:
		prf="TCP";
		break;
	case ROHC_PROFILE_UDPLITE_RTP:
		prf="UDPLITE_RTP";
		break;
	case ROHC_PROFILE_UDPLITE:
		prf="UDPLITE";
		break;
	case ROHC_PROFILE_MAX:
		prf="MAX";
		break;
	case ROHC_PROFILE_GENERAL:
		prf="GENERAL";
		break;
	default:prf="UNKNOWN";
		break;
	}

	va_start(ap,format);
	vsnprintf(bfr,sizeof(bfr),format,ap);
	va_end(ap);

	fprintf(stderr,"ROHC: %s-%s-%s: %s\n",lvl,ent,prf,bfr);
#endif
}

static bool rohcrtp(const unsigned char *const ip,
	const unsigned char *const udp,const unsigned char *const payload,
	const unsigned int payload_size,void *const rtp_private)
{
	struct rtp *rtp=(struct rtp *)payload;
	uint16_t port;

	if(payload_size<sizeof(struct rtp)-sizeof(rtp->csrc)+2)return false;
	if(((port=ntohs(((uint16_t *)udp)[0]))&1)||port<1024||port==1080||
		port==1194||port==1512||port==1812||port==4500)return false;
	if(((port=ntohs(((uint16_t *)udp)[1]))&1)||port<1024||port==1080||
		port==1194||port==1512||port==1812||port==4500)return false;
	if((rtp->verpxcc&0xc0)!=0x80)return false;
	if(payload_size<sizeof(struct rtp)-sizeof(rtp->csrc)+
		((rtp->verpxcc&0x0f)<<2)+2)return false;
	return true;
}

static int rohc_init(void *h,struct rohcdata *r,int cid,void *tunbuf,int tbsize)
{
	int udplite=ROHC_PROFILE_UDPLITE;

	if(!strncmp(rohc_version(),"2.3.",4))udplite=-1;
	memset(&r->rohctunin,0,sizeof(r->rohctunin));
	r->rohctunin.data=tunbuf;
	r->rohctunin.max_len=tbsize;
	memset(&r->rohctunout,0,sizeof(r->rohctunout));
	r->rohctunout.data=r->rohctunbuf;
	r->rohctunout.max_len=sizeof(r->rohctunbuf);
	memset(&r->rohcnetin,0,sizeof(r->rohcnetin));
	memset(&r->rohcnetout,0,sizeof(r->rohcnetout));
	r->rohcnetout.data=r->rohcnetbuf;
	r->rohcnetout.max_len=sizeof(r->rohcnetbuf);
	memset(&r->rohcfbkin,0,sizeof(r->rohcfbkin));
	memset(&r->rohcfbkout,0,sizeof(r->rohcfbkout));
	if(!(r->comp=rohc_comp_new2(ROHC_LARGE_CID,cid,rohcrand,h)))goto err1;
	if(!(r->decomp=rohc_decomp_new2(ROHC_LARGE_CID,cid,ROHC_O_MODE)))
		goto err2;
	if(!rohc_comp_set_traces_cb2(r->comp,rohctrace,NULL))goto err3;
	if(!rohc_decomp_set_traces_cb2(r->decomp,rohctrace,NULL))goto err3;
	if(!rohc_comp_enable_profiles(r->comp,ROHC_PROFILE_UNCOMPRESSED,
		ROHC_PROFILE_UDP,ROHC_PROFILE_IP,ROHC_PROFILE_RTP,
		ROHC_PROFILE_ESP,ROHC_PROFILE_TCP,udplite,-1))goto err3;
	if(!rohc_decomp_enable_profiles(r->decomp,ROHC_PROFILE_UNCOMPRESSED,
		ROHC_PROFILE_UDP,ROHC_PROFILE_IP,ROHC_PROFILE_RTP,
		ROHC_PROFILE_ESP,ROHC_PROFILE_TCP,udplite,-1))goto err3;
	if(!rohc_comp_set_rtp_detection_cb(r->comp,rohcrtp,NULL))goto err3;
	if(!rohc_comp_set_features(r->comp,ROHC_COMP_FEATURE_NONE))goto err3;
	if(!rohc_decomp_set_features(r->decomp,ROHC_DECOMP_FEATURE_NONE))
		goto err3;
	if(!rohc_comp_set_mrru(r->comp,65535-32))goto err3;
	if(!rohc_decomp_set_mrru(r->decomp,65535-32))goto err3;
	r->state=1;
	return 0;

err3:	rohc_decomp_free(r->decomp);
	r->decomp=NULL;
err2:	rohc_comp_free(r->comp);
	r->comp=NULL;
err1:	return -1;
}

static void rohc_exit(struct rohcdata *r)
{
	if(r->decomp)rohc_decomp_free(r->decomp);
	r->decomp=NULL;
	if(r->comp)rohc_comp_free(r->comp);
	r->comp=NULL;
	r->state=0;
}

#endif

static void doexec(void *script,void *mode,void *dev)
{
	if(script)
	{
		sctpomlock();
		switch(fork())
		{
		case -1:sctpomunlock();
			return;
		case 0:	munlockall();
			execl(script,script,mode,dev,NULL);
			exit(255);
		default:sctpomunlock();
			break;
		}
	}
}

HOT static int rxtx_recv(struct conn *c,void *bfr,int max)
{
	switch(c->proto)
	{
	case TCP:
		return tcprecv(c->fd,bfr,max);
	case SCTP_OO:
		return sctpoorecv(c->fd,bfr,max);
	case SCTP_OM:
		return sctpomrecv(c->fd,bfr,max);
	case UDP:
	case UDPLITE:
		return udprecv(c->fd,bfr,max);
	case HELPER:
		return helperrecv(c->fd,bfr,max);
	case PKTHELPER:
		return helperpktrecv(c->fd,bfr,max);
	case UNIX:
		return unixrecv(c->fd,bfr,max);
	case PKTUNIX:
		return unixpktrecv(c->fd,bfr,max);
	default:return -1;
	}
}

HOT static int rxtx_send(struct conn *c,void *bfr,int len)
{
	switch(c->proto)
	{
	case TCP:
		return tcpsend(c->fd,bfr,len);
	case SCTP_OO:
		return sctpoosend(c->fd,bfr,len);
	case SCTP_OM:
		return sctpomsend(c->fd,c->dst,c->dstlen,bfr,len);
	case UDP:
	case UDPLITE:
		return udpsend(c->fd,bfr,len);
	case HELPER:
		return helpersend(c->fd,bfr,len);
	case PKTHELPER:
		return helperpktsend(c->fd,bfr,len);
	case UNIX:
		return unixsend(c->fd,bfr,len);
	case PKTUNIX:
		return unixpktsend(c->fd,bfr,len);
	default:return -1;
	}
}

static void rxtx_clean(int tfd,void *h,struct conn *c,struct localdata *l)
{
	uint64_t dummy;

	memset(&l->it,0,sizeof(l->it));
	timerfd_settime(tfd,0,&l->it,NULL);
	while(read(tfd,&dummy,sizeof(dummy))==sizeof(dummy));
	if(l->tunmsg)free(l->tunmsg);
	l->tunmsg=NULL;
	if(!l->netmsg)
	{
		l->nettot=l->minread;
		l->netpos=0;
	}
	if(c->dst)free(c->dst);
	c->dst=NULL;
	if(c->rtp)rtp_exit(c->rtp);
	c->rtp=NULL;
	crypto_freepair(h,c->conn,3);
	crypto_rekeyfree(h,&l->rkdta,NULL);
	if(c->keys)crypto_keysfree(h,c->keys);
	c->keys=NULL;
	memset(l->saltmem,0,sizeof(l->saltmem));
	if(l->ctrl)free(l->ctrl);
	l->ctrl=NULL;
	if(l->ack)free(l->ack);
	l->ack=NULL;
	crypto_pktrekeyfree(h,&l->pktkey);
	crypto_pktrekeyfree(h,&l->pktpub);
#ifndef NO_ROHC
	if(l->fbk)free(l->fbk);
	l->fbk=NULL;
#endif
	if(l->noop)free(l->noop);
	l->noop=NULL;
	if(l->newpub)free(l->newpub);
	l->newpub=NULL;
	if(l->revoke)free(l->revoke);
	l->revoke=NULL;
	l->keycnt=0;
	l->noopcnt=0;
	l->state=0;
	l->rx=0;
	l->tx=0;
	l->rxbits=0;
	l->txbits=0;
	l->ping=0;
	l->pong=0;
	l->failcnt=0;
	if(c->fd!=-1)switch(c->proto)
	{
	case TCP:
		tcpclose(c->fd);
		break;
	case SCTP_OO:
		sctpooclose(c->fd);
		break;
	case SCTP_OM:
		sctpomclose(c->fd);
		break;
	case UDP:
	case UDPLITE:
		udpclose(c->fd);
		break;
	case HELPER:
	case PKTHELPER:
		helperclose(c->fd);
		break;
	case UNIX:
	case PKTUNIX:
		unix_close(c->fd);
		break;
	default:close(c->fd);
		break;
	}
	c->fd=-1;
}

static int handle_newpub(void *h,struct crypto_keys *keys,struct config *c,
	void *data,int len)
{
	int r=-1;
	int idx=0;
	int clen;
	int plen[2];
	void *curr;
	void *pub[2];

	if(c->mode!=CLIENT||!c->authfile[1])goto done;
	pub[0]=loadpub(h,c->authfile[0],&plen[0]);
	pub[1]=loadpub(h,c->authfile[1],&plen[1]);
	if(!(curr=crypto_getpub(h,keys,&clen)))goto out;
	if(pub[0]&&len==plen[0]&&!memcmp(data,pub[0],len))goto out;
	if(pub[1]&&len==plen[1]&&!memcmp(data,pub[1],len))goto out;
	if(pub[0]&&plen[0]==clen&&!memcmp(pub[0],curr,clen))idx|=1;
	if(pub[1]&&plen[1]==clen&&!memcmp(pub[1],curr,clen))idx|=2;
	switch(idx)
	{
	case 1:	if(savepub(h,c->authfile[1],data,len))break;
		r=0;
		NOTIFY("new auth stored");
		break;
	case 2:	if(savepub(h,c->authfile[0],data,len))break;
		r=0;
		NOTIFY("new auth stored");
		break;
	}
out:	if(curr)free(curr);
	if(pub[0])free(pub[0]);
	if(pub[1])free(pub[1]);
done:	return r;
}

static int handle_revoke(void *h,struct crypto_keys *keys,struct config *c,
	void *data,int len)
{
	int r=-1;
	int idx=0;
	int clen;
	int plen[2];
	void *curr;
	void *pub[2];

	if(c->mode!=CLIENT||!c->authfile[1])goto done;
	pub[0]=loadpub(h,c->authfile[0],&plen[0]);
	pub[1]=loadpub(h,c->authfile[1],&plen[1]);
	if(!(curr=crypto_getpub(h,keys,&clen)))goto out;
	if(pub[0]&&len==plen[0]&&!memcmp(data,pub[0],len))idx|=4;
	if(pub[1]&&len==plen[1]&&!memcmp(data,pub[1],len))idx|=8;
	if(pub[0]&&plen[0]==clen&&!memcmp(pub[0],curr,clen))idx|=1;
	if(pub[1]&&plen[1]==clen&&!memcmp(pub[1],curr,clen))idx|=2;
	switch(idx)
	{
	case 6:	if(unlink(c->authfile[0]))break;
		r=0;
		NOTIFY("revoked auth deleted");
		break;
	case 9:	if(unlink(c->authfile[1]))break;
		r=0;
		NOTIFY("revoked auth deleted");
		break;
	}
out:	if(curr)free(curr);
	if(pub[0])free(pub[0]);
	if(pub[1])free(pub[1]);
done:	return r;
}

static void *get_newpub(void *h,struct crypto_keys *keys,struct config *c,
	struct global *g,int *len)
{
	int plen;
	void *pub;
	void *r;

	if(c->mode!=SERVER||!g->newpub)return NULL;
	if(!(pub=loadpub(h,g->newpub,&plen)))return NULL;
	r=crypto_newrvkpub(h,keys,pub,plen,1,len);
	free(pub);
	return r;
}

static void *get_revoke(void *h,struct crypto_keys *keys,struct config *c,
	struct global *g,int *len)
{
	int plen;
	void *pub;
	void *r;

	if(c->mode!=SERVER||!g->revoke)return NULL;
	if(!(pub=loadpub(h,g->revoke,&plen)))return NULL;
	r=crypto_newrvkpub(h,keys,pub,plen,0,len);
	free(pub);
	return r;
}

HOT void *rxtx(void *data)
{
	struct localdata l;
	int tfd;
	int len;
	int type;
	int ctrllen;
	int acklen;
	int nooplen;
	int newlen;
	int revlen;
	int tmplen;
	int update;
	uint64_t dummy;
	void *h;
	void *tmp;
	struct rxtx *cnf=data;
	struct crypto_conn *ctmp;
	struct conn c;
	struct conn new;
	struct pollfd p[5];
	unsigned char netbuf[65535];
	unsigned char tunbuf[65535-32];
#ifndef NO_ROHC
	struct rohcdata r;
#endif
	sigset_t set;

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,NULL);
#ifdef DEBUG
	prctl(PR_SET_NAME,"rxtx",0,0,0);
#endif

	if(!(h=crypto_thread_init(cnf->g->glob)))goto err1;
	if((tfd=timerfd_create(CLOCK_MONOTONIC,TFD_CLOEXEC|TFD_NONBLOCK))==-1)
		goto err2;

	memset(&l,0,sizeof(l));
#ifndef NO_ROHC
	memset(&r,0,sizeof(r));
#endif
	c.fd=-1;
	l.nettot=l.minread=l.minhdr=crypto_hdrsize(h,NULL)-2;
	update=0;

	p[0].fd=(cnf->c->mode==CLIENT?cnf->g->cefd:cnf->g->sefd);
	p[0].events=POLLIN;
	p[1].fd=cnf->pfd[0];
	p[1].events=POLLIN;
	p[2].fd=tfd;
	p[2].events=POLLIN;
	p[3].fd=cnf->tun->fd;
	p[3].events=0;
	p[4].fd=-1;
	p[4].events=0;

	while(1)
	{
#ifndef NO_ROHC
		if(UNLIKELY(!r.rohcfbkout.data))
			if(LIKELY((r.rohcfbkout.data=malloc(256))!=NULL))
				r.rohcfbkout.max_len=256;
#endif

		if(!l.tunmsg)
		{
			if(UNLIKELY(l.ctrl!=NULL))
			{
				if(!c.packet&&l.tx)ERR(netfail);
				if(!(l.tunmsg=crypto_encode(h,c.conn[l.tx],
					CONTROL,l.ctrl,ctrllen,l.txbits,
					&l.tuntot,c.rtp!=NULL)))ERR(netfail);
				l.tunpos=0;
				free(l.ctrl);
				l.ctrl=NULL;
			}
			else if(UNLIKELY(l.ack!=NULL))
			{
				if(!c.packet)if(l.tx||c.conn[1]==c.conn[0])
					ERR(netfail);
				if(!(l.tunmsg=crypto_encode(h,c.conn[l.tx],
					CONTROL,l.ack,acklen,l.txbits,&l.tuntot,
					c.rtp!=NULL)))ERR(netfail);
				l.tunpos=0;
				free(l.ack);
				l.ack=NULL;
				if(!c.packet)
				{
					l.state|=1;
					if(l.state==3)
					{
						crypto_freepair(h,c.conn,1);
						c.conn[0]=c.conn[1];
						l.state=0;
						l.rx=0;
						l.tx=0;
						l.failcnt=0;
						NOTIFY("rekey complete");
					}
					else l.tx=1;
				}
			}
#ifndef NO_ROHC
			else if(UNLIKELY(l.fbk!=NULL))
			{
				l.tunmsg=crypto_encode(h,c.conn[l.tx],
					ROHCFEEDBK,l.fbk,r.fbklen,0,&l.tuntot,
					c.rtp!=NULL);
				l.tunpos=0;
				free(l.fbk);
				l.fbk=NULL;
			}
#endif
			if(UNLIKELY(l.revoke!=NULL))
			{
				if(!l.tunmsg)
				{
					l.tunmsg=crypto_encode(h,c.conn[l.tx],
						CONTROL,l.revoke,revlen,
						l.txbits,&l.tuntot,c.rtp!=NULL);
					l.tunpos=0;
					free(l.revoke);
					l.revoke=NULL;
				}
			}

			if(UNLIKELY(l.newpub!=NULL))
			{
				if(!l.tunmsg)
				{
					l.tunmsg=crypto_encode(h,c.conn[l.tx],
						CONTROL,l.newpub,newlen,
						l.txbits,&l.tuntot,c.rtp!=NULL);
					l.tunpos=0;
					free(l.newpub);
					l.newpub=NULL;
				}
			}

			if(UNLIKELY(l.noop!=NULL))
			{
				if(!l.tunmsg)
				{
					l.tunmsg=crypto_encode(h,c.conn[l.tx],
						CONTROL,l.noop,nooplen,l.txbits,
						&l.tuntot,c.rtp!=NULL);
					l.tunpos=0;
				}
				free(l.noop);
				l.noop=NULL;
			}

			if(UNLIKELY(c.rtp!=NULL)&&UNLIKELY(l.tunmsg!=NULL))
			{
				if(rtp_pack(c.rtp,l.tunmsg,l.tuntot,&tmp,
					&tmplen))ERR(netfail);
				free(l.tunmsg);
				l.tunmsg=tmp;
				l.tuntot=tmplen;
			}
		}
		p[3].events=0;
		p[4].events=0;
		if(l.tunmsg)p[4].events|=POLLOUT;
		else p[3].events|=POLLIN;
		if(l.netmsg)p[3].events|=POLLOUT;
		else p[4].events|=POLLIN;
		p[4].revents=0;
		if(UNLIKELY(c.fd==-1))
		{
			if(UNLIKELY(poll(p,4,-1)<=0))continue;
		}
		else if(UNLIKELY(poll(p,5,-1)<=0))continue;

		if(UNLIKELY(p[0].revents&POLLIN))break;

		if(UNLIKELY(p[1].revents&POLLIN))
			if(read(p[1].fd,&new,sizeof(new))==sizeof(new))
		{
			if(c.fd!=-1)
			{
			    if(c.packet)if(!crypto_pktfinmsg(h,c.keys,
				tunbuf,&len))
			    {
				if(l.tunmsg)free(l.tunmsg);
				if((l.tunmsg=crypto_encode(h,c.conn[l.tx],
					CONTROL,tunbuf,len,l.txbits,&l.tuntot,
					c.rtp!=NULL)))
				{
					if(c.rtp)
					{
					    if(!rtp_pack(c.rtp,l.tunmsg,
						l.tuntot,&tmp,&tmplen))
					    {
						rxtx_send(&c,tmp,tmplen);
						free(tmp);
					    }
					}
					else rxtx_send(&c,l.tunmsg,l.tuntot);
				}
			    }

			    doexec(cnf->c->script,"disconnected",cnf->tun->dev);
			}
			switch(new.proto)
			{
			case PKTHELPER:
			case PKTUNIX:
				new.keepalive=0;
			}
			l.limit=new.keepalive*3+2;
			l.minhdr=crypto_hdrsize(h,new.conn[0])-2;
			l.minread=(UNLIKELY(new.rtp!=NULL)?2:l.minhdr);
			rxtx_clean(tfd,h,&c,&l);
			c=new;
			p[4].fd=c.fd;
			p[4].revents=0;
			update=0;
			doexec(cnf->c->script,"connected",cnf->tun->dev);
#ifndef NO_ROHC
			/* rohc_comp_force_contexts_reinit fails... */
			rohc_exit(&r);
			if(r.rohcfbkout.data)free(r.rohcfbkout.data);
			if(rohc_init(h,&r,cnf->c->rohccid,tunbuf,
				sizeof(tunbuf)))ERR(netfail);
#endif
			if(c.noop)c.noop=60;
			if(c.packet)
			{
				l.it.it_value.tv_sec=l.it.it_interval.tv_sec=
					(c.rtp?0:1);
				l.it.it_value.tv_nsec=l.it.it_interval.tv_nsec=
					(c.rtp?250000000:0);
				if(c.rtp)
				{
					l.limit<<=2;
					c.rekey<<=2;
					c.noop<<=2;
				}
			}
			else
			{
				l.it.it_value.tv_sec=l.it.it_interval.tv_sec=
					(c.noop?1:c.rekey);
				l.it.it_value.tv_nsec=l.it.it_interval.tv_nsec=
					0;
			}

			l.newpub=get_newpub(h,c.keys,cnf->c,cnf->g,&newlen);
			l.revoke=get_revoke(h,c.keys,cnf->c,cnf->g,&revlen);

			timerfd_settime(tfd,0,&l.it,NULL);
		}

		if(UNLIKELY(p[2].revents&POLLIN))
			if(read(p[2].fd,&dummy,sizeof(dummy))==sizeof(dummy))
				if(c.fd!=-1)
		{
		    if(c.packet)
		    {
			if(c.keepalive)
			{
			    if(++l.ping>=(c.rtp?2:c.keepalive))
			    {
				if(c.rtp)
				{
					if(!rtp_pack(c.rtp,NULL,0,&tmp,&tmplen))
					{
						if(rxtx_send(&c,tmp,tmplen)==
							tmplen)l.ping=0;
						free(tmp);
					}
				}
				else if(!rxtx_send(&c,NULL,0))l.ping=0;
			    }
			    if(++l.pong>=l.limit)ERR(netfail);
			}
			if(c.rekey)if(++l.keycnt>=c.rekey)
			{
				if(cnf->c->rekeyfail)if(l.keycnt>=(c.rekey<<1))
					if(l.state)ERR(netfail);
				if(!(l.keycnt-c.rekey))
				{
					if(l.newpub)free(l.newpub);
					l.newpub=get_newpub(h,c.keys,cnf->c,
						cnf->g,&newlen);
					if(l.revoke)free(l.revoke);
					l.revoke=get_revoke(h,c.keys,cnf->c,
						cnf->g,&revlen);
					if(l.newpub||l.revoke)l.noopcnt=0;
				}
				if(!(l.state&1))
				{
					l.state|=1;
					crypto_pktrekeygen(h,&l.pktkey,c.curve);
				}
				if((l.state&3)==1&&!l.ctrl&&
					!((l.keycnt-c.rekey)&(c.rtp?0xf:0x3)))
				{
					crypto_pktrekeytxmsg(h,c.keys,l.pktkey,
						&l.ctrl,&ctrllen);
					l.noopcnt=0;
				}
			}
			if(c.noop)if(++l.noopcnt>=c.noop)
				if(!crypto_noopmsg(h,&l.noop,&nooplen))
					l.noopcnt=0;
		    }
		    else if(c.noop)
		    {
			if(++l.keycnt>=c.rekey)
			{
				l.keycnt=0;
				l.noopcnt=0;
				goto cmn;
			}
			if(c.noop)if(++l.noopcnt>=c.noop)
				if(!crypto_noopmsg(h,&l.noop,&nooplen))
					l.noopcnt=0;
		    }
		    else
		    {
cmn:			if(cnf->c->rekeyfail)if(l.failcnt++)ERR(netfail);
			if(l.newpub)free(l.newpub);
			l.newpub=get_newpub(h,c.keys,cnf->c,cnf->g,&newlen);
			if(l.revoke)free(l.revoke);
			l.revoke=get_revoke(h,c.keys,cnf->c,cnf->g,&revlen);
			if(l.ctrl)free(l.ctrl);
			l.ctrl=NULL;
			if(crypto_rekeysend(h,c.keys,&l.rkdta,c.curve,
				&l.ctrl,&ctrllen)==1)
			{
				crypto_freepair(h,c.conn,2);
				if(l.ack)free(l.ack);
				l.ack=NULL;
				if(!(c.conn[1]=
					crypto_rekeyderive(h,c.keys,&l.rkdta,
						c.conn[0],&l.ack,&acklen)))
							ERR(netfail);
				crypto_rekeyfree(h,&l.rkdta,l.saltmem);
			}
		    }
		}

		if(p[3].revents&POLLIN)
			if(LIKELY((len=read(p[3].fd,tunbuf,sizeof(tunbuf)))>0))
				if(LIKELY(c.fd!=-1))
		{
#ifndef NO_ROHC
			if(UNLIKELY(cnf->c->rohc)&&!c.packet&&LIKELY(r.state))
			{
				r.rohctunin.len=len;
				r.rohctunout.len=0;
				switch(rohc_compress4(r.comp,r.rohctunin,
					&r.rohctunout))
				{
				case ROHC_STATUS_OK:
					l.tunmsg=crypto_encode(h,c.conn[l.tx],
						ROHCDATA,r.rohctunbuf,
						r.rohctunout.len,l.txbits,
						&l.tuntot,c.rtp!=NULL);
					break;

				case ROHC_STATUS_SEGMENT:
					/* rohc_comp_force_contexts_reinit
					   fails... */
					l.tunmsg=NULL;
					l.tunpos=0;
					ERR(netfail);

				default:l.tunmsg=crypto_encode(h,c.conn[l.tx],
						DATAUNCOMP,tunbuf,len,l.txbits,
						&l.tuntot,c.rtp!=NULL);
					break;
				}
			}
			else
#endif
			l.tunmsg=crypto_encode(h,c.conn[l.tx],DATAUNCOMP,tunbuf,
				len,l.txbits,&l.tuntot,c.rtp!=NULL);
			l.tunpos=0;
			if(UNLIKELY(c.rtp!=NULL)&&LIKELY(l.tunmsg!=NULL))
			{
				if(UNLIKELY(rtp_pack(c.rtp,l.tunmsg,l.tuntot,
					&tmp,&tmplen)))
				{
					free(l.tunmsg);
					l.tunmsg=NULL;
				}
				else
				{
					free(l.tunmsg);
					l.tunmsg=tmp;
					l.tuntot=tmplen;
				}
			}
		}

		if(p[3].revents&POLLOUT)
			if(LIKELY((len=write(p[3].fd,l.netmsg+l.netpos,
				l.nettot-l.netpos))>0))
					if((l.netpos+=len)==l.nettot)
		{
#ifndef NO_ROHC
			if(l.netmsg!=r.rohcnetbuf)
#endif
				free(l.netmsg);
			l.netmsg=NULL;
			l.nettot=l.minread;
			l.netpos=0;
		}

		if(p[4].revents&POLLIN)switch(c.proto)
		{
		case SCTP_OM:
		case UDP:
		case UDPLITE:
		case PKTHELPER:
		case PKTUNIX:
			if(UNLIKELY((l.nettot=rxtx_recv(&c,netbuf,
				sizeof(netbuf)))<l.minread))
			{
				if(l.nettot!=-2)
				{
					if(!c.packet)ERR(netfail);
					switch(c.proto)
					{
					case PKTHELPER:
					case PKTUNIX:
						ERR(netfail);
					}
				}
				else if(c.packet)l.pong=0;
				l.nettot=l.minread;
				l.netpos=0;
				break;
			}
			if(UNLIKELY(c.rtp!=NULL))
			{
				if(!c.packet)
				{
					tmplen=(netbuf[0]<<8)|netbuf[1];
					if(tmplen!=l.nettot-2)ERR(netfail);
					tmplen=2;
				}
				else tmplen=0;
				switch(l.nettot-tmplen)
				{
				case 1328:
				case 1140:
				case 952:
				case 764:
				case 576:
				case 388:
				case 200:
				case 12:break;
				default:goto skip;
				}
				switch(rtp_unpack(c.rtp,netbuf+tmplen,
					l.nettot-tmplen,&tmp,&tmplen))
				{
				case -2:if(c.packet)l.pong=0;
				case -1:goto skip;
				}
				if(UNLIKELY(crypto_unmask(c.conn[l.rx],tmp,
					tmplen,&len,&l.rxbits))||
					UNLIKELY(len!=tmplen))
				{
					free(tmp);
					goto skip;
				}
				if(c.packet)l.rx=l.rxbits&1;
				if(UNLIKELY(!(l.netmsg=crypto_decode(h,
					c.conn[l.rx],&type,tmp,tmplen,
					&l.nettot,NULL))))
				{
					free(tmp);
					goto skip;
				}
				free(tmp);
				goto rtpcnt1;
			}
			if(UNLIKELY(crypto_unmask(c.conn[l.rx],netbuf,l.nettot,
				&len,&l.rxbits))||UNLIKELY(len!=l.nettot))
					goto skip;
			if(c.packet)l.rx=l.rxbits&1;
			if(UNLIKELY(!(l.netmsg=crypto_decode(h,c.conn[l.rx],
				&type,netbuf,l.nettot,&l.nettot,NULL))))
			{
skip:				if(!c.packet)ERR(netfail);
				l.nettot=l.minread;
				l.netpos=0;
				break;
			}
rtpcnt1:		if(UNLIKELY(l.state==0xf)&&c.packet&&
				(l.rxbits&2)==(l.txbits&2))
			{
				l.tx^=1;
				l.txbits=(l.tx<<1)|l.tx;
				l.state=0;
				l.keycnt=0;
				l.failcnt=0;
				NOTIFY("rekey complete");
			}
			goto common;

		default:if(UNLIKELY((len=rxtx_recv(&c,netbuf+l.netpos,
				l.nettot-l.netpos))<=0))ERR(netfail);
			if((l.netpos+=len)==l.nettot)
			{
				if(UNLIKELY(l.nettot==2))
				{
					switch(
					    (l.nettot=(netbuf[0]<<8)|netbuf[1]))
					{
					case 1328:
					case 1140:
					case 952:
					case 764:
					case 576:
					case 388:
					case 200:
						break;
					default:ERR(netfail);
					}
					l.netpos=0;
				}
				else if(l.nettot==l.minhdr)
				{
					if(UNLIKELY(crypto_unmask(c.conn[l.rx],
						netbuf,l.nettot,&len,
						&l.rxbits))||
						UNLIKELY(len<l.minhdr+2)||
						UNLIKELY(len>sizeof(netbuf)))
							ERR(netfail);
					else l.nettot=len;
				}
				else if(UNLIKELY(c.rtp!=NULL))
				{
					if(UNLIKELY(rtp_unpack(c.rtp,netbuf,
						l.nettot,&tmp,&tmplen)))
							ERR(netfail);
					if(UNLIKELY(crypto_unmask(c.conn[l.rx],
						tmp,tmplen,&len,&l.rxbits))||
						UNLIKELY(len<l.minhdr+2)||
						UNLIKELY(len>sizeof(netbuf)))
					{
						free(tmp);
						ERR(netfail);
					}
					if(UNLIKELY(!(l.netmsg=crypto_decode(h,
						c.conn[l.rx],&type,tmp,len,
						&l.nettot,NULL))))
					{
						free(tmp);
						ERR(netfail);
					}
					free(tmp);
					goto common;
				}
				else if(UNLIKELY(!(l.netmsg=crypto_decode(h,
					c.conn[l.rx],&type,netbuf,l.nettot,
					&l.nettot,NULL))))ERR(netfail);
				else
				{
common:				    if(c.packet)l.pong=0;
				    switch(type)
				    {
				    case DATAUNCOMP:
					l.netpos=0;
					break;
#ifndef NO_ROHC
				    case ROHCDATA:
					if(UNLIKELY(!r.state))
					{
						l.nettot=l.minread;
						l.netpos=0;
						free(l.netmsg);
						l.netmsg=NULL;
						break;
					}
					r.rohcnetin.data=l.netmsg;
					r.rohcnetin.max_len=r.rohcnetin.len=
						l.nettot;
					r.rohcnetout.len=0;
					r.rohcfbkout.len=0;
					switch(rohc_decompress3(r.decomp,
						r.rohcnetin,&r.rohcnetout,NULL,
						LIKELY(r.rohcfbkout.data!=NULL)?
						&r.rohcfbkout:NULL))
					{
					case ROHC_STATUS_OK:
						break;
					case ROHC_STATUS_NO_CONTEXT:
					    /* rohc_comp_force_contexts_reinit
					       fails... */
					default:ERR(rohcfail);
					}
					if(!l.netmsg)break;
					free(l.netmsg);
					if(UNLIKELY(r.rohcfbkout.len))
					{
						if(UNLIKELY(l.fbk!=NULL))
							free(l.fbk);
						l.fbk=r.rohcfbkout.data;
						r.fbklen=r.rohcfbkout.len;
						r.rohcfbkout.data=NULL;
					}
					if(UNLIKELY(!r.rohcnetout.len))
						l.netmsg=NULL;
					else l.netmsg=r.rohcnetbuf;
					l.nettot=r.rohcnetout.len;
					l.netpos=0;
					break;

				    case ROHCFEEDBK:
					if(UNLIKELY(!r.state))
					{
						l.nettot=l.minread;
						l.netpos=0;
						free(l.netmsg);
						l.netmsg=NULL;
						break;
					}
					r.rohcfbkin.data=l.netmsg;
					r.rohcfbkin.max_len=r.rohcfbkin.len=
						l.nettot;
					r.x=rohc_comp_deliver_feedback2(r.comp,
						r.rohcfbkin);
					free(l.netmsg);
					l.netmsg=NULL;
					l.nettot=l.minread;
					l.netpos=0;
					break;
#endif
				    case CONTROL:
					if(c.packet)switch(crypto_pkttype(h,
						c.keys,l.netmsg,l.nettot,
						l.pktkey))
					{
					case 0:	ERR(ctrlfail);

					case 1:	if(!(l.state&4))
						    if(!crypto_pktrekeyrxmsg(h,
							c.keys,&l.pktpub,
							l.netmsg,l.nettot))
								l.state|=4;
						if((l.state&4)&&!l.ack)
						    crypto_pktackmsg(h,c.keys,
							&l.ack,&acklen,
							l.pktpub);
						goto pktcmn;

					case 2:	if(l.state&1)l.state|=2;
pktcmn:						if(l.state==7)
						 if((ctmp=crypto_pktrekeyderive(
							h,c.keys,&l.pktkey,
							&l.pktpub,
							c.conn[l.tx])))
						{
						    crypto_freepair(h,c.conn,
								2-l.tx);
						    c.conn[l.tx^1]=ctmp;
						    l.txbits=l.tx|((l.tx^1)<<1);
						    l.state|=8;
						    if((l.rxbits&2)==
							(l.txbits&2))
						    {
							l.tx^=1;
							l.txbits=(l.tx<<1)|l.tx;
							l.state=0;
							l.keycnt=0;
							l.failcnt=0;
							NOTIFY(
							    "rekey complete");
						    }
						}

					case 3:	break;

					case 4:	if(!(tmp=crypto_chkgetpub(h,
							c.keys,l.netmsg,
							l.nettot,&tmplen)))
								break;
						NOTIFY("new pub");
						if(!handle_newpub(h,c.keys,
							cnf->c,tmp,tmplen))
								update=1;
						free(tmp);
						break;

					case 5:	if(!(tmp=crypto_chkgetpub(h,
							c.keys,l.netmsg,
							l.nettot,&tmplen)))
								break;
						NOTIFY("pub revoked");
						if(!handle_revoke(h,c.keys,
							cnf->c,tmp,tmplen))
								update=1;
						free(tmp);
						break;
					}
					else switch(crypto_rekeyrecv(h,c.keys,
						&l.rkdta,c.curve,l.netmsg,
						l.nettot,l.saltmem))
					{
					case 1:	crypto_freepair(h,c.conn,2);
						if(!(c.conn[1]=
						    crypto_rekeyderive(h,c.keys,
						    &l.rkdta,c.conn[0],&l.ack,
						    &acklen)))ERR(ctrlfail);
						crypto_rekeyfree(h,&l.rkdta,
							l.saltmem);
						break;

					case 2:	if(c.conn[1]==c.conn[0])
							ERR(netfail);
						memset(l.saltmem,0,
							sizeof(l.saltmem));
						l.state|=2;
						if(l.state==3)
						{
						    crypto_freepair(h,c.conn,1);
						    c.conn[0]=c.conn[1];
						    l.state=0;
						    l.rx=0;
						    l.tx=0;
						    l.failcnt=0;
						    NOTIFY("rekey complete");
						}
						else l.rx=1;

					case 3:	break;

					case 4:	if(!(tmp=crypto_chkgetpub(h,
							c.keys,l.netmsg,
							l.nettot,&tmplen)))
								break;
						NOTIFY("new pub");
						if(!handle_newpub(h,c.keys,
							cnf->c,tmp,tmplen))
								update=1;
						free(tmp);
						break;

					case 5:	if(!(tmp=crypto_chkgetpub(h,
							c.keys,l.netmsg,
							l.nettot,&tmplen)))
								break;
						NOTIFY("pub revoked");
						if(!handle_revoke(h,c.keys,
							cnf->c,tmp,tmplen))
								update=1;
						free(tmp);
						break;

					case -1:ERR(ctrlfail);
					}

				    default:
					crypto_memclear(h,l.netmsg,l.nettot);
#ifndef NO_ROHC
					if(l.netmsg!=r.rohcnetbuf)
#endif
						free(l.netmsg);
					l.netmsg=NULL;
					l.nettot=l.minread;
					l.netpos=0;
					break;
				    }
				}
			}
			break;
		}

		if(p[4].revents&POLLOUT)
		{
			if(UNLIKELY((len=rxtx_send(&c,l.tunmsg+l.tunpos,
				l.tuntot-l.tunpos))<=0))
			{
				ERROR;
				if(0)
				{
ctrlfail:				crypto_memclear(h,l.netmsg,l.nettot);
#ifndef NO_ROHC
rohcfail:				if(l.netmsg!=r.rohcnetbuf)
#endif
						free(l.netmsg);
					l.netmsg=NULL;
					l.nettot=l.minread;
					l.netpos=0;
				}
netfail:			rxtx_clean(tfd,h,&c,&l);
#ifndef NO_ROHC
				/* rohc_comp_force_contexts_reinit fails... */
				rohc_exit(&r);
#endif
				doexec(cnf->c->script,"disconnected",
					cnf->tun->dev);
				if(cnf->notify)cnf->notify(h,cnf->user,update);
				update=0;
			}
			else
			{
				l.ping=0;
				l.noopcnt=0;
				if((l.tunpos+=len)==l.tuntot)
				{
					free(l.tunmsg);
					l.tunmsg=NULL;
				}
			}
		}
	}

	if(l.netmsg)
#ifndef NO_ROHC
		if(l.netmsg!=r.rohcnetbuf)
#endif
			free(l.netmsg);

	if(c.fd!=-1)
	{
	    if(c.packet)if(!crypto_pktfinmsg(h,c.keys,tunbuf,&len))
	    {
		if(l.tunmsg)free(l.tunmsg);
		if((l.tunmsg=crypto_encode(h,c.conn[l.tx],CONTROL,tunbuf,
			len,l.txbits,&l.tuntot,c.rtp!=NULL)))
		{
			if(c.rtp)
			{
			    if(!rtp_pack(c.rtp,l.tunmsg,l.tuntot,&tmp,&tmplen))
			    {
				rxtx_send(&c,tmp,tmplen);
				free(tmp);
			    }
			}
			else rxtx_send(&c,l.tunmsg,l.tuntot);
		}
	    }

	    doexec(cnf->c->script,"disconnected",cnf->tun->dev);
	}

	rxtx_clean(tfd,h,&c,&l);
#ifndef NO_ROHC
	if(r.rohcfbkout.data)free(r.rohcfbkout.data);
	rohc_exit(&r);
#endif
	close(tfd);
err2:	crypto_thread_exit(h);
err1:	pthread_exit(NULL);
}
