/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#ifndef NO_SCTP
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <netinet/in.h>
#include <netinet/sctp.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <poll.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#ifndef SCTP_SOCKOPT_PEELOFF_FLAGS
#define SCTP_SOCKOPT_PEELOFF_FLAGS 122
typedef struct
{
	sctp_peeloff_arg_t p_arg;
	unsigned flags;
} sctp_peeloff_flags_arg_t;
#endif
#endif
#include "misc.h"
#include "netlink.h"
#include "sctp_om.h"

#ifndef NO_SCTP
/* start of kernel bugfix: they managed to change the struct size as of
   kernel 5.5 and expect userspace to report the exact size, so an
   older kernel with newer userspace headers or vice versa fails
   utterly with EINVAL when setting SCTP_EVENTS. The following
   is a guess the kernel version and struct size workaround */

struct sctp_event_subscribe_v54 {
        __u8 sctp_data_io_event;
        __u8 sctp_association_event;
        __u8 sctp_address_event;
        __u8 sctp_send_failure_event;
        __u8 sctp_peer_error_event;
        __u8 sctp_shutdown_event;
        __u8 sctp_partial_delivery_event;
        __u8 sctp_adaptation_layer_event;
        __u8 sctp_authentication_event;
        __u8 sctp_sender_dry_event;
        __u8 sctp_stream_reset_event;
        __u8 sctp_assoc_reset_event;
        __u8 sctp_stream_change_event;
};

struct sctp_event_subscribe_v55 {
        __u8 sctp_data_io_event;
        __u8 sctp_association_event;
        __u8 sctp_address_event;
        __u8 sctp_send_failure_event;
        __u8 sctp_peer_error_event;
        __u8 sctp_shutdown_event;
        __u8 sctp_partial_delivery_event;
        __u8 sctp_adaptation_layer_event;
        __u8 sctp_authentication_event;
        __u8 sctp_sender_dry_event;
        __u8 sctp_stream_reset_event;
        __u8 sctp_assoc_reset_event;
        __u8 sctp_stream_change_event;
        __u8 sctp_send_failure_event_event;
};

union sctp_event_subscribe_kernelfix
{
	struct sctp_event_subscribe_v54 ev54;
	struct sctp_event_subscribe_v55 ev55;
	struct sctp_event_subscribe ev;
};

/* end of workaround part one */

static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;
static int nolock=0;

/* start of workaround part two */

int geteventsize(void)
{
	int nh;
	int nl;
	char *ptr;
	char *major;
	char *minor;
	struct utsname kinfo;

	if(uname(&kinfo))goto fail;
	major=kinfo.release;
	if(!(ptr=strchr(kinfo.release,'.')))goto fail;
	*ptr++=0;
	minor=ptr;
	if(!(ptr=strchr(ptr,'.')))goto fail;
	*ptr=0;
	if(!*major||!*minor)goto fail;
	nh=atoi(major);
	nl=atoi(minor);
	if(nh>5||(nh==5&&nl>4))
	{
		if(sizeof(struct sctp_event_subscribe)>
			sizeof(struct sctp_event_subscribe_v55))
				return sizeof(struct sctp_event_subscribe);
		return sizeof(struct sctp_event_subscribe_v55);
	}
	else
	{
		if(sizeof(struct sctp_event_subscribe)<
			sizeof(struct sctp_event_subscribe_v54))
				return sizeof(struct sctp_event_subscribe);
		return sizeof(struct sctp_event_subscribe_v54);
	}

fail:	return sizeof(struct sctp_event_subscribe);
}

/* end of workaround part two */

#endif

COLD int sctpomlisten(void *addr,int port,int keepalive,int sendbuf,int maxseg)
{
#ifndef NO_SCTP
	int s;
	int x;
	int v1;
	int v2;
	int family;
	socklen_t l;
	union sctp_event_subscribe_kernelfix ev;
	struct sctp_rtoinfo rto;
	struct sctp_paddrparams ap;
	union
	{
		struct sockaddr_in a4;
		struct sockaddr_in6 a6;
	} a;

	if(port<1||port>65535)goto err1;
	memset(&a,0,sizeof(a));
	if(addr)
	{
		if(inet_pton(AF_INET,addr,&a.a4.sin_addr)==1)
		{
			family=AF_INET;
			a.a4.sin_family=AF_INET;
			a.a4.sin_port=htons(port);
		}
		else if(inet_pton(AF_INET6,addr,&a.a6.sin6_addr)==1)
		{
			family=AF_INET6;
			a.a6.sin6_family=AF_INET6;
			a.a6.sin6_port=htons(port);
		}
		else goto err1;
	}
	else
	{
		family=AF_INET6;
		a.a6.sin6_family=AF_INET6;
		a.a6.sin6_addr=in6addr_any;
		a.a6.sin6_port=htons(port);
	}
	if((s=socket(family,SOCK_SEQPACKET|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_SCTP))==-1)goto err1;
	x=1;
	if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x)))goto err2;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	if(family==AF_INET6)
	{
		x=(addr?1:0);
		if(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
			goto err2;
	}
	x=1;
	if(setsockopt(s,SOL_SCTP,SCTP_NODELAY,&x,sizeof(x)))goto err2;
	if(sendbuf)if(setsockopt(s,SOL_SOCKET,SO_SNDBUF,&sendbuf,
		sizeof(sendbuf)))goto err2;
	if(maxseg)if(setsockopt(s,IPPROTO_SCTP,SCTP_MAXSEG,&maxseg,
		sizeof(maxseg)))goto err2;
	l=sizeof(rto);
	memset(&rto,0,sizeof(rto));
	if(getsockopt(s,SOL_SCTP,SCTP_RTOINFO,&rto,&l))goto err2;
	rto.srto_initial=500;
	rto.srto_min=50;
	rto.srto_max=5000;
	if(setsockopt(s,SOL_SCTP,SCTP_RTOINFO,&rto,sizeof(rto)))goto err2;
	if(keepalive)
	{
		v1=SPP_HB_ENABLE;
		v2=SPP_HB_DISABLE;
	}
	else
	{
		v1=SPP_HB_DISABLE;
		v2=SPP_HB_ENABLE;
	}
	l=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&l))goto err2;
	ap.spp_hbinterval=keepalive*1000;
	ap.spp_pathmaxrxt=3;
	ap.spp_sackdelay=200;
	ap.spp_flags|=v1|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~(v2|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO);
	if(setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap)))
		goto err2;
	memset(&ev,0,sizeof(ev));
	ev.ev.sctp_association_event=1;
	if(setsockopt(s,SOL_SCTP,SCTP_EVENTS,&ev,geteventsize()))goto err2;
	if(bind(s,(struct sockaddr *)(&a),family==AF_INET?
		sizeof(struct sockaddr_in):sizeof(struct sockaddr_in6)))
			goto err2;
	if(listen(s,SOMAXCONN))goto err2;
	return s;

err2:	close(s);
err1:	return -1;
#else
	return -1;
#endif
}

int sctpomaccept(int fd,int keepalive)
{
#ifndef NO_SCTP
	int s;
	int x;
	int v1;
	int v2;
	int family;
	void *ptr;
	struct sctp_getaddrs *ga;
	socklen_t ll;
	struct linger l;
	struct sctp_paddrparams ap;
	struct msghdr msg;
	struct iovec v;
	union sctp_notification notify;
	union sctp_event_subscribe_kernelfix ev;
	struct sctp_prim pa;
	sctp_peeloff_arg_t peel;
	sctp_peeloff_flags_arg_t peel2;
	unsigned char bfr[8192];

	memset(&ev,0,sizeof(ev));
	ev.ev.sctp_association_event=1;
	memset(&msg,0,sizeof(msg));
	msg.msg_iov=&v;
	msg.msg_iovlen=1;
	v.iov_base=&notify;
	v.iov_len=sizeof(notify);
	if(recvmsg(fd,&msg,MSG_NOSIGNAL|MSG_WAITALL)!=
		sizeof(notify.sn_assoc_change))goto err1;
	if((msg.msg_flags&(MSG_NOTIFICATION|MSG_EOR))!=
		(MSG_NOTIFICATION|MSG_EOR))goto err1;
	peel.associd=notify.sn_assoc_change.sac_assoc_id;

	if(nolock)
	{
		ll=sizeof(peel2);
		peel2.flags=SOCK_NONBLOCK|SOCK_CLOEXEC;
		if((s=getsockopt(fd,SOL_SCTP,SCTP_SOCKOPT_PEELOFF_FLAGS,
			&peel2,&ll))==-1)goto err1;
	}
	else
	{
		ll=sizeof(peel);
		pthread_mutex_lock(&mtx);
		if((s=getsockopt(fd,SOL_SCTP,SCTP_SOCKOPT_PEELOFF,&peel,&ll))
			==-1)
		{
			pthread_mutex_unlock(&mtx);
			goto err1;
		}
		if(fcntl(s,F_SETFD,FD_CLOEXEC))
		{
			pthread_mutex_unlock(&mtx);
			goto err2;
		}
		pthread_mutex_unlock(&mtx);
		if(fcntl(s,F_SETFL,O_NONBLOCK))goto err2;
	}
	if(setsockopt(s,SOL_SCTP,SCTP_EVENTS,&ev,geteventsize()))goto err2;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_SCTP,SCTP_NODELAY,&x,sizeof(x)))goto err2;
	if(keepalive)
	{
		v1=SPP_HB_ENABLE;
		v2=SPP_HB_DISABLE;
	}
	else
	{
		v1=SPP_HB_DISABLE;
		v2=SPP_HB_ENABLE;
	}
	ll=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto err2;
	ap.spp_hbinterval=keepalive*1000;
	ap.spp_pathmaxrxt=3;
	ap.spp_sackdelay=200;
	ap.spp_flags|=v1|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~(v2|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO);
	if(setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap)))
		goto err2;
	l.l_onoff=0;
	l.l_linger=5;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))goto err2;

	ll=sizeof(pa);
	if(getsockopt(s,SOL_SCTP,SCTP_PRIMARY_ADDR,&pa,&ll))goto err2;
	family=pa.ssp_addr.ss_family;

	ll=sizeof(bfr);
	ga=(void *)bfr;
	memset(ga,0,sizeof(struct sctp_getaddrs));
	if(!getsockopt(s,SOL_SCTP,SCTP_GET_PEER_ADDRS,ga,&ll))
		for(x=0,ptr=bfr+sizeof(struct sctp_getaddrs);x<ga->addr_num;x++)
			switch(((struct sockaddr *)ptr)->sa_family)
	{
	case AF_INET6:
		if(family!=AF_INET6||
			!memcmp(&((struct sockaddr_in6 *)ptr)->sin6_addr,
			  &((struct sockaddr_in6 *)&pa.ssp_addr)->sin6_addr,16))
				goto n6;
		memset(&ap,0,sizeof(ap));
		memcpy(&ap.spp_address,ptr,sizeof(struct sockaddr_in6));
		ll=sizeof(ap);
		if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto n6;
		ap.spp_flags|=SPP_HB_DISABLE;
		ap.spp_flags&=~SPP_HB_ENABLE;
		ll=setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap));
n6:		ptr+=sizeof(struct sockaddr_in6);
		break;
	case AF_INET:
		if(family!=AF_INET||
			!memcmp(&((struct sockaddr_in *)ptr)->sin_addr,
			    &((struct sockaddr_in *)&pa.ssp_addr)->sin_addr,4))
				goto n4;
		memset(&ap,0,sizeof(ap));
		memcpy(&ap.spp_address,ptr,sizeof(struct sockaddr_in));
		ll=sizeof(ap);
		if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto n4;
		ap.spp_flags|=SPP_HB_DISABLE;
		ap.spp_flags&=~SPP_HB_ENABLE;
		setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap));
n4:		ptr+=sizeof(struct sockaddr_in);
		break;
	}

	return s;

err2:	close(s);
err1:	return -1;
#else
	return -1;
#endif
}

int sctpomconnect(void *dst,int timeout,int keepalive,int sendbuf,int maxseg,
	int srcaddr,int efd)
{
#ifndef NO_SCTP
	int family;
	int s;
	int x;
	int v1;
	int v2;
	int id;
	unsigned int y;
	socklen_t ll;
	void *ptr;
	struct sockaddr *sa;
	struct sctp_getaddrs *ga;
	struct linger l;
	struct pollfd p[2];
	struct sctp_rtoinfo rto;
	struct sctp_paddrparams ap;
	union sctp_event_subscribe_kernelfix ev;
	struct sctp_prim pa;
	struct msghdr msg;
	struct iovec v;
	union sctp_notification *notify;
	unsigned char bfr[8192];

	family=((struct sockaddr_in *)dst)->sin_family;
	if((s=socket(family,SOCK_SEQPACKET|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_SCTP))==-1)goto err1;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_SCTP,SCTP_NODELAY,&x,sizeof(x)))goto err2;
	ll=sizeof(rto);
	memset(&rto,0,sizeof(rto));
	if(getsockopt(s,SOL_SCTP,SCTP_RTOINFO,&rto,&ll))goto err2;
	rto.srto_initial=500;
	rto.srto_min=50;
	rto.srto_max=5000;
	if(setsockopt(s,SOL_SCTP,SCTP_RTOINFO,&rto,sizeof(rto)))goto err2;
	if(keepalive)
	{
		v1=SPP_HB_ENABLE;
		v2=SPP_HB_DISABLE;
	}
	else
	{
		v1=SPP_HB_DISABLE;
		v2=SPP_HB_ENABLE;
	}
	ll=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto err2;
	ap.spp_hbinterval=keepalive*1000;
	ap.spp_pathmaxrxt=3;
	ap.spp_sackdelay=200;
	ap.spp_flags|=v1|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~(v2|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO);
	if(setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap)))
		goto err2;
	if(sendbuf)if(setsockopt(s,SOL_SOCKET,SO_SNDBUF,&sendbuf,
		sizeof(sendbuf)))goto err2;
	if(maxseg)if(setsockopt(s,IPPROTO_SCTP,SCTP_MAXSEG,&maxseg,
		sizeof(maxseg)))goto err2;
	l.l_onoff=0;
	l.l_linger=5;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))goto err2;

	if(srcaddr)
	{
		if(!(sa=getsrcaddr(dst,family==AF_INET6?
		    sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in))))
			goto err2;
		if(bind(s,sa,family==AF_INET6?
			sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)))
		{
			free(sa);
			goto err2;
		}
		free(sa);
	}

	memset(&ev,0,sizeof(ev));
	ev.ev.sctp_association_event=1;
	if(setsockopt(s,SOL_SCTP,SCTP_EVENTS,&ev,geteventsize()))goto err2;

	if(connect(s,(struct sockaddr *)dst,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)))
	{
		if(errno!=EINPROGRESS)goto err2;
		p[0].fd=efd;
		p[0].events=POLLIN;
		p[1].fd=s;
		p[1].events=POLLOUT;
		if(poll(p,2,timeout*1000)<1)goto err2;
		if(p[0].revents&POLLIN)goto err2;
		if(!(p[1].revents&&POLLOUT))goto err2;
		y=sizeof(x);
		if(getsockopt(s,SOL_SOCKET,SO_ERROR,&x,&y))goto err2;
		if(x)goto err2;
	}

	p[0].fd=efd;
	p[0].events=POLLIN;
	p[1].fd=s;
	p[1].events=POLLIN;
	if(poll(p,2,500)<1)goto err2;
	if(p[0].revents&POLLIN)goto err2;
	if(!(p[1].revents&&POLLIN))goto err2;
	memset(&msg,0,sizeof(msg));
	msg.msg_iov=&v;
	msg.msg_iovlen=1;
	v.iov_base=bfr;
	v.iov_len=sizeof(bfr);
	if(((x=recvmsg(s,&msg,MSG_NOSIGNAL|MSG_WAITALL))==-1))goto err2;
	if(!(msg.msg_flags&MSG_NOTIFICATION)||
		x!=sizeof(notify->sn_assoc_change))goto err2;
	notify=(union sctp_notification *)bfr;
	if(notify->sn_header.sn_type!=SCTP_ASSOC_CHANGE)goto err2;
	if(notify->sn_assoc_change.sac_state!=SCTP_COMM_UP)goto err2;
	id=notify->sn_assoc_change.sac_assoc_id;

	pa.ssp_assoc_id=id;
	memcpy(&pa.ssp_addr,dst,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in));
	if(setsockopt(s,SOL_SCTP,SCTP_PRIMARY_ADDR,&pa,sizeof(pa)))goto err2;
	ll=sizeof(bfr);
	ga=(void *)bfr;
	memset(ga,0,sizeof(struct sctp_getaddrs));
	ga->assoc_id=id;
	if(!getsockopt(s,SOL_SCTP,SCTP_GET_PEER_ADDRS,ga,&ll))
		for(x=0,ptr=bfr+sizeof(struct sctp_getaddrs);x<ga->addr_num;x++)
			switch(((struct sockaddr *)ptr)->sa_family)
	{
	case AF_INET6:
		if(family!=AF_INET6||
			!memcmp(&((struct sockaddr_in6 *)ptr)->sin6_addr,
				&((struct sockaddr_in6 *)dst)->sin6_addr,16))
					goto n6;
		memset(&ap,0,sizeof(ap));
		memcpy(&ap.spp_address,ptr,sizeof(struct sockaddr_in6));
		ap.spp_assoc_id=id;
		ll=sizeof(ap);
		if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto n6;
		ap.spp_flags|=SPP_HB_DISABLE;
		ap.spp_flags&=~SPP_HB_ENABLE;
		ll=setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap));
n6:		ptr+=sizeof(struct sockaddr_in6);
		break;
	case AF_INET:
		if(family!=AF_INET||
			!memcmp(&((struct sockaddr_in *)ptr)->sin_addr,
				&((struct sockaddr_in *)dst)->sin_addr,4))
					goto n4;
		memset(&ap,0,sizeof(ap));
		memcpy(&ap.spp_address,ptr,sizeof(struct sockaddr_in));
		ap.spp_assoc_id=id;
		ll=sizeof(ap);
		if(getsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,&ll))goto n4;
		ap.spp_flags|=SPP_HB_DISABLE;
		ap.spp_flags&=~SPP_HB_ENABLE;
		setsockopt(s,SOL_SCTP,SCTP_PEER_ADDR_PARAMS,&ap,sizeof(ap));
n4:		ptr+=sizeof(struct sockaddr_in);
		break;
	}

	return s;

err2:	close(s);
err1:	return -1;
#else
	return -1;
#endif
}

void sctpomclose(int fd)
{
#ifndef NO_SCTP
	close(fd);
#endif
}

HOT int sctpomsend(int fd,void *dst,int dstlen,void *data,
	int len)
{
#ifndef NO_SCTP
	int l;

	if(UNLIKELY((l=sendto(fd,data,len,MSG_NOSIGNAL|MSG_EOR,dst,dstlen))
		==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	else if(UNLIKELY(l!=len))return -1;
	return l;
#else
	return -1;
#endif
}

HOT int sctpomrecv(int fd,void *data,int max)
{
#ifndef NO_SCTP
	int l;
	struct msghdr msg;
	struct iovec v;
	union sctp_notification *notify;

	memset(&msg,0,sizeof(msg));
	msg.msg_iov=&v;
	msg.msg_iovlen=1;
	v.iov_base=data;
	v.iov_len=max;

	if(UNLIKELY((l=recvmsg(fd,&msg,MSG_NOSIGNAL|MSG_WAITALL))==-1))
		switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	if(UNLIKELY(!(msg.msg_flags&MSG_EOR)))return -1;
	else if(UNLIKELY(msg.msg_flags&MSG_NOTIFICATION))
	{
		notify=data;
		if(l!=sizeof(notify->sn_assoc_change))return -1;
		if(notify->sn_header.sn_type!=SCTP_ASSOC_CHANGE)return -1;
		if(notify->sn_assoc_change.sac_state==SCTP_COMM_UP)return -2;
		return -1;
	}
	return l;
#else
	return -1;
#endif
}

COLD void sctpomlockselect(void)
{
#ifndef NO_SCTP
	int s;
	socklen_t ll;
	sctp_peeloff_flags_arg_t peel;

	if((s=socket(AF_INET6,SOCK_SEQPACKET,IPPROTO_SCTP))==-1)return;
	memset(&peel,0,sizeof(peel));
	ll=sizeof(peel);
	if(getsockopt(s,SOL_SCTP,SCTP_SOCKOPT_PEELOFF_FLAGS,&peel,&ll))
		if(errno==EINVAL)nolock=1;
	close(s);
#endif
}

void sctpomlock(void)
{
#ifndef NO_SCTP
	if(!nolock)pthread_mutex_lock(&mtx);
#endif
}

void sctpomunlock(void)
{
#ifndef NO_SCTP
	if(!nolock)pthread_mutex_unlock(&mtx);
#endif
}
