/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <pthread.h>
#include <sys/epoll.h>
#include <sys/timerfd.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <signal.h>
#include <poll.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef DEBUG
#include <sys/prctl.h>
#include <stdio.h>
#endif
#include "misc.h"
#include "crypto.h"
#include "tcp.h"
#include "sctp_oo.h"
#include "sctp_om.h"
#include "udp.h"
#include "rtp.h"
#include "unix.h"
#include "proto.h"
#include "keyio.h"
#include "tun.h"
#include "rxtx.h"
#include "parse.h"
#include "server.h"

#ifdef DEBUG
#define ERR(a) do							\
{									\
	fprintf(stderr,"Server abort: %s line %d\n",__FILE__,__LINE__);	\
	goto a;								\
} while(0)
#else
#define ERR(a)  goto a
#endif

struct srvconn
{
	struct srvconn *right;
	struct srvconn *left;
	int fd;
	int tfd;
	int state;
	int len;
	int setuplimit;
	int total;
	int minhdr;
	struct global *g;
	struct listen *l;
	void *rtp;
	struct crypto_x25519 *x25519;
	struct crypto_setup *setup;
	struct crypto_conn *conn;
	unsigned char bfr[0];
};

struct server_socket
{
	int fd;
	struct listen *l;
	void *ctx;
};

struct server
{
	void *user;
	struct crypto_conf *cnf;
	int (*checkauth)(void *data,int len,void *user,void **id);
	void (*handoff)(void *h,struct conn *c,void *user,void *id);
	struct srvconn *head;
	struct srvconn *tail;
	struct global *g;
	int total;
	struct server_socket listen[0];
};

struct srvusr
{
	int publen;
	void *pub;
	struct config *c;
	struct rxtx rt;
	pthread_t th;
};

struct srvctx
{
	int nuser;
	pthread_t th;
	void *h;
	struct server *s;
	struct srvusr u[0];
};

static void server_free(void *h,struct server *cnf,int fd,struct srvconn *c,
	struct crypto_keys **keys)
{
	epoll_ctl(fd,EPOLL_CTL_DEL,c->tfd,NULL);
	epoll_ctl(fd,EPOLL_CTL_DEL,c->fd,NULL);
	if(!keys)switch(c->l->proto)
	{
	case TCP:
		tcpclose(c->fd);
		break;
	case SCTP_OO:
		sctpooclose(c->fd);
		break;
	case SCTP_OM:
		sctpomclose(c->fd);
		break;
	case UDP:
	case UDPLITE:
		udpclose(c->fd);
		break;
	case UNIX:
	case PKTUNIX:
		unix_close(c->fd);
		break;
	default:close(c->fd);
		break;
	}
	close(c->tfd);
	if(c->x25519)crypto_initialfree(h,cnf->cnf,c->x25519);
	if(c->setup)crypto_setupfree(h,cnf->cnf,c->setup,keys);
	if(!keys)
	{
		if(c->conn)crypto_freeconn(h,c->conn);
		if(c->rtp)rtp_exit(c->rtp);
	}
	crypto_memclear(h,c->bfr,c->setuplimit);
	if(c->left)c->left->right=c->right;
	else cnf->head=c->right;
	if(c->right)c->right->left=c->left;
	else cnf->tail=c->left;
	free(c);
}

static inline void server_reap(void *h,struct server *cnf,int fd,int limit,
	int *total)
{
	int i;
	struct srvconn **c;
	struct itimerspec it;

	if(*total<=limit)return;

	for(i=1;i<cnf->g->setuptimeout;i++)for(c=&cnf->head;*c;)
	{
		if((*c)->state>=2||(LIKELY(!timerfd_gettime((*c)->tfd,&it))&&
			it.it_value.tv_sec>=i))c=&(*c)->right;
		else
		{
			server_free(h,cnf,fd,*c,NULL);
			if(--(*total)<=limit)return;
		}
	}

	for(i=1;i<=4;i++)for(c=&cnf->head;*c;)
	{
		if((*c)->state>=i)c=&(*c)->right;
		else
		{
			server_free(h,cnf,fd,*c,NULL);
			if(--(*total)<=limit)return;
		}
	}
}

static inline int server_recv(struct srvconn *c,void *bfr,int max)
{
	switch(c->l->proto)
	{
	case TCP:
		return tcprecv(c->fd,bfr,max);
	case SCTP_OO:
		return sctpoorecv(c->fd,bfr,max);
	case SCTP_OM:
		return sctpomrecv(c->fd,bfr,max);
	case UDP:
	case UDPLITE:
		return udprecv(c->fd,bfr,max);
	case UNIX:
		return unixrecv(c->fd,bfr,max);
	case PKTUNIX:
		return unixpktrecv(c->fd,bfr,max);
	default:return -1;
	}
}

static inline int server_send(struct srvconn *c,void *bfr,int len)
{
	switch(c->l->proto)
	{
	case TCP:
		return tcpsend(c->fd,bfr,len);
	case SCTP_OO:
		return sctpoosend(c->fd,bfr,len);
	case SCTP_OM:
		return sctpomsend(c->fd,NULL,0,bfr,len);
	case UDP:
	case UDPLITE:
		return udpsend(c->fd,bfr,len);
	case UNIX:
		return unixsend(c->fd,bfr,len);
	case PKTUNIX:
		return unixpktsend(c->fd,bfr,len);
	default:return -1;
	}
}

HOT static void *serverworker(void *data)
{
	int fd;
	int total=0;
	int len;
	int type;
	int mlen;
	int limit;
	int bits;
	int tmplen;
	int status;
	uint64_t dummy;
	void *h;
	void *msg;
	void *mem;
	void *tmp;
	struct srvconn *c;
	struct conn *conn;
	struct server_socket *a;
	struct server *cnf;
	struct pollfd p[3];
	struct epoll_event e;
	struct itimerspec it;
	sigset_t set;
	unsigned char bfr[1330];

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,NULL);
#ifdef DEBUG
	prctl(PR_SET_NAME,"serverworker",0,0,0);
#endif

	cnf=(struct server *)data;
	cnf->head=NULL;
	cnf->tail=NULL;
	limit=(cnf->g->pendinglimit*90)/100;

	memset(&it,0,sizeof(it));
	it.it_value.tv_sec=cnf->g->setuptimeout;

	p[0].fd=cnf->g->sefd;
	p[0].events=POLLIN;
	if((p[1].fd=epoll_create1(EPOLL_CLOEXEC))==-1)ERR(err1);
	p[1].events=POLLIN;
	if((p[2].fd=epoll_create1(EPOLL_CLOEXEC))==-1)ERR(err2);
	p[2].events=POLLIN;

	if(!(h=crypto_thread_init(cnf->g->glob)))ERR(err3);

	for(e.events=EPOLLIN,len=0;len<cnf->total;len++)
	{
		e.data.ptr=&cnf->listen[len];
		if(epoll_ctl(p[1].fd,EPOLL_CTL_ADD,cnf->listen[len].fd,&e))
			ERR(err4);
	}

	while(1)
	{
		if(UNLIKELY(poll(p,3,-1)<=0))continue;

		if(UNLIKELY(p[0].revents&POLLIN))break;

		if(p[1].revents&POLLIN)if(LIKELY(epoll_wait(p[1].fd,&e,1,0)==1))
				if(LIKELY(e.events&EPOLLIN))
		{
			a=e.data.ptr;

			switch(a->l->proto)
			{
			case TCP:
				fd=tcpaccept(a->fd,cnf->g->keepalive,
					cnf->g->tcpdefer,cnf->g->tcpuser);
				break;
			case SCTP_OO:
				fd=sctpooaccept(a->fd,cnf->g->keepalive);
				break;
			case SCTP_OM:
				fd=sctpomaccept(a->fd,cnf->g->keepalive);
				break;
			case UDP:
			case UDPLITE:
				fd=udp_accept(a->fd);
				break;
			case UNIX:
			case PKTUNIX:
				fd=unix_accept(a->fd);
				break;
			default:fd=-1;
			}

			if(LIKELY(fd!=-1))if(cnf->g->pendinglimit)
			{
				if(LIKELY(limit))server_reap(h,cnf,p[2].fd,
					limit,&total);
				if(UNLIKELY(total>cnf->g->pendinglimit))
					switch(a->l->proto)
				{
				case TCP:
					tcpclose(fd);
					fd=-1;
					break;
				case SCTP_OO:
					sctpooclose(fd);
					fd=-1;
					break;
				case SCTP_OM:
					sctpomclose(fd);
					fd=-1;
					break;
				case UDP:
				case UDPLITE:
					udpclose(fd);
					fd=-1;
					break;
				case UNIX:
				case PKTUNIX:
					unix_close(fd);
					fd=-1;
					break;
				default:close(fd);
					fd=-1;
					break;
				}
			}
			if(LIKELY(fd!=-1))
			{
				total++;
				tmplen=cnf->g->setuplimit;
				if(UNLIKELY(a->l->rtp))switch(a->l->proto)
				{
				case UDP:
				case UDPLITE:
				case PKTUNIX:
				    if((tmplen=(((cnf->g->setuplimit+183)/184)
					*188)+12)>1328)tmplen=1328;
					break;
				default:
				    if((tmplen=(((cnf->g->setuplimit+183)/184)
					*188)+14)>1330)tmplen=1330;
					break;
				}
				else tmplen=cnf->g->setuplimit;
				if(UNLIKELY(!(c=malloc(sizeof(struct srvconn)+
					tmplen))))ERR(neterr1);
				memset(c,0,sizeof(struct srvconn));
				c->l=a->l;
				c->g=cnf->g;
				c->setuplimit=tmplen;
				c->minhdr=crypto_hdrsize(h,NULL);
				if(UNLIKELY(c->l->rtp))switch(c->l->proto)
				{
				case UDP:
				case UDPLITE:
				case PKTUNIX:
					if(!(c->rtp=rtp_init(h,1)))ERR(neterr2);
					break;
				default:if(!(c->rtp=rtp_init(h,0)))ERR(neterr2);
					break;
				}
				c->fd=fd;
				if(UNLIKELY((c->tfd=timerfd_create(
					CLOCK_MONOTONIC,
					TFD_NONBLOCK|TFD_CLOEXEC))==-1))
					ERR(neterr3);
				if(UNLIKELY(timerfd_settime(c->tfd,0,&it,NULL)))
					ERR(neterr4);
				e.events=POLLIN;
				e.data.ptr=c;
				if(UNLIKELY(epoll_ctl(p[2].fd,EPOLL_CTL_ADD,
					c->tfd,&e)))ERR(neterr3);
				if(UNLIKELY(epoll_ctl(p[2].fd,EPOLL_CTL_ADD,
					fd,&e)))
				{
					epoll_ctl(p[2].fd,EPOLL_CTL_DEL,c->tfd,
						NULL);
neterr4:				close(c->tfd);
neterr3:				if(c->rtp)rtp_exit(c->rtp);
neterr2:				free(c);
neterr1:				switch(a->l->proto)
					{
					case TCP:
						tcpclose(fd);
						break;
					case SCTP_OO:
						sctpooclose(fd);
						break;
					case SCTP_OM:
						sctpomclose(fd);
						break;
					case UDP:
					case UDPLITE:
						udpclose(fd);
						break;
					case UNIX:
					case PKTUNIX:
						unix_close(fd);
						break;
					default:close(fd);
						break;
					}
					total--;
				}
				else
				{
					c->right=NULL;
					if(cnf->tail)
					{
						c->left=cnf->tail;
						c->left->right=c;
					}
					else
					{
						c->left=NULL;
						cnf->head=c;
					}
					cnf->tail=c;
				}
			}
		}

		if(p[2].revents&POLLIN)if(LIKELY(epoll_wait(p[2].fd,&e,1,0)==1))
			if(LIKELY(e.events&EPOLLIN))
		{
			c=e.data.ptr;
			if(UNLIKELY(read(c->tfd,&dummy,sizeof(dummy))==
				sizeof(dummy)))ERR(fail1);

			switch(c->l->proto)
			{
			case SCTP_OM:
			case UDP:
			case UDPLITE:
			case PKTUNIX:
				if(UNLIKELY((c->len=server_recv(c,c->bfr,
					c->setuplimit))<=0))
				{
					if(c->l->proto==SCTP_OM&&c->len==-2)
						ERR(skip);
					ERR(fail1);
				}
				if(UNLIKELY(c->rtp!=NULL))switch(c->l->proto)
				{
				case SCTP_OM:
					if(c->len<2)ERR(fail1);
					tmplen=(((unsigned char *)c->bfr)[0]<<8)
						|((unsigned char *)c->bfr)[1];
					if(tmplen!=c->len-2)ERR(fail1);
					goto rtpcnt1;
				default:goto rtpcnt2;
				}
				break;

			default:if(UNLIKELY((len=server_recv(c,c->bfr+c->len,
					c->setuplimit-c->len))<=0))ERR(fail1);
				c->len+=len;
				if(LIKELY(c->rtp==NULL))break;
				if(c->len<2)ERR(skip);
				tmplen=(((unsigned char *)c->bfr)[0]<<8)|
					((unsigned char *)c->bfr)[1];
				if(c->len!=tmplen+2)ERR(skip);
rtpcnt1:			memmove(c->bfr,c->bfr+2,tmplen);
				c->len=tmplen;
rtpcnt2:			if(rtp_unpack(c->rtp,c->bfr,c->len,&tmp,
					&tmplen))ERR(fail1);
				memcpy(c->bfr,tmp,tmplen);
				free(tmp);
				c->len=tmplen;
				break;
			}

			switch(c->state)
			{
			case 0:	if(UNLIKELY(!(c->x25519=crypto_initialreply(h,
					cnf->cnf,c->bfr,c->len,&msg,&len,
					c->rtp!=NULL))))ERR(fail1);
				if(UNLIKELY(c->rtp!=NULL))
				{
					if(UNLIKELY(rtp_pack(c->rtp,msg,len,
						&tmp,&tmplen)))ERR(fail1);
					memcpy(bfr,tmp,tmplen);
					free(tmp);
					msg=bfr;
					len=tmplen;
				}
				if(UNLIKELY(server_send(c,msg,len)!=len))
					ERR(fail1);
				c->state++;
				c->len=0;
				break;

			case 1:	if(c->len<c->minhdr-2)
				{
					switch(c->l->proto)
					{
					case SCTP_OM:
					case UDP:
					case UDPLITE:
					case PKTUNIX:
						ERR(fail1);
					default:break;
					}
					break;
				}

				switch(c->l->proto)
				{
				case UDP:
				case UDPLITE:
				case PKTUNIX:
					if(UNLIKELY(!(c->conn=
						crypto_initialderive(h,cnf->cnf,
						c->x25519,1,1,c->rtp!=NULL,
						c->l->proto==UDPLITE?69:0,
						c->l->small,c->l->strip,
						c->g->alt))))ERR(fail1);
					break;

				default:if(UNLIKELY(!(c->conn=
						crypto_initialderive(h,cnf->cnf,
						c->x25519,0,1,c->rtp!=NULL,0,
						c->l->small,c->l->strip,
						c->g->alt))))ERR(fail1);
					break;
				}
				c->minhdr=crypto_hdrsize(h,c->conn);

				if(UNLIKELY(crypto_unmask(c->conn,c->bfr,c->len,
					&c->total,&bits))||c->total<c->minhdr||
					bits||c->total>c->setuplimit)ERR(fail1);
				c->state++;

			case 2:	if(c->len<c->total)
				{
					switch(c->l->proto)
					{
					case SCTP_OM:
					case UDP:
					case UDPLITE:
					case PKTUNIX:
						ERR(fail1);
					default:break;
					}
					break;
				}

				if(!(msg=crypto_decode(h,c->conn,&type,c->bfr,
					c->len,&len,NULL))||
					UNLIKELY(type!=CONTROL))ERR(fail1);
				if(UNLIKELY(len>cnf->g->setuplimit))ERR(fail3);

				if(UNLIKELY(!(c->setup=crypto_realreply(h,
					cnf->cnf,msg,len,&mem,&mlen))))
						ERR(fail3);
				crypto_memclear(h,msg,len);
				free(msg);

				if(UNLIKELY(!(msg=crypto_encode(h,c->conn,
					CONTROL,mem,mlen,0,&len,c->rtp!=NULL))))
						ERR(fail1);

				if(UNLIKELY(c->rtp!=NULL))
				{
					if(UNLIKELY(rtp_pack(c->rtp,msg,len,
						&tmp,&tmplen)))ERR(fail2);
					free(msg);
					msg=tmp;
					len=tmplen;
				}

				if(UNLIKELY(server_send(c,msg,len)!=len))
					ERR(fail2);
				free(msg);

				crypto_freeconn(h,c->conn);
				switch(c->l->proto)
				{
				case UDP:
				case UDPLITE:
				case PKTUNIX:
					if(UNLIKELY(!(c->conn=
						crypto_derivekey(h,cnf->cnf,
						c->setup,1,c->rtp!=NULL,
						c->l->proto==UDPLITE?69:0,
						c->l->small,c->l->strip,
						c->g->alt))))ERR(fail1);
					break;

				default:if(UNLIKELY(!(c->conn=
						crypto_derivekey(h,cnf->cnf,
						c->setup,0,c->rtp!=NULL,0,
						c->l->small,c->l->strip,
						c->g->alt))))ERR(fail1);
					break;
				}
				c->minhdr=crypto_hdrsize(h,c->conn);

				c->state++;
				c->len=0;
				break;

			case 3:	if(c->len<c->minhdr-2)
				{
					switch(c->l->proto)
					{
					case SCTP_OM:
					case UDP:
					case UDPLITE:
					case PKTUNIX:
						ERR(fail1);
					default:break;
					}
					break;
				}

				if(UNLIKELY(crypto_unmask(c->conn,c->bfr,c->len,
					&c->total,&bits))||bits||
					UNLIKELY(c->total<c->minhdr)||
					UNLIKELY(c->total>c->setuplimit))
						ERR(fail1);
				c->state++;

			case 4:	if(c->len<c->total)
				{
					switch(c->l->proto)
					{
					case SCTP_OM:
					case UDP:
					case UDPLITE:
					case PKTUNIX:
						ERR(fail1);
					default:break;
					}
					break;
				}

				if(UNLIKELY(!(msg=crypto_decode(h,c->conn,&type,
					c->bfr,c->len,&len,NULL)))||
					UNLIKELY(type!=CONTROL))ERR(fail1);
				if(UNLIKELY(len>cnf->g->setuplimit))ERR(fail3);

				if(UNLIKELY(crypto_authparse(h,cnf->cnf,
					c->setup,msg,len)))ERR(fail3);
				crypto_memclear(h,msg,len);
				free(msg);

				if(!crypto_anonymousauth(h,cnf->cnf,c->setup))
				{
					msg=NULL;
					len=0;
				}
				else if(UNLIKELY(!(msg=crypto_getauthdata(h,
					cnf->cnf,c->setup,&len))))ERR(fail1);
				if(cnf->checkauth(msg,len,cnf->user,&mem))
					status=0;
				else status=1;

				if(UNLIKELY(crypto_authverify(h,cnf->cnf,
					c->setup)))ERR(fail1);

				if(UNLIKELY(crypto_authreply(h,cnf->cnf,
					c->setup,status,&msg,&len)))ERR(fail1);

				if(UNLIKELY(!(msg=crypto_encode(h,c->conn,
					CONTROL,msg,len,0,&len,c->rtp!=NULL))))
						ERR(fail1);

				if(UNLIKELY(c->rtp!=NULL))
				{
					if(UNLIKELY(rtp_pack(c->rtp,msg,len,
						&tmp,&tmplen)))ERR(fail2);
					free(msg);
					msg=tmp;
					len=tmplen;
				}

				if(UNLIKELY(server_send(c,msg,len)!=len)||
					!status)ERR(fail2);
				free(msg);

				if(status&&
					!crypto_wantsrekey(h,cnf->cnf,c->setup))
						switch(c->l->proto)
				{
				case UDP:
				case UDPLITE:
				case PKTUNIX:
					crypto_freeconn(h,c->conn);
					if(UNLIKELY(!(c->conn=
						crypto_authderive(h,cnf->cnf,
						c->setup,1,c->rtp!=NULL,
						c->l->proto==UDPLITE?69:0,
						c->l->small,c->l->strip,
						c->g->alt))))ERR(fail1);
					break;

				default:crypto_freeconn(h,c->conn);
					if(UNLIKELY(!(c->conn=
						crypto_authderive(h,cnf->cnf,
						c->setup,0,c->rtp!=NULL,0,
						c->l->small,c->l->strip,
						c->g->alt))))ERR(fail1);
					break;
				}

				if(UNLIKELY(!(conn=malloc(
					sizeof(struct conn)))))ERR(fail1);

				switch(c->l->proto)
				{
				case UDP:
				case UDPLITE:
				case PKTUNIX:
					conn->packet=1;
					break;
				default:conn->packet=0;
					break;
				}
				conn->proto=c->l->proto;
				conn->fd=c->fd;
				conn->curve=
					crypto_getcurve(h,cnf->cnf,c->setup);
				conn->keepalive=cnf->g->keepalive;
				conn->noop=0;
				conn->rtp=c->rtp;
				conn->dst=NULL;
				conn->dstlen=0;
				conn->conn[0]=c->conn;
				conn->conn[1]=c->conn;

				server_free(h,cnf,p[2].fd,c,&conn->keys);
				total--;

				cnf->handoff(h,conn,cnf->user,mem);
				break;
			}
		}

skip:		if(0)
		{
fail3:			crypto_memclear(h,msg,len);
fail2:			free(msg);
fail1:			server_free(h,cnf,p[2].fd,c,NULL);
			total--;
		}
	}

	while(total--)server_free(h,cnf,p[2].fd,cnf->head,NULL);

err4:	for(len=0;len<cnf->total;len++)
		epoll_ctl(p[1].fd,EPOLL_CTL_DEL,cnf->listen[len].fd,NULL);

	crypto_thread_exit(h);

err3:	close(p[2].fd);
err2:	close(p[1].fd);
err1:	pthread_exit(NULL);
}

static void handoff(void *h,struct conn *c,void *user,void *id)
{
	struct srvusr *u=id;

	c->rekey=u->c->rekey;
	if(write(u->rt.pfd[1],c,sizeof(struct conn))!=sizeof(struct conn))
	{
		switch(c->proto)
		{
		case TCP:
			tcpclose(c->fd);
			break;
		case SCTP_OO:
			sctpooclose(c->fd);
			break;
		case SCTP_OM:
			sctpomclose(c->fd);
			break;
		case UDP:
		case UDPLITE:
			udpclose(c->fd);
			break;
		case UNIX:
		case PKTUNIX:
			unix_close(c->fd);
			break;
		default:close(c->fd);
			break;
		}
		crypto_freepair(h,c->conn,3);
		if(c->rtp)free(c->rtp);
	}
	free(c);
}

static int checkauth(void *data,int len,void *user,void **id)
{
	int i;
	struct srvctx *ctx=user;

	if(!data)return -1;
	for(i=0;i<ctx->nuser;i++)
		if(len==ctx->u[i].publen&&!memcmp(data,ctx->u[i].pub,len))
	{
		*id=&ctx->u[i];
		return 0;
	}
	return -1;
}

COLD static int doexec(void *script,void *mode,void *dev)
{
	int status;
	pid_t pid;

	if(script)switch((pid=fork()))
	{
	case -1:return -1;
	case 0:	munlockall();
		execl(script,script,mode,dev,NULL);
		exit(255);
	default:if(waitpid(pid,&status,0)!=pid||status)return -1;
		break;
	}
	return 0;
}

COLD void *server_init(struct global *g,void *pass,int plen)
{
	int i;
	int ncnt;
	int scnt;
	struct config *c;
	struct listen *l;
	struct server *s;
	struct srvctx *ctx;
	void *h;
	void *key;

	if(!(h=crypto_thread_init(g->glob)))ERR(err1);
	if(!(key=loadkey(h,g->keyfile,pass,plen)))ERR(err2);

	for(ncnt=0,c=g->cfg;c;c=c->next)if(c->mode==SERVER&&c->connection)
		ncnt++;
	for(scnt=0,l=g->sockets;l;l=l->next)scnt++;
	if(!(s=malloc(sizeof(struct server)+scnt*sizeof(struct server_socket))))
		ERR(err3);
	s->g=g;
	if(!(s->cnf=crypto_config(h,key,g->setuplimit,g->x25519,g->secp,
		g->brainpool,g->minbits,g->maxbits,g->sigsecp,g->sigbrainpool,
		g->sigminbits,g->sigmaxbits,g->maxdata,g->datapad)))ERR(err4);
	key=NULL;
	s->checkauth=checkauth;
	s->handoff=handoff;
	s->total=scnt;
	for(i=0;i<scnt;i++)
	{
			s->listen[i].fd=-1;
			s->listen[i].ctx=NULL;
	}
	for(i=0,l=g->sockets;i<scnt;i++,l=l->next)switch(l->proto)
	{
	case TCP:
		s->listen[i].l=l;
		if((s->listen[i].fd=tcplisten(l->addr,l->port,g->keepalive,
			g->sendbuf,g->maxseg,g->tcpdefer,g->tcpuser))==-1)
				ERR(err5);
		break;
	case SCTP_OO:
		s->listen[i].l=l;
		if((s->listen[i].fd=sctpoolisten(l->addr,l->port,g->keepalive,
			g->sendbuf,g->maxseg))==-1)ERR(err5);
		break;
	case SCTP_OM:
		s->listen[i].l=l;
		if((s->listen[i].fd=sctpomlisten(l->addr,l->port,g->keepalive,
			g->sendbuf,g->maxseg))==-1)ERR(err5);
		break;
	case UDP:
		s->listen[i].l=l;
		if(!(s->listen[i].ctx=udp_listen_init(l->addr,l->port,
			g->portmin,g->portmax,g->sendbuf,0,g->frag,l->rtp,
			g->alt,g->sefd,g->glob)))ERR(err5);
		break;
	case UDPLITE:
		s->listen[i].l=l;
		if(!(s->listen[i].ctx=udp_listen_init(l->addr,l->port,
			g->portmin,g->portmax,g->sendbuf,l->rtp?20:8,g->frag,
			l->rtp,g->alt,g->sefd,g->glob)))ERR(err5);
		break;
	case UNIX:
	case PKTUNIX:
		s->listen[i].l=l;
		s->listen[i].ctx=l->fn;
		break;
	default:ERR(err5);
	}
	if(!(ctx=malloc(sizeof(struct srvctx)+ncnt*sizeof(struct srvusr))))
		ERR(err5);
	memset(ctx,0,sizeof(struct srvctx)+ncnt*sizeof(struct srvusr));
	for(c=g->cfg,i=0;i<ncnt;i++)ctx->u[i].rt.pfd[0]=ctx->u[i].rt.pfd[1]=-1;
	s->user=ctx;
	ctx->s=s;
	ctx->nuser=ncnt;
	ctx->h=h;
	for(c=g->cfg,i=0;i<ncnt;i++,c=c->next)
	{
		while(c)if(c->mode!=SERVER||!c->connection)c=c->next;
		else break;
		if(!(ctx->u[i].pub=loadpub(h,c->authfile[0],&ctx->u[i].publen)))
			ERR(err6);
		ctx->u[i].c=c;
		ctx->u[i].rt.c=c;
		ctx->u[i].rt.g=g;
		if(pipe2(ctx->u[i].rt.pfd,O_CLOEXEC|O_NONBLOCK|O_DIRECT))
			ERR(err6);
		if(!(ctx->u[i].rt.tun=opentun()))ERR(err6);
	}
	return ctx;

err6:	for(i=0;i<ncnt;i++)
	{
		if(ctx->u[i].pub)free(ctx->u[i].pub);
		if(ctx->u[i].rt.pfd[0]!=-1)close(ctx->u[i].rt.pfd[0]);
		if(ctx->u[i].rt.pfd[1]!=-1)close(ctx->u[i].rt.pfd[1]);
		if(ctx->u[i].rt.tun)closetun(ctx->u[i].rt.tun);
	}
	free(ctx);
err5:	for(i=0;i<scnt;i++)
	{
		if(s->listen[i].fd!=-1)close(s->listen[i].fd);
		if(s->listen[i].ctx)udp_listen_exit(s->listen[i].ctx);
	}
	crypto_free(h,s->cnf);
err4:	free(s);
err3:	if(key)crypto_free_key(h,key);
err2:	crypto_thread_exit(h);
err1:	return NULL;
}

COLD int server_unix_init(void *data)
{
	int i;
	struct srvctx *ctx=data;

	for(i=0;i<ctx->s->total;i++)switch(ctx->s->listen[i].l->proto)
	{
	case UNIX:
		if((ctx->s->listen[i].fd=unix_listen(ctx->s->listen[i].ctx,0))
			==-1)return -1;
		break;
	case PKTUNIX:
		if((ctx->s->listen[i].fd=unix_listen(ctx->s->listen[i].ctx,1))
			==-1)return -1;
		break;
	}
	return 0;
}

COLD int server_start(void *data,int exec)
{
	int i;
	struct srvctx *ctx=data;

	for(i=0;i<ctx->s->total;i++)if(ctx->s->listen[i].l->proto!=UNIX&&
		ctx->s->listen[i].l->proto!=PKTUNIX)if(ctx->s->listen[i].ctx)
		if((ctx->s->listen[i].fd=
			udp_listen_start(ctx->s->listen[i].ctx))==-1)
	{
		for(;i>=0;i--)if(ctx->s->listen[i].l->proto!=UNIX&&
			ctx->s->listen[i].l->proto!=PKTUNIX)
			if(ctx->s->listen[i].ctx)
				udp_listen_stop(ctx->s->listen[i].ctx,1);
		goto err1;
	}
	for(i=0;i<ctx->nuser;i++)
	{
		if(pthread_create(&ctx->u[i].th,NULL,rxtx,&ctx->u[i].rt))
			goto fail;
		if(exec&&doexec(ctx->u[i].rt.c->script,"up",
			ctx->u[i].rt.tun->dev))
		{
			pthread_cancel(ctx->u[i].th);
			pthread_join(ctx->u[i].th,NULL);
fail:			while(--i>=0)
			{
				if(exec)doexec(ctx->u[i].rt.c->script,"down",
					ctx->u[i].rt.tun->dev);
				pthread_cancel(ctx->u[i].th);
				pthread_join(ctx->u[i].th,NULL);
			}
			goto err2;
		}
	}
	if(pthread_create(&ctx->th,NULL,serverworker,ctx->s))
	{
		for(i=0;i<ctx->nuser;i++)
		{
			if(exec)doexec(ctx->u[i].rt.c->script,"down",
				ctx->u[i].rt.tun->dev);
			pthread_cancel(ctx->u[i].th);
			pthread_join(ctx->u[i].th,NULL);
		}
		goto err2;
	}
	return 0;

err2:	for(i=0;i<ctx->s->total;i++)if(ctx->s->listen[i].l->proto!=UNIX&&
		ctx->s->listen[i].l->proto!=PKTUNIX)if(ctx->s->listen[i].ctx)
			udp_listen_stop(ctx->s->listen[i].ctx,1);
err1:	return -1;
}

COLD void server_stop(void *data,int exec)
{
	int i;
	struct srvctx *ctx=data;

	pthread_join(ctx->th,NULL);
	for(i=0;i<ctx->nuser;i++)
	{
		if(exec)doexec(ctx->u[i].rt.c->script,"down",
			ctx->u[i].rt.tun->dev);
		pthread_join(ctx->u[i].th,NULL);
	}
	for(i=0;i<ctx->s->total;i++)if(ctx->s->listen[i].l->proto!=UNIX&&
		ctx->s->listen[i].l->proto!=PKTUNIX)if(ctx->s->listen[i].ctx)
			udp_listen_stop(ctx->s->listen[i].ctx,0);
}

COLD void server_exit(void *data)
{
	int i;
	struct srvctx *ctx=data;
	struct pollfd p;
	struct conn c;

	for(p.events=POLLIN,i=0;i<ctx->nuser;i++)
	{
		closetun(ctx->u[i].rt.tun);
		free(ctx->u[i].pub);
		p.fd=ctx->u[i].rt.pfd[0];
		while(poll(&p,1,0)==1)if(!(p.revents&POLLIN))break;
		else if(read(p.fd,&c,sizeof(c))!=sizeof(c))break;
		else
		{
			switch(c.proto)
			{
			case TCP:
				tcpclose(c.fd);
				break;
			case SCTP_OO:
				sctpooclose(c.fd);
				break;
			case SCTP_OM:
				sctpomclose(c.fd);
				break;
			case UDP:
			case UDPLITE:
				udpclose(c.fd);
				break;
			case UNIX:
			case PKTUNIX:
				unix_close(c.fd);
				break;
			default:close(c.fd);
				break;
			}
			crypto_freepair(ctx->h,c.conn,3);
		}
		close(ctx->u[i].rt.pfd[0]);
		close(ctx->u[i].rt.pfd[1]);
	}
	for(i=0;i<ctx->s->total;i++)
	{
		if(ctx->s->listen[i].l->proto==UNIX||
			ctx->s->listen[i].l->proto==PKTUNIX)
		{
			unix_close(ctx->s->listen[i].fd);
			unix_clean(ctx->s->listen[i].ctx);
		}
		else if(ctx->s->listen[i].ctx)
			udp_listen_exit(ctx->s->listen[i].ctx);
		else close(ctx->s->listen[i].fd);
	}
	crypto_free(ctx->h,ctx->s->cnf);
	free(ctx->s);
	crypto_thread_exit(ctx->h);
	free(ctx);
}

void server_keyupdate(void *data)
{
	void *key;
	struct srvctx *ctx=data;

	if(!(key=loadkey(ctx->h,ctx->s->g->keyfile,NULL,0)))return;
	crypto_reconfig(ctx->h,ctx->s->cnf,key);
}
