/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/timerfd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <poll.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include "misc.h"
#include "tcp.h"

static int netread(int s,void *bfr,int len,int tfd,int efd)
{
	int pos;
	int l;
	struct pollfd p[3];

	p[0].fd=efd;
	p[0].events=POLLIN;
	p[1].fd=tfd;
	p[1].events=POLLIN;
	p[2].fd=s;
	p[2].events=POLLIN;

	for(pos=0;pos<len;pos+=l)
	{
		while(poll(p,3,-1)<1);
		if(p[0].revents&POLLIN)return -1;
		if(p[0].revents&POLLIN)return -1;
		if(!(p[2].revents&&POLLIN))return -1;
		if((l=read(s,bfr+pos,len-pos))<=0)
		{
			if(l!=-1)return -1;
			if(errno!=EAGAIN&&errno!=EINTR)return -1;
			l=0;
		}
	}

	return 0;
}

COLD int tcplisten(void *addr,int port,int keepalive,int sendbuf,int maxseg,
	int defer,int user)
{
	int s;
	int x;
	int family;
	union
	{
		struct sockaddr_in a4;
		struct sockaddr_in6 a6;
	} a;

	if(port<1||port>65535)goto err1;
	memset(&a,0,sizeof(a));
	if(addr)
	{
		if(inet_pton(AF_INET,addr,&a.a4.sin_addr)==1)
		{
			family=AF_INET;
			a.a4.sin_family=AF_INET;
			a.a4.sin_port=htons(port);
		}
		else if(inet_pton(AF_INET6,addr,&a.a6.sin6_addr)==1)
		{
			family=AF_INET6;
			a.a6.sin6_family=AF_INET6;
			a.a6.sin6_port=htons(port);
		}
		else goto err1;
	}
	else
	{
		family=AF_INET6;
		a.a6.sin6_family=AF_INET6;
		a.a6.sin6_addr=in6addr_any;
		a.a6.sin6_port=htons(port);
	}
	if((s=socket(family,SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_TCP))==-1)goto err1;
	x=1;
	if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x)))goto err2;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	if(family==AF_INET6)
	{
		x=(addr?1:0);
		if(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
			goto err2;
	}
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_QUICKACK,&x,sizeof(x)))goto err2;
	if(keepalive)
	{
		x=3;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPCNT,&x,sizeof(x)))goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPIDLE,&x,sizeof(x)))
			goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPINTVL,&x,sizeof(x)))
			goto err2;
		x=1;
	}
	else x=0;
	if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&x,sizeof(x)))goto err2;
	x=user*1000;
	if(setsockopt(s,IPPROTO_TCP,TCP_USER_TIMEOUT,&x,sizeof(x)))goto err2;
	if(defer)if(setsockopt(s,IPPROTO_TCP,TCP_DEFER_ACCEPT,&defer,
		sizeof(defer)))goto err2;
	if(sendbuf)if(setsockopt(s,SOL_SOCKET,SO_SNDBUF,&sendbuf,
		sizeof(sendbuf)))goto err2;
	if(maxseg)if(setsockopt(s,IPPROTO_TCP,TCP_MAXSEG,&maxseg,
		sizeof(maxseg)))goto err2;
	if(bind(s,(struct sockaddr *)(&a),family==AF_INET?
		sizeof(struct sockaddr_in):sizeof(struct sockaddr_in6)))
			goto err2;
	if(listen(s,SOMAXCONN))goto err2;
	return s;

err2:	close(s);
err1:	return -1;
}

int tcpaccept(int fd,int keepalive,int defer,int user)
{
	int s;
	int x;
	struct linger l;
	struct pollfd p;

	if((s=accept4(fd,NULL,NULL,SOCK_NONBLOCK|SOCK_CLOEXEC))==-1)goto err1;
	if(defer)
	{
		p.fd=s;
		p.events=POLLIN;
		if(poll(&p,1,0)!=1||p.revents!=POLLIN)goto err2;
	}
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_QUICKACK,&x,sizeof(x)))goto err2;
	if(keepalive)
	{
		x=3;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPCNT,&x,sizeof(x)))goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPIDLE,&x,sizeof(x)))
			goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPINTVL,&x,sizeof(x)))
			goto err2;
		x=1;
	}
	else x=0;
	if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&x,sizeof(x)))goto err2;
	x=user*1000;
	if(setsockopt(s,IPPROTO_TCP,TCP_USER_TIMEOUT,&x,sizeof(x)))goto err2;
	x=5;
	if(setsockopt(s,IPPROTO_TCP,TCP_LINGER2,&x,sizeof(x)))goto err2;
	l.l_onoff=0;
	l.l_linger=5;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))goto err2;
	return s;

err2:	close(s);
err1:	return -1;
}

int tcpconnect(void *dst,int timeout,int keepalive,int sendbuf,int maxseg,
	int defer,int user,int efd)
{
	int family;
	int s;
	int x;
	unsigned int y;
	struct linger l;
	struct pollfd p[2];

	family=((struct sockaddr_in *)dst)->sin_family;
	if((s=socket(family,SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_TCP))==-1)goto err1;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_QUICKACK,&x,sizeof(x)))goto err2;
	if(keepalive)
	{
		x=3;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPCNT,&x,sizeof(x)))goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPIDLE,&x,sizeof(x)))
			goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPINTVL,&x,sizeof(x)))
			goto err2;
		x=1;
	}
	else x=0;
	if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&x,sizeof(x)))goto err2;
	x=5;
	if(setsockopt(s,IPPROTO_TCP,TCP_LINGER2,&x,sizeof(x)))goto err2;
	if(sendbuf)if(setsockopt(s,SOL_SOCKET,SO_SNDBUF,&sendbuf,
		sizeof(sendbuf)))goto err2;
	if(maxseg)if(setsockopt(s,IPPROTO_TCP,TCP_MAXSEG,&maxseg,
		sizeof(maxseg)))goto err2;
	x=defer?1:0;
	if(setsockopt(s,IPPROTO_TCP,TCP_DEFER_ACCEPT,&x,sizeof(x)))goto err2;
	x=user*1000;
	if(setsockopt(s,IPPROTO_TCP,TCP_USER_TIMEOUT,&x,sizeof(x)))goto err2;
	l.l_onoff=0;
	l.l_linger=5;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))goto err2;
	if(connect(s,(struct sockaddr *)dst,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)))
	{
		if(errno!=EINPROGRESS)goto err2;
		p[0].fd=efd;
		p[0].events=POLLIN;
		p[1].fd=s;
		p[1].events=POLLOUT;
		if(poll(p,2,timeout*1000)<1)goto err2;
		if(p[0].revents&POLLIN)goto err2;
		if(!(p[1].revents&&POLLOUT))goto err2;
		y=sizeof(x);
		if(getsockopt(s,SOL_SOCKET,SO_ERROR,&x,&y))goto err2;
		if(x)goto err2;
	}
	return s;

err2:	close(s);
err1:	return -1;
}

int tcpsocks5connect(void *proxy,void *host,int port,int timeout,int keepalive,
	int efd)
{
	int family;
	int mode;
	int hlen;
	int tfd;
	int s;
	int x;
	unsigned int y;
	struct linger l;
	struct in_addr a4;
	struct in6_addr a6;
	struct itimerspec it;
	struct pollfd p[3];
	unsigned char bfr[262];

	if(!((unsigned char *)host)[0]||(hlen=strlen(host))>255||
		port<1||port>65535)goto err1;

	if(inet_pton(AF_INET,host,&a4)==1)mode=1;
	else if(inet_pton(AF_INET6,host,&a6)==1)mode=2;
	else mode=0;

	family=((struct sockaddr_in *)proxy)->sin_family;
	if((s=socket(family,SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_TCP))==-1)goto err1;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_QUICKACK,&x,sizeof(x)))goto err2;
	if(keepalive)
	{
		x=3;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPCNT,&x,sizeof(x)))goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPIDLE,&x,sizeof(x)))
			goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPINTVL,&x,sizeof(x)))
			goto err2;
		x=1;
	}
	else x=0;
	if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&x,sizeof(x)))goto err2;
	x=5;
	if(setsockopt(s,IPPROTO_TCP,TCP_LINGER2,&x,sizeof(x)))goto err2;
	l.l_onoff=0;
	l.l_linger=5;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))goto err2;

	if((tfd=timerfd_create(CLOCK_MONOTONIC,TFD_NONBLOCK|TFD_CLOEXEC))==-1)
		goto err2;
	memset(&it,0,sizeof(it));
	it.it_value.tv_sec=timeout;
	if(timerfd_settime(tfd,0,&it,NULL))goto err3;

	if(connect(s,(struct sockaddr *)proxy,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)))
	{
		if(errno!=EINPROGRESS)goto err3;
		p[0].fd=efd;
		p[0].events=POLLIN;
		p[1].fd=tfd;
		p[1].events=POLLIN;
		p[2].fd=s;
		p[2].events=POLLOUT;
		if(poll(p,3,-1)<1)goto err3;
		if(p[0].revents&POLLIN)goto err3;
		if(p[1].revents&POLLIN)goto err3;
		if(!(p[2].revents&&POLLOUT))goto err3;
		y=sizeof(x);
		if(getsockopt(s,SOL_SOCKET,SO_ERROR,&x,&y))goto err3;
		if(x)goto err3;
	}

	bfr[0]=0x05;
	bfr[1]=0x01;
	bfr[2]=0x00;
	if(write(s,bfr,3)!=3)goto err3;

	if(netread(s,bfr,2,tfd,efd))goto err3;
	if(bfr[0]!=0x05||(bfr[1]&&bfr[1]!=0xff))goto err3;
	if(bfr[1])goto err3;

	switch(mode)
	{
	case 0:	bfr[0]=0x05;
		bfr[1]=0x01;
		bfr[2]=0x00;
		bfr[3]=0x03;
		bfr[4]=(unsigned char)hlen;
		memcpy(bfr+5,host,hlen);
		bfr[5+hlen]=port>>8;
		bfr[6+hlen]=port&0xff;
		x=hlen+7;
		break;

	case 1:	bfr[0]=0x05;
		bfr[1]=0x01;
		bfr[2]=0x00;
		bfr[3]=0x01;
		memcpy(bfr+4,&a4.s_addr,4);
		bfr[8]=port>>8;
		bfr[9]=port&0xff;
		x=10;
		break;

	case 2:	bfr[0]=0x05;
		bfr[1]=0x01;
		bfr[2]=0x00;
		bfr[3]=0x04;
		memcpy(bfr+4,a6.s6_addr,16);
		bfr[20]=port>>8;
		bfr[21]=port&0xff;
		x=22;
		break;

	default:goto err3;
	}
	if(write(s,bfr,x)!=x)goto err3;

	if(netread(s,bfr,4,tfd,efd))goto err3;
	if(bfr[0]!=0x05||bfr[1]>8||bfr[2]!=0x00)goto err3;
	switch(bfr[3])
	{
	case 0x01:
		if(netread(s,bfr+4,4,tfd,efd))goto err3;
		x=8;
		break;
	case 0x03:
		if(netread(s,bfr+4,1,tfd,efd))goto err3;
		if(netread(s,bfr+5,bfr[4],tfd,efd))goto err3;
		x=5+bfr[4];
		break;
	case 0x04:
		if(netread(s,bfr+4,16,tfd,efd))goto err3;
		x=20;
		break;
	}
	if(netread(s,bfr+x,2,tfd,efd))goto err3;
	if(bfr[1])goto err3;
	if(bfr[3]!=0x01&&bfr[3]!=0x04)goto err3;

	close(tfd);
	return s;

err3:	close(tfd);
err2:	close(s);
err1:	return -1;
}

int tcpproxyconnect(void *proxy,void *host,int port,int timeout,int keepalive,
	int efd)
{
	int family;
	int mode;
	int hlen;
	int tfd;
	int s;
	int x;
	unsigned int y;
	struct linger l;
	struct in_addr a4;
	struct in6_addr a6;
	struct itimerspec it;
	struct pollfd p[3];
	unsigned char bfr[2048];

	if(!((unsigned char *)host)[0]||(hlen=strlen(host))>255||
		port<1||port>65535)goto err1;

	if(inet_pton(AF_INET,host,&a4)==1)mode=1;
	else if(inet_pton(AF_INET6,host,&a6)==1)mode=2;
	else mode=0;

	family=((struct sockaddr_in *)proxy)->sin_family;
	if((s=socket(family,SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_TCP))==-1)goto err1;
	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))goto err2;
	x=1;
	if(setsockopt(s,SOL_TCP,TCP_QUICKACK,&x,sizeof(x)))goto err2;
	if(keepalive)
	{
		x=3;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPCNT,&x,sizeof(x)))goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPIDLE,&x,sizeof(x)))
			goto err2;
		x=keepalive;
		if(setsockopt(s,IPPROTO_TCP,TCP_KEEPINTVL,&x,sizeof(x)))
			goto err2;
		x=1;
	}
	else x=0;
	if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&x,sizeof(x)))goto err2;
	x=5;
	if(setsockopt(s,IPPROTO_TCP,TCP_LINGER2,&x,sizeof(x)))goto err2;
	l.l_onoff=0;
	l.l_linger=5;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))goto err2;

	if((tfd=timerfd_create(CLOCK_MONOTONIC,TFD_NONBLOCK|TFD_CLOEXEC))==-1)
		goto err2;
	memset(&it,0,sizeof(it));
	it.it_value.tv_sec=timeout;
	if(timerfd_settime(tfd,0,&it,NULL))goto err3;

	if(connect(s,(struct sockaddr *)proxy,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)))
	{
		if(errno!=EINPROGRESS)goto err3;
		p[0].fd=efd;
		p[0].events=POLLIN;
		p[1].fd=tfd;
		p[1].events=POLLIN;
		p[2].fd=s;
		p[2].events=POLLOUT;
		if(poll(p,3,-1)<1)goto err3;
		if(p[0].revents&POLLIN)goto err3;
		if(p[1].revents&POLLIN)goto err3;
		if(!(p[2].revents&&POLLOUT))goto err3;
		y=sizeof(x);
		if(getsockopt(s,SOL_SOCKET,SO_ERROR,&x,&y))goto err3;
		if(x)goto err3;
	}

	if(mode!=2)sprintf((char *)bfr,
		"CONNECT %s:%d HTTP/1.0\r\nHost: %s:%d\r\n\r\n",
		(char *)host,port,(char *)host,port);
	else sprintf((char *)bfr,
		"CONNECT [%s]:%d HTTP/1.0\r\nHost: [%s]:%d\r\n\r\n",
		(char *)host,port,(char *)host,port);
	x=strlen((char *)bfr);
	if(write(s,bfr,x)!=x)goto err3;

	p[0].fd=efd;
	p[0].events=POLLIN;
	p[1].fd=tfd;
	p[1].events=POLLIN;
	p[2].fd=s;
	p[2].events=POLLIN;
	for(x=0,y=0;y<2&&x<sizeof(bfr);x++)
	{
		while(poll(p,3,-1)<1);
		if(p[0].revents&POLLIN)goto err3;
		if(p[1].revents&POLLIN)goto err3;
		if(!(p[2].revents&&POLLIN))goto err3;
		switch(read(s,bfr+x,1))
		{
		case -1:if(errno==EAGAIN||errno==EINTR)
			{
				x--;
				continue;
			}
		case 0:	goto err3;
		}
		if(bfr[x]=='\r')continue;
		if(bfr[x]=='\n')y++;
		else y=0;
	}
	if(y!=2)goto err3;
	bfr[x]=0;
	if(strncmp((char *)bfr,"HTTP/1.0 200",12)&&
		strncmp((char *)bfr,"HTTP/1.1 200",12))goto err3;

	close(tfd);
	return s;

err3:	close(tfd);
err2:	close(s);
err1:	return -1;
}

void tcpclose(int fd)
{
	shutdown(fd,SHUT_RDWR);
	close(fd);
}

HOT int tcpsend(int fd,void *data,int len)
{
	int l;

	if(UNLIKELY((l=write(fd,data,len))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
}

HOT int tcprecv(int fd,void *data,int max)
{
	int l;

	if(UNLIKELY((l=read(fd,data,max))==-1))switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	return l;
}
