/*
 * This file is part of the vpnd2 project
 *
 * (C) 2020 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#define _GNU_SOURCE
#include <pthread.h>
#include <sys/eventfd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include "dns.h"

typedef struct
{
	int efd;
	int family;
	int valid;
	char *host;
	struct in_addr a4;
	struct in6_addr a6;
} DNSQUERY;

static void *dnsworker(void *data)
{
	int r;
	DNSQUERY *q=(DNSQUERY *)data;
	struct addrinfo *list=NULL;
	struct addrinfo *ptr;
	struct addrinfo hints;
	sigset_t set;
	uint64_t dummy=1;

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,NULL);

	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,NULL);
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	memset(&hints,0,sizeof(hints));
	hints.ai_family=AF_UNSPEC;
	hints.ai_flags=AI_IDN;

	r=getaddrinfo(q->host,NULL,&hints,&list);

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	if(!r&&list)for(ptr=list;ptr&&q->valid!=3;ptr=ptr->ai_next)
		if(ptr->ai_addr)
	{
		switch(ptr->ai_family)
		{
		case AF_INET:
			if(q->valid&1)break;
			q->a4=((struct sockaddr_in *)ptr->ai_addr)->sin_addr;
			q->valid|=1;
			break;

		case AF_INET6:
			if(q->valid&2)break;
			q->a6=((struct sockaddr_in6 *)ptr->ai_addr)->sin6_addr;
			q->valid|=2;
			break;
		}
	}

	if(!r&&list)freeaddrinfo(list);

	dummy=write(q->efd,&dummy,sizeof(dummy));

	pthread_exit(NULL);
}

int lookup(int timeout,int *family,char *host,void *result,int efd)
{
	pthread_t h;
	struct pollfd p[2];
	DNSQUERY q;

	if(inet_pton(AF_INET,host,result)==1)
	{
		*family=AF_INET;
		return 1;
	}

	if(inet_pton(AF_INET6,host,result)==1)
	{
		*family=AF_INET6;
		return 1;
	}

	if(timeout!=-1)
	{
		if(timeout<1||timeout>86400)return -1;
		timeout*=1000;
	}

	q.efd=eventfd(0,EFD_CLOEXEC);
	q.valid=0;
	q.host=host;

	if(q.efd==-1)
	{
		perror("eventfd");
		return -1;
	}

	if(pthread_create(&h,NULL,dnsworker,&q))
	{
		perror("pthread_create");
		close(q.efd);
		return -1;
	}

	p[0].fd=q.efd;
	p[0].events=POLLIN;
	p[1].fd=efd;
	p[1].events=POLLIN;
	p[1].revents=0;

	if(poll(p,efd==-1?1:2,timeout)<=0)pthread_cancel(h);
	else if(p[1].revents&POLLIN)pthread_cancel(h);
	else if(!(p[0].revents&POLLIN))pthread_cancel(h);
	pthread_join(h,NULL);
	close(q.efd);

	if(p[1].revents)return -2;

	if(q.valid&1)
	{
		*family=AF_INET;
		memcpy(result,&q.a4,sizeof(q.a4));
		return 1;
	}

	if(q.valid&2)
	{
		*family=AF_INET6;
		memcpy(result,&q.a6,sizeof(q.a6));
		return 1;
	}

	return 0;
}
