/*
 * This file is part of the vpnd2 project
 *
 * (C) 2020 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <limits.h>
#include <poll.h>
#include <time.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#ifdef HTTP2
#include <sys/timerfd.h>
#include <nghttp2/nghttp2.h>
#endif
#include "dns.h"
#include "network.h"
#include "s4client.h"
#include "s5client.h"
#include "httpconn.h"
#include "tlsclient.h"

#define BUFSIZE	65536
#define STRING	0
#define NUMBER	1
#define BOOLEAN	2

struct parsecfg
{
	const char *name;
	const int type;
	const int min;
	const int max;
	const int offset;
};

struct config
{
	char *proxyhost;
	char *proxytype;
	char *cafile;
	char *servername;
	char *rcafile;
	char *rservername;
	char *rurl;
	char *ragent;
	char *emupath;
	char *emutype;
#ifdef HTTP2
	char *h2page;
#endif
	unsigned int proxyport;
	unsigned int conntimeout;
	unsigned int fastconnect;
	unsigned int keepalive;
	unsigned int verifyhost;
	unsigned int verifychain;
	unsigned int redirect;
	unsigned int rverifyhost;
	unsigned int rverifychain;
	unsigned int usealpn;
	unsigned int http11;
#ifdef HTTP2
	unsigned int h2ping;
#endif
};

typedef struct
{
	int fd;
	void *tlscfg;
	void *tlsconn;
} CTX;

#ifdef HTTP2

#define ALPNENTRIES     2

typedef struct
{
	int total;
	int bfrfill;
	char *bfr;
} H2DATA;

typedef struct
{
	int fd;
	int flags;
	int32_t id;
	void *tlsctx;
	nghttp2_session *sess;
	H2DATA h2in;
	H2DATA h2out;
} H2CTX;

/* http/1.1 must be last! */

static const char *alpn[2]=
{
	"h2",
	"http/1.1"
};

static const nghttp2_settings_entry settings[2]=
{
	{NGHTTP2_SETTINGS_ENABLE_PUSH,0},
	{NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS,10},
};

#else

#define ALPNENTRIES     1

static const char *alpn[1]=
{       
	"http/1.1"
};

#endif

static const struct parsecfg items[]=
{
	{"proxyhost",STRING,0,0,offsetof(struct config,proxyhost)},
	{"proxyport",NUMBER,1,65535,offsetof(struct config,proxyport)},
	{"proxytype",STRING,0,0,offsetof(struct config,proxytype)},
	{"connecttimeout",NUMBER,1,60,offsetof(struct config,conntimeout)},
	{"fastconnect",BOOLEAN,0,0,offsetof(struct config,fastconnect)},
	{"keepalive",NUMBER,15,120,offsetof(struct config,keepalive)},
	{"servername",STRING,0,0,offsetof(struct config,servername)},
	{"verifyhost",BOOLEAN,0,0,offsetof(struct config,verifyhost)},
	{"verifychain",BOOLEAN,0,0,offsetof(struct config,verifychain)},
	{"cafile",STRING,0,0,offsetof(struct config,cafile)},
	{"redirect",BOOLEAN,0,0,offsetof(struct config,redirect)},
	{"redirurl",STRING,0,0,offsetof(struct config,rurl)},
	{"rediragent",STRING,0,0,offsetof(struct config,ragent)},
	{"redirservername",STRING,0,0,offsetof(struct config,rservername)},
	{"redirverifyhost",BOOLEAN,0,0,offsetof(struct config,rverifyhost)},
	{"redirverifychain",BOOLEAN,0,0,offsetof(struct config,rverifychain)},
	{"redircafile",STRING,0,0,offsetof(struct config,rcafile)},
	{"alpn",BOOLEAN,0,0,offsetof(struct config,usealpn)},
	{"http11",BOOLEAN,0,0,offsetof(struct config,http11)},
#ifdef HTTP2
	{"h2ping",NUMBER,15,120,offsetof(struct config,h2ping)},
	{"h2page",STRING,0,0,offsetof(struct config,h2page)},
	{"emupath",STRING,0,0,offsetof(struct config,emupath)},
	{"emutype",STRING,0,0,offsetof(struct config,emutype)},
#endif
	{NULL,0,0,0,0}
};

static const struct
{
	char *name;
	int emu;
} emulist[]=
{
#ifdef HTTP2
	{"brave110lx",TLS_CLIENT_EMULATION_BRAVE_110_LINUX},
	{"brave111a10",TLS_CLIENT_EMULATION_BRAVE_111_ANDROID_10},
	{"chrome84a10",TLS_CLIENT_EMULATION_CHROME_84_ANDROID_10},
	{"chromium84lx",TLS_CLIENT_EMULATION_CHROMIUM_84_LINUX},
	{"firefox68a10",TLS_CLIENT_EMULATION_FIREFOX_68_ANDROID_10},
	{"firefox78lx",TLS_CLIENT_EMULATION_FIREFOX_78_LINUX},
	{"kiwi77a10",TLS_CLIENT_EMULATION_KIWI_77_ANDROID_10},
	{"konqueror50lx",TLS_CLIENT_EMULATION_KONQUEROR_5_O_LINUX},
	{"opera59a10",TLS_CLIENT_EMULATION_OPERA_59_ANDROID_10},
	{"opera69lx",TLS_CLIENT_EMULATION_OPERA_69_LINUX},
	{"vivaldi31a10",TLS_CLIENT_EMULATION_VIVALDI_31_ANDROID_10},
	{"opera70lx",TLS_CLIENT_EMULATION_OPERA_70_LINUX},
#endif
	{NULL,-1},
};

static int tmocalc(struct timespec *start,int timeout)
{
	struct timespec curr;

	if(clock_gettime(CLOCK_MONOTONIC,&curr))return 0;
	if(curr.tv_nsec<start->tv_nsec)curr.tv_sec-=1;
	curr.tv_sec-=start->tv_sec;
	if(timeout<curr.tv_sec)return 0;
	return timeout-curr.tv_sec;
}

typedef union
{
	struct sockaddr_in a4;
	struct sockaddr_in6 a6;
} ADDR;

typedef struct
{
	ADDR dest;
	ADDR proxy;
	struct timespec *start;
	char *tlshost;
	int family;
	int proxyfamily;
	int port;
	int timeout;
	int keepalive;
	int fastconnect;
	int mode;
	int ip;
	int efd;
	int fd;
} CONNECTION;

static int doconnect(void *user)
{
	CONNECTION *c=user;
	int tmo;
	int port;

	if(!(tmo=tmocalc(c->start,c->timeout)))goto err1;

	switch(c->mode)
	{
	case 0:	switch(c->family)
		{
		case AF_INET:
			port=c->port;
			if((c->fd=tcp4connect(NULL,&c->dest.a4.sin_addr,&port,
				tmo,c->efd,(c->keepalive?KEEPALIVE:0)|
				(c->fastconnect?FAST_CONNECT:0),NULL,
				c->keepalive))==-1)goto err1;
			break;

		case AF_INET6:
			port=c->port;
			if((c->fd=tcp6connect(NULL,&c->dest.a6.sin6_addr,&port,
				tmo,c->efd,(c->keepalive?KEEPALIVE:0)|
				(c->fastconnect?FAST_CONNECT:0),NULL,
				c->keepalive))==-1)goto err1;
			break;

		default:goto err1;
		}
		break;

	case 1:	if(c->family!=AF_INET)goto err1;
		if((c->fd=s4connect((void *)&c->proxy,"",&c->dest.a4,NULL,tmo,
			c->efd,c->keepalive,c->fastconnect))==-1)goto err1;
		break;

	case 2:	if(c->ip)if(c->family!=AF_INET)goto err1;
		c->dest.a4.sin_port=htons(port);
		if((c->fd=s4connect((void *)&c->proxy,"",&c->dest.a4,
			c->ip?NULL:c->tlshost,tmo,c->efd,c->keepalive,
			c->fastconnect))==-1)goto err1;
		break;

	case 3:	if((c->fd=s5connect((void *)&c->proxy,(void *)&c->dest,c->port,
			NULL,tmo,c->efd,c->keepalive,c->fastconnect))==-1)
				goto err1;
		break;

	case 4:	if((c->fd=s5connect((void *)&c->proxy,(void *)&c->dest,c->port,
			c->ip?NULL:c->tlshost,tmo,c->efd,c->keepalive,
			c->fastconnect))==-1)goto err1;
		break;

	case 5:	if((c->fd=httpconnect((void *)&c->proxy,(void *)&c->dest,
			c->tlshost,c->port,tmo,c->efd,c->keepalive,
			c->fastconnect))==-1)goto err1;
		break;
	default:goto err1;
	}

	return c->fd;

err1:	return -1;
}

static int remoteconnect(char *tlshost,int port,char *svrhost,int svrport,
	int efd,struct config *c,CTX *ctx,int mode,struct timespec *start,
	int emu)
{
	int tmo;
	int verifyhost=(mode?c->verifyhost:c->rverifyhost);
	int verifychain=(mode?c->verifychain:c->rverifychain);
	char *cafile=(mode?c->cafile:c->rcafile);
	char *servername=(mode?c->servername:c->rservername);
	CONNECTION con;
	union
	{
		struct in_addr a4;
		struct in6_addr a6;
	} proxy;
	union
	{
		struct in_addr a4;
		struct in6_addr a6;
	} dest;

	con.start=start;
	con.tlshost=tlshost;
	con.family=AF_UNSPEC;
	con.proxyfamily=AF_UNSPEC;
	con.port=port;
	con.timeout=c->conntimeout;
	con.keepalive=c->keepalive;
	con.fastconnect=c->fastconnect;
	con.ip=0;
	con.efd=efd;

	if(c->proxyhost)
	{
		if(!strcmp(c->proxytype,"socks4"))con.mode=1;
		else if(!strcmp(c->proxytype,"socks4a"))con.mode=2;
		else if(!strcmp(c->proxytype,"socks5"))con.mode=3;
		else if(!strcmp(c->proxytype,"socks5h"))con.mode=4;
		else if(!strcmp(c->proxytype,"http"))con.mode=5;
	}
	else con.mode=0;

	if(inet_pton(AF_INET,tlshost,&dest)==1)con.ip=1;
	else if(inet_pton(AF_INET6,tlshost,&dest)==1)con.ip=1;

	if(emu!=-1)
	{
		if(!(ctx->tlscfg=tls_client_emulation_init(emu,
			TLS_CLIENT_EMULATION_STRICT)))goto err1;
	}
	else if(!(ctx->tlscfg=tls_client_init(TLS_CLIENT_TLS_1_2)))goto err1;
	if(verifychain)if(tls_client_add_cafile(ctx->tlscfg,cafile))
		goto err2;

	if(!(tmo=tmocalc(start,c->conntimeout)))goto err2;
	if(c->proxyhost)
	{
		if(!lookup(tmo,&con.proxyfamily,c->proxyhost,&proxy,efd))
			goto err2;
		if(!strcmp(c->proxytype,"socks4")||
			!strcmp(c->proxytype,"socks5")||con.ip)
		{
			if(!(tmo=tmocalc(start,c->conntimeout)))goto err2;
			if(!lookup(tmo,&con.family,tlshost,&dest,efd))
				goto err2;
		}
	}
	else if(!lookup(tmo,&con.family,tlshost,&dest,efd))goto err2;

	if(!(tmo=tmocalc(start,c->conntimeout)))goto err2;

	memset(&con.dest,0,sizeof(con.dest));
	switch(con.family)
	{
	case AF_INET:
		con.dest.a4.sin_family=AF_INET;
		con.dest.a4.sin_addr=dest.a4;
		con.dest.a4.sin_port=htons(port);
		break;
	case AF_INET6:
		con.dest.a6.sin6_family=AF_INET6;
		con.dest.a6.sin6_addr=dest.a6;
		con.dest.a6.sin6_port=htons(port);
		break;
	}

	if(con.mode)
	{
		memset(&con.proxy,0,sizeof(con.proxy));
		switch(con.proxyfamily)
		{
		case AF_INET:
			con.proxy.a4.sin_family=AF_INET;
			con.proxy.a4.sin_addr=proxy.a4;
			con.proxy.a4.sin_port=htons(c->proxyport);
			break;
		case AF_INET6:
			con.proxy.a6.sin6_family=AF_INET6;
			con.proxy.a6.sin6_addr=proxy.a6;
			con.proxy.a6.sin6_port=htons(c->proxyport);
			break;
		default:goto err2;
		}

	}

	if(emu!=-1)
	{
		if(!(ctx->tlsconn=tls_client_emulation_connect(ctx->tlscfg,5000,
			servername?servername:tlshost,verifyhost,NULL,NULL,NULL,
			doconnect,&con)))goto err2;
		ctx->fd=con.fd;
	}
	else
	{
		if(doconnect(&con)==-1)goto err2;
		ctx->fd=con.fd;

		if(!(tmo=tmocalc(start,c->conntimeout)))goto err3;
		tmo*=1000;

		if(c->usealpn)
		{
			if(c->http11)
			{
				if(tls_client_set_alpn(ctx->tlscfg,1,
					(char **)(&alpn[ALPNENTRIES-1])))
						goto err3;
			}
			else if(tls_client_set_alpn(ctx->tlscfg,ALPNENTRIES,
				(char **)alpn))goto err3;
		}

		if(!(ctx->tlsconn=tls_client_connect(ctx->tlscfg,con.fd,tmo,
			servername?servername:tlshost,verifyhost,NULL)))
			goto err3;
	}

	return 0;

err3:	close(con.fd);
err2:	tls_client_fini(ctx->tlscfg);
err1:	return -1;
}

#ifdef HTTP2

static ssize_t send_callback(nghttp2_session *session,const uint8_t *data,
	size_t length,int flags, void *user_data)
{       
	int len;
	H2CTX *ctx=user_data;

	if((len=tls_client_write(ctx->tlsctx,(void *)data,length))==-1)
	{       
		if(errno==EAGAIN)return NGHTTP2_ERR_WOULDBLOCK;
		return NGHTTP2_ERR_CALLBACK_FAILURE;
	}
	return len;
}

static int on_header_callback(nghttp2_session *session,
	const nghttp2_frame *frame,const uint8_t *name, size_t namelen,
	const uint8_t *value,size_t valuelen,uint8_t flags,void *user_data)
{
	H2CTX *ctx=user_data;

	if(frame->hd.type==NGHTTP2_HEADERS&&frame->hd.stream_id==ctx->id)
		if(!strcasecmp((char *)name,":status"))
			if(!strcmp((char *)value,"200"))ctx->flags|=2;
	return 0;
}

static int on_frame_recv_callback(nghttp2_session *session,
	const nghttp2_frame *frame, void *user_data)
{
	H2CTX *ctx=user_data;

	if(frame->hd.type==NGHTTP2_HEADERS&&
		frame->headers.cat==NGHTTP2_HCAT_RESPONSE&&
		frame->hd.stream_id==ctx->id)
			ctx->flags|=1;
	return 0;
}

static int on_data_chunk_recv_callback(nghttp2_session *session,uint8_t flags,
	int32_t stream_id,const uint8_t *data,size_t len,void *user_data)
{
	int i;
	H2CTX *ctx=user_data;

	if(stream_id==ctx->id)
	{
		if(ctx->flags&0x8)
		{
			if(ctx->h2in.bfrfill+len>=ctx->h2in.total)
				return NGHTTP2_ERR_CALLBACK_FAILURE;
			if(ctx->flags&4)return 0;
			memcpy(ctx->h2in.bfr+ctx->h2in.bfrfill,data,len);
			ctx->h2in.bfrfill+=len;
			for(i=0;i<ctx->h2in.bfrfill;i++)
				if(ctx->h2in.bfr[i]=='\r'||
					ctx->h2in.bfr[i]=='\n')
			{
				if(!i)return NGHTTP2_ERR_CALLBACK_FAILURE;
				ctx->h2in.bfr[i]=0;
				ctx->flags|=4;
			}
			return 0;
		}
		if(len)
		{
			if(len>ctx->h2in.total-ctx->h2in.bfrfill)
				return NGHTTP2_ERR_PAUSE;
			memcpy(ctx->h2in.bfr+ctx->h2in.bfrfill,data,len);
			ctx->h2in.bfrfill+=len;
		}
	}
	return 0;
}

static int on_stream_close_callback(nghttp2_session *session,int32_t stream_id,
	uint32_t error_code,void *user_data)
{
	H2CTX *ctx=user_data;

	if(stream_id==ctx->id)
	{
		ctx->flags|=0x10;
		if(nghttp2_session_terminate_session(session,NGHTTP2_NO_ERROR))
			return NGHTTP2_ERR_CALLBACK_FAILURE;
	}
	return 0;
}

static void hdrset(nghttp2_nv *hdr,char *name,char *value)
{
	hdr->name=(void *)name;
	hdr->value=(void *)value;
	hdr->namelen=strlen(name);
	hdr->valuelen=strlen(value);
	hdr->flags=NGHTTP2_NV_FLAG_NONE;
}

static ssize_t data_source_read_callback(nghttp2_session *session,
	int32_t stream_id,uint8_t *buf,size_t length,uint32_t *data_flags,
	nghttp2_data_source *source,void *user_data)
{
	int len;
	H2DATA *h2out=source->ptr;

	len=(length<h2out->total?length:h2out->total);
	if(!len)return 0;
	memcpy(buf,h2out->bfr,len);
	if((h2out->bfrfill-=len))memmove(h2out->bfr,h2out->bfr+len,
		h2out->bfrfill);
	if(!(h2out->total-=len))*data_flags|=
		NGHTTP2_DATA_FLAG_EOF|NGHTTP2_DATA_FLAG_NO_END_STREAM;
	return len;
}

static ssize_t dummy_source_read_callback(nghttp2_session *session,
	int32_t stream_id,uint8_t *buf,size_t length,uint32_t *data_flags,
	nghttp2_data_source *source,void *user_data)
{
	*data_flags|=NGHTTP2_DATA_FLAG_EOF|NGHTTP2_DATA_FLAG_NO_END_STREAM;
	return 0;
}

static void *http2_connect(int fd,void *tlsctx,char *host,int port,
	struct timespec *start,int timeout,char *tlshost,char *page)
{
	int r;
	int w;
	int len;
	int pos=0;
	int fill=0;
	int n=0;
	int tmo;
	nghttp2_session_callbacks *cb;
	H2CTX *ctx;
	struct pollfd p;
	nghttp2_nv hdrs[6];
	nghttp2_data_provider prd;
	unsigned char bfr[16384];
	char hp[260];

	prd.read_callback=dummy_source_read_callback;

	if(!(ctx=malloc(sizeof(H2CTX))))goto err1;
	memset(ctx,0,sizeof(H2CTX));
	p.fd=fd;
	ctx->tlsctx=tlsctx;
	ctx->fd=fd;
	snprintf(hp,sizeof(hp),"%s:%u",host,port);
	hdrset(&hdrs[0],":method","CONNECT");
	hdrset(&hdrs[1],":authority",hp);
	hdrset(&hdrs[2],":method","GET");
	hdrset(&hdrs[3],":scheme","https");
	hdrset(&hdrs[4],":authority",tlshost);
	hdrset(&hdrs[5],":path",page?page:"/");

	if(nghttp2_session_callbacks_new(&cb))goto err1;
	nghttp2_session_callbacks_set_send_callback(cb,send_callback);
	nghttp2_session_callbacks_set_on_frame_recv_callback(cb,
		on_frame_recv_callback);
	nghttp2_session_callbacks_set_on_data_chunk_recv_callback(cb,
		on_data_chunk_recv_callback);
	nghttp2_session_callbacks_set_on_stream_close_callback(cb,
		on_stream_close_callback);
	nghttp2_session_callbacks_set_on_header_callback(cb,on_header_callback);
	if(nghttp2_session_client_new(&ctx->sess,cb,ctx))
	{
		nghttp2_session_callbacks_del(cb);
		goto err2;
	}
	nghttp2_session_callbacks_del(cb);

	if(nghttp2_submit_settings(ctx->sess,NGHTTP2_FLAG_NONE,settings,2))
		goto err2;
	if(page)if(nghttp2_submit_request(ctx->sess,NULL,&hdrs[2],4,NULL,&ctx)
		<0)goto err3;
	if((ctx->id=nghttp2_submit_request(ctx->sess,NULL,hdrs,2,&prd,&ctx))<0)
		goto err3;

	while(1)
	{
repeat:		p.events=0;
		if((r=nghttp2_session_want_read(ctx->sess)))p.events|=POLLIN;
		if((w=nghttp2_session_want_write(ctx->sess)))p.events|=POLLOUT;
		if(fill)p.events&=~POLLIN;

		if(!r&&!w)goto err3;

		if(!(tmo=tmocalc(start,timeout)))goto err3;
		tmo*=1000;

		if(poll(&p,1,fill?100:tmo)<1)
		{
			if(fill)
			{
				len=nghttp2_session_mem_recv(ctx->sess,
					bfr+pos,fill);
				if(len<0)goto err3;
				if(len==fill)n=pos=fill=0;
				else
				{
					pos+=len;
					fill-=len;
				}
				if((n+=100)>=tmo)goto err3;
				goto repeat;
			}
			else goto err3;
		}
		n=0;

		if(p.revents&(POLLERR|POLLHUP))goto err3;

		if(p.revents&POLLIN)
		{
			fill=tls_client_read(tlsctx,bfr,sizeof(bfr));
			switch(fill)
			{
			case -1:if(errno!=EAGAIN)goto err3;
				fill=0;
			case 0: break;
			default:len=nghttp2_session_mem_recv(ctx->sess,
					bfr,fill);
				if(len<0)goto err3;
				if(len==fill)fill=0;
				else
				{
					pos=len;
					fill-=len;
				}
				break;
			}
		}

		if(p.revents&POLLOUT)
			if(w)if(nghttp2_session_send(ctx->sess))goto err3;

		if(ctx->flags&1)
		{
			if((ctx->flags&3)!=3)goto err3;
			return ctx;
		}
	}

err3:	nghttp2_session_del(ctx->sess);
err2:	free(ctx);
err1:	return NULL;
}

void http2_disconnect(void *context)
{
	H2CTX *ctx=context;

	nghttp2_session_del(ctx->sess);
	free(ctx);
}

static int http2_getaddr(char *host,int port,CTX *ctx,struct config *c,
	char *buf,int size,struct timespec *start)
{
	int r;
	int w;
	int len;
	int pos=0;
	int fill=0;
	int res=0;
	int n=0;
	int tmo;
	nghttp2_session *sess;
	nghttp2_session_callbacks *cb;
	H2CTX h2;
	struct pollfd p;
	nghttp2_nv hdrs[5];
	unsigned char bfr[16384];
	char hp[260];

	memset(&h2,0,sizeof(H2CTX));
	p.fd=ctx->fd;
	h2.tlsctx=ctx->tlsconn;
	h2.h2in.bfr=buf;
	h2.h2in.total=size;
	h2.h2in.bfrfill=0;
	h2.flags=0x08;

	if(port!=443)
	{
		snprintf(hp,sizeof(hp),
			"%s:%u",c->rservername?c->rservername:host,port);
		host=hp;
	}
	hdrset(&hdrs[0],":method","GET");
	hdrset(&hdrs[1],":scheme","https");
	hdrset(&hdrs[2],":authority",host);
	hdrset(&hdrs[3],":path",c->rurl);
	if(c->ragent)hdrset(&hdrs[4],"user-agent",c->ragent);

	if(nghttp2_session_callbacks_new(&cb))goto err1;
	nghttp2_session_callbacks_set_send_callback(cb,send_callback);
	nghttp2_session_callbacks_set_on_frame_recv_callback(cb,
		on_frame_recv_callback);
	nghttp2_session_callbacks_set_on_data_chunk_recv_callback(cb,
		on_data_chunk_recv_callback);
	nghttp2_session_callbacks_set_on_stream_close_callback(cb,
		on_stream_close_callback);
	nghttp2_session_callbacks_set_on_header_callback(cb,on_header_callback);
	if(nghttp2_session_client_new(&sess,cb,&h2))
	{
		nghttp2_session_callbacks_del(cb);
		goto err1;
	}
	nghttp2_session_callbacks_del(cb);

	if(nghttp2_submit_settings(sess,NGHTTP2_FLAG_NONE,settings,2))goto err2;
	if((h2.id=nghttp2_submit_request(sess,NULL,hdrs,c->ragent?5:4,NULL,&
		h2))<0)goto err2;

	while(1)
	{
repeat:		p.events=0;
		if((r=nghttp2_session_want_read(sess)))p.events|=POLLIN;
		if((w=nghttp2_session_want_write(sess)))p.events|=POLLOUT;
		if(fill)p.events&=~POLLIN;

		if(!r&&!w)break;

		if(!(tmo=tmocalc(start,c->conntimeout)))goto err2;
		tmo*=1000;

		if(poll(&p,1,fill?100:tmo)<1)
		{
			if(fill)
			{
				len=nghttp2_session_mem_recv(sess,bfr+pos,fill);
				if(len<0)goto err2;
				if(len==fill)n=pos=fill=0;
				else
				{
					pos+=len;
					fill-=len;
				}
				if((n+=100)>=tmo)goto err2;
				goto repeat;
			}
			else goto err2;
		}
		n=0;

		if(p.revents&(POLLERR|POLLHUP))goto err2;

		if(p.revents&POLLIN)
		{
			fill=tls_client_read(h2.tlsctx,bfr,sizeof(bfr));
			switch(fill)
			{
			case -1:if(errno!=EAGAIN)goto err2;
				fill=0;
			case 0: break;
			default:len=nghttp2_session_mem_recv(sess,bfr,fill);
				if(len<0)goto err2;
				if(len==fill)fill=0;
				else
				{
					pos=len;
					fill-=len;
				}
				break;
			}
		}

		if(p.revents&POLLOUT)
			if(w)if(nghttp2_session_send(sess))goto err2;
	}

	if((h2.flags&0xf)!=0xf)goto err2;

	res=0;

err2:	nghttp2_session_del(sess);
err1:	tls_client_disconnect(ctx->tlsconn,NULL);
	tls_client_fini(ctx->tlscfg);
	return res;
}

static void http2_forward(int lcl,H2CTX *ctx,int h2ping)
{
	int len;
	int lev;
	int rev;
	int state;
	int rh2fill;
	int h2rfill;
	int h2rlen;
	uint64_t dummy;
	const unsigned char *h2rptr;
	struct pollfd p[3];
	nghttp2_data_provider prd;
	struct itimerspec it;
	char lclh2[BUFSIZE];
	char rmth2[BUFSIZE];
	char h2lcl[BUFSIZE];
	char h2rmt[BUFSIZE];

	p[0].fd=lcl;
	p[1].fd=ctx->fd;
	if(!h2ping)p[2].fd=-1;
	else if((p[2].fd=timerfd_create(CLOCK_MONOTONIC,
		TFD_NONBLOCK|TFD_CLOEXEC))==-1)h2ping=0;
	p[2].events=(h2ping?POLLIN:0);
	p[2].revents=0;

	state=0;
	lev=POLLIN;
	rev=POLLIN;
	h2rlen=0;
	h2rptr=NULL;
	rh2fill=0;
	h2rfill=0;
	ctx->h2in.bfrfill=0;
	ctx->h2in.total=BUFSIZE;
	ctx->h2in.bfr=h2lcl;
	ctx->h2out.bfrfill=0;
	ctx->h2out.total=0;
	ctx->h2out.bfr=lclh2;
	prd.source.ptr=&ctx->h2out;
	prd.read_callback=data_source_read_callback;
	it.it_interval.tv_sec=0;
	it.it_interval.tv_nsec=0;
	it.it_value.tv_sec=h2ping;
	it.it_value.tv_nsec=0;

	if(h2ping)timerfd_settime(p[2].fd,0,&it,NULL);

	while(lev||rev)
	{
		if((state&9)==9)break;

		p[0].events=lev;
		p[1].events=rev;

		while(poll(p,3,-1)<1);

		if(p[2].revents&POLLIN)
			if(read(p[2].fd,&dummy,sizeof(dummy))==sizeof(dummy))
				if(!(state&2)&&!h2rfill&&!p[0].revents&POLLIN)
		{
			if(nghttp2_submit_ping(ctx->sess,NGHTTP2_FLAG_NONE,
				NULL))break;
			timerfd_settime(p[2].fd,0,&it,NULL);
		}

		if(p[0].revents&(POLLERR|POLLHUP|POLLNVAL))break;

		if(p[0].revents&POLLIN)
		{
			if((len=read(lcl,lclh2+ctx->h2out.bfrfill,
				BUFSIZE-ctx->h2out.bfrfill))<=0)state|=1;
			else ctx->h2out.bfrfill+=len;
		}

		if(!(state&2))
		{
			if(h2rptr)
			{
				len=BUFSIZE-h2rfill;
				if(h2rlen<len)len=h2rlen;
				memcpy(h2rmt+h2rfill,h2rptr,len);
				h2rfill+=len;
				if(!(h2rlen-=len))h2rptr=NULL;
				else h2rptr+=len;
			}
			if(!h2rptr)
			{
				if((h2rlen=nghttp2_session_mem_send(ctx->sess,
					&h2rptr))<0)break;
				if(!h2rlen)h2rptr=NULL;
				else
				{
					len=BUFSIZE-h2rfill;
					if(h2rlen<len)len=h2rlen;
					memcpy(h2rmt+h2rfill,h2rptr,len);
					h2rfill+=len;
					if(!(h2rlen-=len))h2rptr=NULL;
					else h2rptr+=len;
				}
			}
			if(!h2rptr&&ctx->h2out.bfrfill&&!ctx->h2out.total)
			{
				ctx->h2out.total=ctx->h2out.bfrfill;
				if(nghttp2_submit_data(ctx->sess,0,ctx->id,
					&prd))break;
				if((h2rlen=nghttp2_session_mem_send(ctx->sess,
					&h2rptr))<0)break;
				if(!h2rlen)h2rptr=NULL;
				else
				{
					len=BUFSIZE-h2rfill;
					if(h2rlen<len)len=h2rlen;
					memcpy(h2rmt+h2rfill,h2rptr,len);
					h2rfill+=len;
					if(!(h2rlen-=len))h2rptr=NULL;
					else h2rptr+=len;
				}
			}
		}

		if(p[1].revents&(POLLERR|POLLHUP|POLLNVAL))break;

		if(p[1].revents&POLLIN)
		{
			if((len=tls_client_read(ctx->tlsctx,rmth2+rh2fill,
				BUFSIZE-rh2fill))>0)rh2fill+=len;
			else if(len!=-1||errno!=EAGAIN)state|=4;
		}

		if(!(state&8)&&rh2fill)
		{
			if((len=nghttp2_session_mem_recv(ctx->sess,
				(void *)rmth2,rh2fill))<0)break;
			if(ctx->flags&0x10)break;
			rh2fill-=len;
			if(rh2fill)memmove(rmth2,rmth2+len,rh2fill);
		}

		if(h2rfill&&(p[1].revents&POLLOUT))
		{
			if((len=tls_client_write(ctx->tlsctx,(void *)h2rmt,
				h2rfill))>=0)
			{
				if((h2rfill-=len))memmove(h2rmt,h2rmt+len,
					h2rfill);
				else if(h2ping)
				{
					timerfd_settime(p[2].fd,0,&it,NULL);
					len=read(p[2].fd,&dummy,sizeof(dummy));
				}
			}
			else if(len!=-1||errno!=EAGAIN)
			{
				h2rfill=0;
				shutdown(p[0].fd,SHUT_RD);
				state|=2;
			}
		}

		if(ctx->h2in.bfrfill&&(p[0].revents&POLLOUT))
		{
			if((len=write(lcl,h2lcl,ctx->h2in.bfrfill))>=0)
			{
				if((ctx->h2in.bfrfill-=len))memmove(h2lcl,
					h2lcl+len,ctx->h2in.bfrfill);
			}
			else
			{
				ctx->h2in.bfrfill=0;
				shutdown(p[1].fd,SHUT_RD);
				state|=8;
			}
		}

		if(state&1)lev=0;
		else if(ctx->h2out.bfrfill==BUFSIZE)lev=0;
		else lev=POLLIN;
		if(state&2)
		{
			h2rfill=0;
			ctx->h2out.bfrfill=0;
		}

		if(state&4)rev=0;
		else if(rh2fill==BUFSIZE)rev=0;
		else rev=POLLIN;
		if(state&8)
		{
			rh2fill=0;
			ctx->h2in.bfrfill=0;
		}

		if(rh2fill||ctx->h2in.bfrfill)lev|=POLLOUT;
		if(h2rfill||ctx->h2out.bfrfill)rev|=POLLOUT;
	}

	if(p[2].fd!=-1)close(p[2].fd);
}

#endif

static int http1_connect(int fd,void *tlsctx,char *host,int port,
	struct timespec *start,int timeout)
{
	int len;
	int l;
	int pos;
	int tmo;
	struct pollfd p;
	char bfr[1024];

	len=snprintf(bfr,sizeof(bfr),
		"CONNECT %s:%d HTTP/1.1\r\nHost: %s:%d\r\n\r\n",
		host,port,host,port);

	if(tls_client_write(tlsctx,bfr,len)!=len)goto err1;

	p.fd=fd;
	p.events=POLLIN;

	if(!(tmo=tmocalc(start,timeout)))goto err1;
	tmo*=1000;

	for(len=0;len<sizeof(bfr);)
	{
		if(poll(&p,1,tmo)<1)goto err1;
		if(!(p.revents&POLLIN))goto err1;

		if((l=tls_client_read(tlsctx,bfr+len,sizeof(bfr)-len))<0)
		{
			if(errno==EAGAIN)l=0;
			else goto err1;
		}
		len+=l;

		for(pos=0,l=0;pos<len;pos++)
		{
			if(bfr[pos]=='\r')continue;
			else if(bfr[pos]=='\n')l++;
			else l=0;
			if(l==2)
			{
				if(len<12)goto err1;
				if(strncmp(bfr,"HTTP/1.0 200",12)&&
					strncmp(bfr,"HTTP/1.1 200",12))
						goto err1;
				return 0;
			}
		}

		if(!(tmo=tmocalc(start,timeout)))goto err1;
		tmo*=1000;
	}

err1:	return -1;
}

static int http1_getaddr(char *host,int port,CTX *ctx,struct config *c,
	char *buf,int size,struct timespec *start)
{
	int l;
	int len;
	int pos;
	int tmo;
	struct pollfd p;
	char bfr[1024];

	len=snprintf(bfr,sizeof(bfr),"GET %s HTTP/1.0\r\nHost: %s\r\n",c->rurl,
		c->rservername?c->rservername:host);
	if(c->ragent)len+=snprintf(bfr+len,sizeof(bfr)-len,
		"User-Agent: %s\r\n",c->ragent);
	len+=snprintf(bfr+len,sizeof(bfr)-len,"\r\n");

	if(tls_client_write(ctx->tlsconn,bfr,len)!=len)goto err1;

	p.fd=ctx->fd;
	p.events=POLLIN;

	for(len=0;len<sizeof(bfr);)
	{
		if(!(tmo=tmocalc(start,c->conntimeout)))goto err1;
		tmo*=1000;
		if(poll(&p,1,tmo)<1)goto err1;
		if(!(p.revents&POLLIN))goto err1;
		if((l=tls_client_read(ctx->tlsconn,bfr+len,sizeof(bfr)-len))<0)
		{
			if(errno==EAGAIN)l=0;
			else goto err1;
		}
		len+=l;

		for(pos=0,l=0;pos<len;pos++)
		{
			if(bfr[pos]=='\r')continue;
			else if(bfr[pos]=='\n')l++;
			else l=0;
			if(l==2)
			{
				if(len<12)goto err1;
				if(strncmp(bfr,"HTTP/1.0 200",12)&&
					strncmp(bfr,"HTTP/1.1 200",12))
						goto err1;
				goto body;
			}
		}
	}

	goto err1;

body:	len-=pos+1;
	if(len)memmove(bfr,bfr+pos+1,len);

	for(l=len;len<sizeof(bfr);l=0)
	{
		if(!l)
		{
			if(!(tmo=tmocalc(start,c->conntimeout)))goto err1;
			tmo*=1000;
			if(poll(&p,1,tmo)<1)goto err1;
			if(!(p.revents&POLLIN))goto err1;
			if((l=tls_client_read(ctx->tlsconn,bfr+len,
				sizeof(bfr)-len))<0)
			{
				if(errno==EAGAIN)l=0;
				else goto err1;
			}
			len+=l;
		}

		for(pos=0;pos<len;pos++)
		{
			if(bfr[pos]=='\r'||bfr[pos]=='\n')
			{
				bfr[pos]=0;
				if(!pos||pos>=size)goto err1;
				strcpy(buf,bfr);
				tls_client_disconnect(ctx->tlsconn,NULL);
				tls_client_fini(ctx->tlscfg);
				return 0;
			}
		}
	}

err1:	tls_client_disconnect(ctx->tlsconn,NULL);
	tls_client_fini(ctx->tlscfg);
	return -1;
}

static void http1_forward(int lcl,CTX *ctx)
{
	int len;
	int lrfill;
	int rlfill;
	int lev;
	int rev;
	int state;
	struct pollfd p[2];
	char lclrmt[BUFSIZE];
	char rmtlcl[BUFSIZE];

	p[0].fd=lcl;
	p[1].fd=ctx->fd;

	state=0;
	lev=POLLIN;
	rev=POLLIN;
	lrfill=0;
	rlfill=0;

	while(lev||rev)
	{
		p[0].events=lev;
		p[1].events=rev;

		if((state&9)==9)break;

		while(poll(p,2,-1)<1);

		if(p[0].revents&(POLLERR|POLLHUP|POLLNVAL))break;

		if(p[0].revents&POLLIN)
		{
			if((len=read(lcl,lclrmt+lrfill,BUFSIZE-lrfill))<=0)
				state|=1;
			else lrfill+=len;
		}

		if(p[1].revents&(POLLERR|POLLHUP|POLLNVAL))break;

		if(p[1].revents&POLLIN)
		{
			if((len=tls_client_read(ctx->tlsconn,rmtlcl+rlfill,
				BUFSIZE-rlfill))>0)rlfill+=len;
			else if(len!=-1||errno!=EAGAIN)state|=4;
		}

		if(lrfill&&(p[1].revents&POLLOUT))
		{
			if((len=tls_client_write(ctx->tlsconn,lclrmt,lrfill))
				>=0)
			{
				lrfill-=len;
				if(len)memmove(lclrmt,lclrmt+len,lrfill);
			}
			else if(len!=-1||errno!=EAGAIN)
			{
				shutdown(p[0].fd,SHUT_RD);
				state|=2;
			}
		}

		if(rlfill&&(p[0].revents&POLLOUT))
		{
			if((len=write(lcl,rmtlcl,rlfill))>=0)
			{
				rlfill-=len;
				if(len)memmove(rmtlcl,rmtlcl+len,rlfill);
			}
			else
			{
				shutdown(p[1].fd,SHUT_RD);
				state|=8;
			}
		}

		if(state&1)lev=0;
		else if(lrfill==BUFSIZE)lev=0;
		else lev=POLLIN;
		if(state&2)lrfill=0;

		if(state&4)rev=0;
		else if(rlfill==BUFSIZE)rev=0;
		else rev=POLLIN;
		if(state&8)rlfill=0;

		if(rlfill)lev|=POLLOUT;
		if(lrfill)rev|=POLLOUT;
	}
}

static int selectproto(struct config *c,CTX *ctx)
{
	int i;
	char *sel;

	if(!c->usealpn&&!c->emutype)return 0;
	if(!(sel=tls_client_get_alpn(ctx->tlsconn)))return -1;
	for(i=0;i<ALPNENTRIES;i++)if(!strcmp(sel,alpn[i]))break;
	if(i==ALPNENTRIES)return -1;
	if(ALPNENTRIES==2&&!i)return 1;
	return 0;
}

static int parseline(char *line,int num,void *base)
{       
	int i;
	unsigned long number;
	char *mem;
	char *item;
	char *value;
	char *ptr;
	
	item=strtok_r(line," \t\r\n",&mem);
	value=strtok_r(NULL," \t\r\n",&mem);
	
	if(!item||!value||!*item||!*value)goto err1;
	for(i=0;items[i].name;i++)if(!strcmp(item,items[i].name))break;
	if(!items[i].name)goto err1;
	
	switch(items[i].type)
	{
	case STRING:
		if(!*value||strlen(value)>255)goto err1;
		if(!(ptr=strdup(value)))goto err2;
		*((char **)(base+items[i].offset))=ptr;
		break;
	
	case NUMBER:
		if((number=strtoul(value,&mem,10))==ULONG_MAX||*mem)goto err1;
		if(number<items[i].min||number>items[i].max)goto err1;
		*((unsigned int *)(base+items[i].offset))=(unsigned int)number;
		break;

	case BOOLEAN:
		if(!strcmp(value,"enable"))
			*((unsigned int *)(base+items[i].offset))=1;
		else if(!strcmp(value,"disable"))
			*((unsigned int *)(base+items[i].offset))=0;
		else goto err1;
		break;

	default:goto err1;
	}

	return 0;

err2:   fprintf(stderr,"internal error\n");
	return -1;

err1:   fprintf(stderr,"configuration syntax error line %d\n",num);
	return -1;
}

static int parseconfig(char *fn,struct config *cfg)
{
	int num=0;
	FILE *fp;
	char *ptr;
	char line[8192];

        if(!(fp=fopen(fn,"r")))
        {
                fprintf(stderr,"can't open %s\n",fn);
                goto err1;
        }

        while(fgets(line,sizeof(line),fp))
        {
                num++;
                for(ptr=line;*ptr;ptr++)if(*ptr!=' '||*ptr!='\t')break;
                if(!*ptr||*ptr=='#'||*ptr=='\r'||*ptr=='\n')continue;
		if(parseline(ptr,num,cfg))goto err2;
	}

	if(cfg->redirect)
	{
		if(!cfg->rurl||strlen(cfg->rurl)>512||cfg->rurl[0]!='/')
			goto err3;
		if(cfg->ragent&&strlen(cfg->ragent)>256)goto err3;
		if(cfg->rverifychain&&!cfg->rcafile)goto err3;
	}

	if(cfg->verifychain&&!cfg->cafile)goto err3;

	if(cfg->proxytype||cfg->proxyhost||cfg->proxyport)
	{
		if(!cfg->proxytype||!cfg->proxyhost||!cfg->proxyport)goto err3;
		if(strcmp(cfg->proxytype,"http")&&strcmp(cfg->proxytype,"https")
			&&strcmp(cfg->proxytype,"socks4")
			&&strcmp(cfg->proxytype,"socks4a")
			&&strcmp(cfg->proxytype,"socks5")
			&&strcmp(cfg->proxytype,"socks5h"))goto err3;
	}

	fclose(fp);

	return 0;

err3:	fprintf(stderr,"configuration syntax error\n");
err2:	fclose(fp);
err1:	return -1;
}

int main(int argc,char *argv[])
{
	int i;
	int mode;
	int port;
	int hport;
	int emu=-1;
	CTX ctx;
#ifdef HTTP2
	void *h2;
#endif
	char *host;
	char *ptr;
	struct config cfg;
	struct timespec start;
	char bfr[256];

	if(!(host=getenv("VPN_PORT")))goto err1;
	port=atoi(host);
	if(port<1||port>65535)goto err1;
	if(!(host=getenv("VPN_HOST")))goto err1;
	if(!*host||strlen(host)>255)goto err1;
	memset(&cfg,0,sizeof(cfg));
	if(argc==2)
	{
		if(parseconfig(argv[1],&cfg))goto err1;
	}
	else if(argc!=1)goto err1;
	if(!cfg.conntimeout)cfg.conntimeout=10;

	if(cfg.emutype)
	{
		for(i=0;emulist[i].name;i++)
			if(!strcmp(emulist[i].name,cfg.emutype))break;
		if(!emulist[i].name)goto err1;
		emu=emulist[i].emu;
	}

	if(cfg.emupath)
	{
		snprintf(bfr,sizeof(bfr)-16,"%s",cfg.emupath);
		if(!(ptr=getenv("LD_LIBRARY_PATH"))||strcmp(ptr,bfr))
		{
			snprintf(bfr,sizeof(bfr),"LD_LIBRARY_PATH=%s",
				cfg.emupath);
			putenv(bfr);
			return execv(argv[0],argv);
		}
	}

	close(1);
	if(fcntl(0,F_SETFL,O_RDWR|O_NONBLOCK))goto err1;

	if((ptr=strchr(host,']')))
	{
		if(*host++!='[')goto err1;
		*ptr++=0;
		ptr=strchr(ptr,':');
	}
	else ptr=strchr(host,':');

	if(!ptr)hport=443;
	else
	{
		*ptr++=0;
		if((hport=atoi(ptr))<1||hport>65535)goto err1;
	}

	if(clock_gettime(CLOCK_MONOTONIC,&start))goto err1;

	if(tls_client_global_init())goto err1;
	if(cfg.redirect)
	{
		if(remoteconnect(host,hport,NULL,0,0,&cfg,&ctx,0,&start,emu))
			goto err2;
		switch(selectproto(&cfg,&ctx))
		{
		case 0:	if(http1_getaddr(host,port,&ctx,&cfg,bfr,sizeof(bfr),
				&start))goto err2;
			break;
#ifdef HTTP2
		case 1:	if(http2_getaddr(host,port,&ctx,&cfg,bfr,sizeof(bfr),
				&start))goto err2;
			break;
#endif
		case -1:goto err2;
		}

		if(!*(host=bfr)||strlen(host)>255)goto err2;

		if((ptr=strchr(host,']')))
		{
			if(*host++!='[')goto err2;
			*ptr++=0;
			ptr=strchr(ptr,':');
		}
		else ptr=strchr(host,':');

		if(!ptr)hport=443;
		else
		{
			*ptr++=0;
			if((hport=atoi(ptr))<1||hport>65535)goto err2;
		}
	}

	if(remoteconnect(host,hport,"localhost",port,0,&cfg,&ctx,1,&start,emu))
		goto err2;
	switch((mode=selectproto(&cfg,&ctx)))
	{
	case 0:	if(http1_connect(ctx.fd,ctx.tlsconn,"localhost",port,&start,
			cfg.conntimeout))goto err3;
		break;
#ifdef HTTP2
	case 1:	if(!(h2=http2_connect(ctx.fd,ctx.tlsconn,"localhost",port,
			&start,cfg.conntimeout,host,cfg.h2page)))goto err3;
		break;
#endif
	case -1:goto err3;
	}

	if(write(0,"OK\n",3)!=3)goto err3;

	switch(mode)
	{
	case 0:	http1_forward(0,&ctx);
		break;
#ifdef HTTP2
	case 1:	http2_forward(0,h2,cfg.h2ping);
		http2_disconnect(h2);
		break;
#endif
	}

	tls_client_disconnect(ctx.tlsconn,NULL);
	tls_client_fini(ctx.tlscfg);
	tls_client_global_fini();
	close(0);
	return 0;

err3:	tls_client_disconnect(ctx.tlsconn,NULL);
	tls_client_fini(ctx.tlscfg);
err2:	tls_client_global_fini();
err1:	port=write(0,"NO\n",3);
	return 1;
}
