/*
 * This file is part of the vpnd2 project
 *
 * (C) 2020 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <ifaddrs.h>
#include <poll.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include "network.h"

static int setkeepalive(int s,int idle)
{
	int x;

	x=3;
	if(setsockopt(s,IPPROTO_TCP,TCP_KEEPCNT,&x,sizeof(x)))
	{
		perror("setsockopt IPPROTO_TCP TCP_KEEPCNT");
		return -1;
	}

	x=idle;
	if(setsockopt(s,IPPROTO_TCP,TCP_KEEPIDLE,&x,sizeof(x)))
	{
		perror("setsockopt IPPROTO_TCP TCP_KEEPIDLE");
		return -1;
	}

	x=idle;
	if(setsockopt(s,IPPROTO_TCP,TCP_KEEPINTVL,&x,sizeof(x)))
	{
		perror("setsockopt IPPROTO_TCP TCP_KEEPINTVL");
		return -1;
	}

	x=1;
	if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&x,sizeof(x)))
	{
		perror("setsockopt SOL_SOCKET SO_KEEPALIVE");
		return -1;
	}

	return 0;
}

int tcp4connect(struct in_addr *src,struct in_addr *dst,int *port,int timeout,
	int efd,int flags,char *dev,int idle)
{
	int s;
	int x;
	int n=0;
	socklen_t y;
	struct linger l;
	struct sockaddr_in a;
	struct pollfd p[2];
	struct ifreq ifr;

	if(timeout!=-1)if(timeout<1||timeout>86400)goto e1;

	if(*port<1||*port>65535)goto e1;

	if((s=socket(AF_INET,
		SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,IPPROTO_TCP))==-1)
	{
		perror("socket AF_INET SOCK_STREAM");
		goto e1;
	}

	if(flags&KEEPALIVE)if(setkeepalive(s,idle))goto e2;

	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))
	{
		perror("setsockopt SOL_SOCKET SO_OOBINLINE");
		goto e2;
	}

	l.l_onoff=1;
	l.l_linger=10;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))
	{
		perror("setsockopt SOL_SOCKET SO_LINGER");
		goto e2;
	}

	x=1;
	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))
	{
		perror("setsockopt SOL_TCP TCP_NODELAY");
		goto e2;
	}

	if(flags&FAST_CONNECT)
	{
#ifdef TCP_FASTOPEN_CONNECT
		x=1;
		if(setsockopt(s,SOL_TCP,TCP_FASTOPEN_CONNECT,&x,sizeof(x)))
		{
			perror("setsockopt SOL_TCP TCP_FASTOPEN_CONNECT");
			goto e2;
		}
#endif
	}

	if(dev)
	{
		memset(&ifr,0,sizeof(ifr));
		ifr.ifr_addr.sa_family=AF_INET;
		strncpy(ifr.ifr_name,dev,sizeof(ifr.ifr_name)-1);
		ifr.ifr_name[sizeof(ifr.ifr_name)-1]=0;

		if(setsockopt(s,SOL_SOCKET,SO_BINDTODEVICE,&ifr,sizeof(ifr)))
		{
			perror("setsockopt SOL_SOCKET SO_BINDTODEVICE");
			goto e2;
		}
	}

	if(flags&SPECIFIC_ADDR)
	{
		memset(&a,0,sizeof(a));
		a.sin_family=AF_INET;
		a.sin_addr.s_addr=src->s_addr;

		if(bind(s,(struct sockaddr *)(&a),sizeof(a)))
		{
			perror("bind AF_INET SOCK_STREAM");
			goto e2;
		}
	}

	memset(&a,0,sizeof(a));
	a.sin_family=AF_INET;
	a.sin_addr.s_addr=dst->s_addr;
	a.sin_port=htons(*port);

	if(connect(s,(struct sockaddr *)(&a),sizeof(a)))
	{
		if(errno==EINPROGRESS)
		{
			p[0].fd=s;
			p[0].events=POLLOUT;
			p[1].fd=efd;
			p[1].events=POLLIN;

			if(timeout!=-1)timeout*=1000;

			switch(poll(p,efd!=-1?2:1,timeout))
			{
			case -1:perror("poll");
			case 0:	goto e2;
			}

			if(efd!=-1)if(p[1].revents&POLLIN)
			{
				n++;
				goto e2;
			}

			y=sizeof(x);
			if(getsockopt(s,SOL_SOCKET,SO_ERROR,&x,&y))
			{
				perror("getsockopt SOL_SOCKET SO_ERROR");
				goto e2;
			}

			if(x)
			{
				errno=x;
				goto e2;
			}
		}
		else goto e2;
	}

	y=sizeof(a);
	if(getsockname(s,(struct sockaddr *)(&a),&y))
	{
		perror("getsockname AF_INET SOCK_STREAM");
		goto e2;
	}

	if(a.sin_family!=AF_INET||y!=sizeof(a))
	{
e2:		x=errno;
		close(s);
		errno=x;
e1:		return -1-n;
	}

	if(src)src->s_addr=a.sin_addr.s_addr;
	*port=ntohs(a.sin_port);

	return s;
}

int tcp6connect(struct in6_addr *src,struct in6_addr *dst,int *port,int timeout,
	int efd,int flags,char *dev,int idle)
{
	int s;
	int x;
	int n=0;
	socklen_t y;
	struct linger l;
	struct sockaddr_in6 a;
	struct pollfd p[2];
	struct ifreq ifr;

	if(timeout!=-1)if(timeout<1||timeout>86400)goto e1;

	if(*port<1||*port>65535)goto e1;

	if((s=socket(AF_INET6,
		SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,IPPROTO_TCP))==-1)
	{
		perror("socket AF_INET6 SOCK_STREAM");
		goto e1;
	}

	if(flags&KEEPALIVE)if(setkeepalive(s,idle))goto e2;

	x=0;
	if(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x)))
	{
		perror("setsockopt SOL_SOCKET SO_OOBINLINE");
		goto e2;
	}

	l.l_onoff=1;
	l.l_linger=10;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))
	{
		perror("setsockopt SOL_SOCKET SO_LINGER");
		goto e2;
	}

	x=1;
	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))
	{
		perror("setsockopt SOL_TCP TCP_NODELAY");
		goto e2;
	}

	if(flags&FAST_CONNECT)
	{
#ifdef TCP_FASTOPEN_CONNECT
		x=1;
		if(setsockopt(s,SOL_TCP,TCP_FASTOPEN_CONNECT,&x,sizeof(x)))
		{
			perror("setsockopt SOL_TCP TCP_FASTOPEN_CONNECT");
			goto e2;
		}
#endif
	}

	x=1;
	if(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
	{
		perror("setsockopt IPPROTO_IPV6 IPV6_V6ONLY");
		goto e2;
	}

	if(dev)
	{
		memset(&ifr,0,sizeof(ifr));
		ifr.ifr_addr.sa_family=AF_INET;
		strncpy(ifr.ifr_name,dev,sizeof(ifr.ifr_name)-1);
		ifr.ifr_name[sizeof(ifr.ifr_name)-1]=0;

		if(setsockopt(s,SOL_SOCKET,SO_BINDTODEVICE,&ifr,sizeof(ifr)))
		{
			perror("setsockopt SOL_SOCKET SO_BINDTODEVICE");
			goto e2;
		}
	}

	if(flags&SPECIFIC_ADDR)
	{
		memset(&a,0,sizeof(a));
		a.sin6_family=AF_INET6;
		a.sin6_addr=*src;

		if(bind(s,(struct sockaddr *)(&a),sizeof(a)))
		{
			perror("bind AF_INET6 SOCK_STREAM");
			goto e2;
		}
	}

	memset(&a,0,sizeof(a));
	a.sin6_family=AF_INET6;
	a.sin6_addr=*dst;
	a.sin6_port=htons(*port);

	if(connect(s,(struct sockaddr *)(&a),sizeof(a)))
	{
		if(errno==EINPROGRESS)
		{
			p[0].fd=s;
			p[0].events=POLLOUT;
			p[1].fd=efd;
			p[1].events=POLLIN;

			if(timeout!=-1)timeout*=1000;

			switch(poll(p,efd!=-1?2:1,timeout))
			{
			case -1:perror("poll");
			case 0:	goto e2;
			}

			if(efd!=-1)if(p[1].revents&POLLIN)
			{
				n++;
				goto e2;
			}

			y=sizeof(x);
			if(getsockopt(s,SOL_SOCKET,SO_ERROR,&x,&y))
			{
				perror("getsockopt SOL_SOCKET SO_ERROR");
				goto e2;
			}

			if(x)
			{
				errno=x;
				goto e2;
			}
		}
		else goto e2;
	}

	y=sizeof(a);
	if(getsockname(s,(struct sockaddr *)(&a),&y))
	{
		perror("getsockname AF_INET6 SOCK_STREAM");
		goto e2;
	}

	if(a.sin6_family!=AF_INET6||y!=sizeof(a))
	{
e2:		x=errno;
		close(s);
		errno=x;
e1:		return -1-n;
	}

	if(src)*src=a.sin6_addr;
	*port=ntohs(a.sin6_port);

	return s;
}

int tcpconnect(struct sockaddr *src,struct sockaddr *dst,int timeout,int efd,
	int flags,char *dev,int idle)
{
	int s;
	int port;
	void *dest;
	void *source;

	if(dst->sa_family==AF_INET)
	{
		port=ntohs(((struct sockaddr_in *)dst)->sin_port);
		dest=&((struct sockaddr_in *)dst)->sin_addr;
		if(src)source=&((struct sockaddr_in *)src)->sin_addr;
		else source=NULL;
		if((s=tcp4connect(source,dest,&port,timeout,efd,flags,dev,
			idle))<0)return s;
		if(src)((struct sockaddr_in *)src)->sin_port=htons(port);
	}
	else
	{
		port=ntohs(((struct sockaddr_in6 *)dst)->sin6_port);
		dest=&((struct sockaddr_in6 *)dst)->sin6_addr;
		if(src)source=&((struct sockaddr_in6 *)src)->sin6_addr;
		else source=NULL;
		if((s=tcp6connect(source,dest,&port,timeout,efd,flags,dev,
			idle))<0)return s;
		if(src)((struct sockaddr_in6 *)src)->sin6_port=htons(port);
	}
	return s;
}
