/*
 * This file is part of the vpnd2 project
 *
 * (C) 2020 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <poll.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include "s5client.h"
#include "network.h"

static int readbytes(int s,char *bfr,int len)
{
	int rlen;
	int f;

	for(rlen=0;rlen<len;rlen+=f)if((f=read(s,bfr+rlen,len-rlen))<=0)
	{
		if(f==-1&&errno==EINTR)
		{
			f=0;
			continue;
		}
		perror("read");
		return -1;
	}

	return 0;
}

static int reply(int s,char *bfr)
{
	int len;

	if(readbytes(s,bfr,4))return -1;

	if(bfr[0]!=0x05||bfr[1]>8||bfr[2]!=0x00)
	{
		fprintf(stderr,"format error\n");
		return -1;
	}

	switch(bfr[3])
	{
	case 0x01:
		if(readbytes(s,bfr+4,4))return -1;
		len=8;
		break;

	case 0x03:
		if(readbytes(s,bfr+4,1))return -1;
		if(readbytes(s,bfr+5,bfr[4]))return -1;
		len=5+bfr[4];
		break;

	case 0x04:
		if(readbytes(s,bfr+4,16))return -1;
		len=20;
		break;

	default:fprintf(stderr,"format error\n");
		return -1;
	}

	if(readbytes(s,bfr+len,2))return -1;

	return len+2;
}

static int doauth(int s,int efd)
{
	struct pollfd p[2];
	unsigned char bfr[3];

	p[0].fd=s;
	p[0].events=POLLIN;
	p[1].fd=efd;
	p[1].events=(efd==-1?0:POLLIN);
	p[1].revents=0;

	bfr[0]=0x05;
	bfr[1]=0x01;
	bfr[2]=0x00;

	if(write(s,bfr,3)!=3)
	{
		perror("write");
		return -1;
	}

	if(poll(p,efd==-1?1:2,5000)<1||p[1].revents||!p[0].revents)
	{
		if(p[1].revents)
		{
			fprintf(stderr,"abort\n");
			return -2;
		}
		fprintf(stderr,"timeout\n");
		return -1;
	}

	if(readbytes(s,(char *)bfr,2))return -1;

	if(bfr[0]!=0x05||(bfr[1]&&bfr[1]!=0xff))
	{
		fprintf(stderr,"format error\n");
		return -1;
	}

	if(bfr[1])
	{
		fprintf(stderr,"request denied by server\n");
		return -1;
	}

	return 0;
}

int s5connect(struct sockaddr *server,struct sockaddr *dest,int port,char *host,
	int contimeout,int efd,int idle,int fast)
{
	int s;
	int hlen=0;
	int e;
	union
	{
		struct sockaddr_in *s4;
		struct sockaddr_in6 *s6;
	} a;
	struct pollfd p[2];
	unsigned char bfr[262];

	if(host)if((hlen=strlen(host))>255)
	{
		fprintf(stderr,"oversize host name\n");
		return -1;
	}

	if((s=tcpconnect(NULL,server,5,efd,
		(idle?KEEPALIVE:0)|(fast?FAST_CONNECT:0),NULL,idle))<0)
			return s;

	p[0].fd=s;
	p[0].events=POLLOUT;
	p[1].fd=efd;
	p[1].events=(efd==-1?0:POLLIN);
	p[1].revents=0;

	if((e=doauth(s,efd)))
	{
		close(s);
		return e;
	}

	a.s4=(struct sockaddr_in *)dest;

	if(host)
	{
		bfr[0]=0x05;
		bfr[1]=0x01;
		bfr[2]=0x00;
		bfr[3]=0x03;
		bfr[4]=(unsigned char)hlen;
		memcpy(bfr+5,host,hlen);
		bfr[5+hlen]=port>>8;
		bfr[6+hlen]=port&0xff;

		if(write(s,bfr,hlen+7)!=hlen+7)
		{
			perror("write");
			close(s);
			return -1;
		}
	}
	else if(a.s4->sin_family==AF_INET)
	{
		port=ntohs(a.s4->sin_port);
		bfr[0]=0x05;
		bfr[1]=0x01;
		bfr[2]=0x00;
		bfr[3]=0x01;
		memcpy(bfr+4,&a.s4->sin_addr,4);
		bfr[8]=port>>8;
		bfr[9]=port&0xff;

		if(write(s,bfr,10)!=10)
		{
			perror("write");
			close(s);
			return -1;
		}
	}
	else
	{
		port=ntohs(a.s6->sin6_port);
		bfr[0]=0x05;
		bfr[1]=0x01;
		bfr[2]=0x00;
		bfr[3]=0x04;
		memcpy(bfr+4,&a.s6->sin6_addr,16);
		bfr[20]=port>>8;
		bfr[21]=port&0xff;

		if(write(s,bfr,22)!=22)
		{
			perror("write");
			close(s);
			return -1;
		}
	}

	p[0].events=POLLIN;
	if(poll(p,efd==-1?1:2,contimeout*1000)<1||p[1].revents||!p[0].revents)
	{
		close(s);
		if(p[1].revents)
		{
			fprintf(stderr,"abort\n");
			return -2;
		}
		fprintf(stderr,"timeout\n");
		return -1;
	}

	if(reply(s,(char *)bfr)<0)
	{
		close(s);
		return -1;
	}

	if(bfr[1])
	{
		fprintf(stderr,"request denied by server\n");
		close(s);
		return -1;
	}

	switch(bfr[3])
	{
	case 0x01:
		a.s4->sin_family=AF_INET;
		memcpy(&a.s4->sin_addr,bfr+4,4);
		memcpy(&a.s4->sin_port,bfr+8,2);
		break;

	case 0x04:
		a.s6->sin6_family=AF_INET6;
		memcpy(&a.s6->sin6_addr,bfr+4,16);
		memcpy(&a.s6->sin6_port,bfr+20,2);
		break;

	default:fprintf(stderr,"format error\n");
		close(s);
		return -1;
	}

	return s;
}
