/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <pthread.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/eventfd.h>
#include <sys/timerfd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <signal.h>
#include <poll.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "misc.h"
#include "crypto.h"
#include "hash.h"
#include "udp.h"

#ifndef UDPLITE_SEND_CSCOV
#define UDPLITE_SEND_CSCOV	10
#endif

#ifndef UDPLITE_RECV_CSCOV
#define UDPLITE_RECV_CSCOV	11
#endif

#define MAXQUEUE		16

#pragma pack(push,1)

struct tuple4
{
	struct in_addr src;
	struct in_addr dst;
	unsigned short sport;
	unsigned short dport;
};

struct tuple6
{
	struct in6_addr src;
	struct in6_addr dst;
	unsigned short sport;
	unsigned short dport;
};

#pragma pack(pop)

struct list
{
	struct list *next;
	struct list *prev;
	struct list *left;
	struct list *right;
	unsigned long long stamp;
	int id;
	union
	{
		struct tuple4 t4;
		struct tuple6 t6;
	};
};

struct peer
{
	int fd;
	int rfd;
};

struct udpctx
{
	int fd;
	int efd;
	int tfd;
	int rfd;
	int pfd[2];
	int sendbuf;
	int lite;
	int alt;
	int frag;
	int rtp;
	int portmin;
	int portmax;
	int family;
	unsigned short port;
	pthread_t th;
	void *glob;
};

static void *udpworker(void *ctx)
{
	int len;
	int queued;
	int total;
	int s;
	int x;
	int start;
	socklen_t sl;
	unsigned int id;
	unsigned long long stamp;
	uint64_t dummy;
	void *h;
	struct udpctx *c=ctx;
	struct cmsghdr *cm;
	struct list *e;
	struct list *head;
	struct list *tail;
	struct in6_pktinfo *p6;
	struct in_pktinfo *p4;
	struct pollfd p[4];
	struct msghdr mh;
	struct iovec io;
	struct sockaddr_in6 a6;
	struct sockaddr_in6 l6;
	struct sockaddr_in6 r6;
	struct tuple6 t6;
	struct sockaddr_in a4;
	struct sockaddr_in l4;
	struct sockaddr_in r4;
	struct tuple4 t4;
	struct itimerspec it;
	struct peer peer;
	char cmsg[256];
	struct list *lookup[8192];
	unsigned char bfr[65536];
	sigset_t set;

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,NULL);
#ifdef DEBUG
	prctl(PR_SET_NAME,"udpworker",0,0,0);
#endif

	if(!(h=crypto_thread_init(c->glob)))goto out;

	start=c->portmin;
	total=0;
	queued=0;
	stamp=0;
	head=NULL;
	tail=NULL;
	memset(lookup,0,sizeof(lookup));
	memset(&it,0,sizeof(it));

	p[0].fd=c->efd;
	p[0].events=POLLIN;
	p[1].fd=c->rfd;
	p[1].events=POLLIN;
	p[2].fd=c->tfd;
	p[2].events=POLLIN;
	p[3].fd=c->fd;
	p[3].events=POLLIN;

	while(1)
	{
		if(poll(p,4,-1)<1)continue;

		if(p[0].revents&POLLIN)break;

		if(p[1].revents&POLLIN)if(read(p[1].fd,&dummy,sizeof(dummy))==
			sizeof(dummy))if((queued-=(int)dummy)<0)queued=0;

		if(p[2].revents&POLLIN)if(read(p[2].fd,&dummy,sizeof(dummy))==
			sizeof(dummy))
		{
			stamp+=(unsigned long long)dummy;
			while(head&&head->stamp<=stamp)
			{
				e=head;
				head=e->next;
				if(head)head->prev=NULL;
				else tail=NULL;
				x=e->id&0x3ff;
				if(lookup[x]==e)
				{
					lookup[x]=e->right;
					if(lookup[x])lookup[x]->left=NULL;
				}
				else
				{
					if(e->right)e->right->left=e->left;
					if(e->left)e->left->right=e->right;
				}
				free(e);
				if(!--total)
				{
					stamp=0;
					it.it_value.tv_nsec=
						it.it_interval.tv_nsec=0;
					timerfd_settime(c->tfd,0,&it,NULL);
					dummy=read(p[2].fd,&dummy,
						sizeof(dummy));
				}
			}
		}

		if(!(p[3].revents&POLLIN))continue;

		mh.msg_name=(c->family==AF_INET?(void *)&a4:(void *)&a6);
		mh.msg_namelen=(c->family==AF_INET?sizeof(a4):sizeof(a6));
		mh.msg_iov=&io;
		mh.msg_iovlen=1;
		mh.msg_control=cmsg;
		mh.msg_controllen=sizeof(cmsg);
		mh.msg_flags=0;
		io.iov_base=bfr;
		io.iov_len=sizeof(bfr);
		p4=NULL;
		p6=NULL;

		if((len=recvmsg(p[3].fd,&mh,MSG_NOSIGNAL|MSG_WAITALL))<=0)
			continue;

		if(c->family==AF_INET)
		{
		  for(cm=CMSG_FIRSTHDR(&mh);cm;cm=CMSG_NXTHDR(&mh,cm))
		    if(cm->cmsg_level==IPPROTO_IP)
		       if(a4.sin_family==AF_INET&&cm->cmsg_type==IP_PKTINFO)
			    p4=(struct in_pktinfo *)CMSG_DATA(cm);
		}
		else for(cm=CMSG_FIRSTHDR(&mh);cm;cm=CMSG_NXTHDR(&mh,cm))
		    if(cm->cmsg_level==IPPROTO_IPV6)
		       if(a6.sin6_family==AF_INET6&&cm->cmsg_type==IPV6_PKTINFO)
			    p6=(struct in6_pktinfo *)CMSG_DATA(cm);
		if(!p4&&!p6)continue;

		if(c->family==AF_INET)
		{
			t4.src=a4.sin_addr;
			t4.dst=p4->ipi_addr;
			t4.sport=a4.sin_port;
			t4.dport=c->port;
			id=hash_ipv4(t4.src.s_addr,t4.dst.s_addr,t4.sport,
				t4.dport);
		}
		else
		{
			t6.src=a6.sin6_addr;
			t6.dst=p6->ipi6_addr;
			t6.sport=a6.sin6_port;
			t6.dport=c->port;
			id=hash_ipv6(&t6.src,&t6.dst,t6.sport,t6.dport);
		}

		x=id&0x3ff;
		for(e=lookup[x];e;e=e->right)if(e->id==id)
		    if((c->family==AF_INET&&!memcmp(&e->t4,&t4,sizeof(t4)))||
			(c->family==AF_INET6&&!memcmp(&e->t6,&t6,sizeof(t6))))
		{
			if(e!=tail)
			{
				if(e->next)e->next->prev=e->prev;
				if(e->prev)e->prev->next=e->next;
				if(tail)tail->next=e;
				e->next=NULL;
				e->prev=tail;
				tail=e;
			}
			e->stamp=stamp+100;
			goto found;
		}

		if(!(e=malloc(sizeof(struct list))))continue;
		{
			e->stamp=stamp+100;
			e->id=id;
			if(c->family==AF_INET)e->t4=t4;
			else e->t6=t6;
			e->next=NULL;
			e->prev=tail;
			if(tail)tail->next=e;
			else head=e;
			tail=e;
			e->right=lookup[x];
			e->left=NULL;
			if(lookup[x])lookup[x]->left=e;
			lookup[x]=e;
			if(!total++)
			{
				it.it_value.tv_nsec=it.it_interval.tv_nsec=
					10000000;
				timerfd_settime(c->tfd,0,&it,NULL);
			}
		}

found:		if(queued==MAXQUEUE)continue;

		if(c->family==AF_INET)
		{
			memset(&l4,0,sizeof(l4));
			l4.sin_family=AF_INET;
			l4.sin_addr=p4->ipi_addr;
			l4.sin_port=0;
		}
		else
		{
			memset(&l6,0,sizeof(l6));
			l6.sin6_family=AF_INET6;
			l6.sin6_addr=p6->ipi6_addr;
			l6.sin6_port=0;
		}

		if((s=socket(c->family,SOCK_DGRAM|SOCK_CLOEXEC|SOCK_NONBLOCK,
			c->lite?IPPROTO_UDPLITE:0))==-1)continue;
		if(c->family==AF_INET6)
		{
			x=0;
			if(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
				goto err1;
		}
		if(c->sendbuf)if(setsockopt(s,SOL_SOCKET,SO_SNDBUF,&c->sendbuf,
			sizeof(c->sendbuf)))goto err1;
		if(c->lite)
		{
			if(setsockopt(s,IPPROTO_UDPLITE,UDPLITE_SEND_CSCOV,
				&c->lite,sizeof(c->lite)))goto err1;
			if(setsockopt(s,IPPROTO_UDPLITE,UDPLITE_RECV_CSCOV,
				&c->lite,sizeof(c->lite)))goto err1;
		}
		x=IP_PMTUDISC_DONT;
		if(c->frag)if(setsockopt(s,IPPROTO_IPV6,IPV6_MTU_DISCOVER,&x,
			sizeof(x)))goto err1;
		if(start)
		{
			for(x=start;;)
			{
				if(++x==c->portmax)x=c->portmin;
				if(c->rtp&&(x&1))goto skip;
				if(c->family==AF_INET)l4.sin_port=htons(x);
				else l6.sin6_port=htons(x);
				if(!bind(s,(struct sockaddr *)
					(c->family==AF_INET?
					(void *)&l4:(void *)&l6),
					c->family==AF_INET?
					sizeof(l4):sizeof(l6)))break;
				if(c->family==AF_INET)l4.sin_port=0;
				else l6.sin6_port=0;
skip:				if(x==start)break;
			}
			if(c->family==AF_INET)
			{
				if(!l4.sin_port)continue;
			}
			else if(!l6.sin6_port)continue;
			else start=x;
		}
		else if(bind(s,(struct sockaddr *)(c->family==AF_INET?
			(void *)&l4:(void *)&l6),c->family==AF_INET?
			sizeof(l4):sizeof(l6)))goto err1;
		if(connect(s,(struct sockaddr *)(c->family==AF_INET?
			(void *)&a4:(void *)&a6),c->family==AF_INET?
			sizeof(a4):sizeof(a6)))goto err1;
		sl=(c->family==AF_INET?sizeof(r4):sizeof(r6));
		if(getsockname(s,(struct sockaddr *)(c->family==AF_INET?
			(void *)&r4:(void *)&r6),&sl)||sl!=(c->family==AF_INET?
			sizeof(r4):sizeof(r6)))goto err1;

		if(crypto_portreply(h,bfr,&len,c->family==AF_INET?
			&r4.sin_port:&r6.sin6_port,c->lite?69:26,c->alt))
				goto err1;

		mh.msg_name=(c->family==AF_INET?(void *)&a4:(void *)&a6);
		mh.msg_namelen=(c->family==AF_INET?sizeof(a4):sizeof(a6));
		mh.msg_iov=&io;
		mh.msg_iovlen=1;
		mh.msg_control=cmsg;
		mh.msg_controllen=CMSG_SPACE((c->family==AF_INET?
			sizeof(struct in_pktinfo):sizeof(struct in6_pktinfo)));
		mh.msg_flags=0;
		io.iov_base=bfr;
		io.iov_len=len;
		cm=CMSG_FIRSTHDR(&mh);
		cm->cmsg_level=(c->family==AF_INET?IPPROTO_IP:IPPROTO_IPV6);
		cm->cmsg_type=(c->family==AF_INET?IP_PKTINFO:IPV6_PKTINFO);
		cm->cmsg_len=CMSG_LEN((c->family==AF_INET?
			sizeof(struct in_pktinfo):sizeof(struct in6_pktinfo)));
		if(c->family==AF_INET)
		{
			p4=(struct in_pktinfo *)CMSG_DATA(cm);
			memset(p4,0,sizeof(struct in_pktinfo));
			p4->ipi_spec_dst=l4.sin_addr;
		}
		else
		{
			p6=(struct in6_pktinfo *)CMSG_DATA(cm);
			memset(p6,0,sizeof(struct in6_pktinfo));
			p6->ipi6_addr=l6.sin6_addr;
		}

		if(sendmsg(p[3].fd,&mh,MSG_NOSIGNAL|MSG_EOR)!=len)goto err1;

		peer.fd=s;
		peer.rfd=c->rfd;

		if(write(c->pfd[1],&peer,sizeof(peer))!=sizeof(peer))goto err1;
		queued++;
		continue;

err1:		close(s);
	}

	while(read(c->pfd[0],&peer,sizeof(peer))==sizeof(peer))close(peer.fd);

	while(head)
	{
		e=head;
		head=e->next;
		free(e);
	}

	crypto_thread_exit(h);

out:	pthread_exit(NULL);
}

COLD void *udp_listen_init(void *addr,int port,int portmin,int portmax,
	int sendbuf,int lite,int frag,int rtp,int alt,int efd,void *glob)
{
	int fd;
	int x;
	int family;
	struct udpctx *c;
	union
	{
		struct sockaddr_in a4;
		struct sockaddr_in6 a6;
	} a;

	if(port<1||port>65535)goto err1;
	memset(&a,0,sizeof(a));
	if(addr)
	{
		if(inet_pton(AF_INET,addr,&a.a4.sin_addr)==1)
		{
			family=AF_INET;
			a.a4.sin_family=AF_INET;
			a.a4.sin_port=htons(port);
		}
		else if(inet_pton(AF_INET6,addr,&a.a6.sin6_addr)==1)
		{
			family=AF_INET6;
			a.a6.sin6_family=AF_INET6;
			a.a6.sin6_port=htons(port);
		}
		else goto err1;
	}
	else
	{
		family=AF_INET6;
		a.a6.sin6_family=AF_INET6;
		a.a6.sin6_addr=in6addr_any;
		a.a6.sin6_port=htons(port);
	}
	if((fd=socket(family,SOCK_DGRAM|SOCK_CLOEXEC|SOCK_NONBLOCK,
		lite?IPPROTO_UDPLITE:0))==-1)goto err1;
	if(family==AF_INET6)
	{
		x=(addr?1:0);
		if(setsockopt(fd,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
			goto err2;
		x=1;
		if(setsockopt(fd,IPPROTO_IPV6,IPV6_RECVPKTINFO,&x,sizeof(x)))
			goto err2;
	}
	else
	{
		x=1;
		if(setsockopt(fd,IPPROTO_IP,IP_PKTINFO,&x,sizeof(x)))goto err2;
	}
	if(sendbuf)if(setsockopt(fd,SOL_SOCKET,SO_SNDBUF,&sendbuf,
		sizeof(sendbuf)))goto err2;
	if(lite)
	{
		if(setsockopt(fd,IPPROTO_UDPLITE,UDPLITE_SEND_CSCOV,&lite,
			sizeof(lite)))goto err2;
		if(setsockopt(fd,IPPROTO_UDPLITE,UDPLITE_RECV_CSCOV,&lite,
			sizeof(lite)))goto err2;
	}
	x=IP_PMTUDISC_DONT;
	if(frag)if(setsockopt(fd,IPPROTO_IPV6,IPV6_MTU_DISCOVER,&x,sizeof(x)))
		goto err2;
	if(bind(fd,(struct sockaddr *)&a,family==AF_INET?
		sizeof(struct sockaddr_in):sizeof(struct sockaddr_in6)))
			goto err2;
	if(!(c=malloc(sizeof(struct udpctx))))goto err2;
	c->fd=fd;
	c->efd=efd;
	if((c->tfd=timerfd_create(CLOCK_MONOTONIC,TFD_CLOEXEC|TFD_NONBLOCK))
		==-1)goto err3;
	if((c->rfd=eventfd(0,EFD_CLOEXEC|EFD_NONBLOCK))==-1)goto err4;
	if(pipe2(c->pfd,O_CLOEXEC|O_NONBLOCK|O_DIRECT))goto err5;
	c->family=family;
	c->sendbuf=sendbuf;
	c->lite=lite;
	c->alt=alt;
	c->frag=frag;
	c->rtp=rtp;
	c->portmin=portmin;
	c->portmax=portmax;
	if(family==AF_INET)c->port=a.a4.sin_port;
	else c->port=a.a6.sin6_port;
	c->glob=glob;
	return c;

err5:	close(c->rfd);
err4:	close(c->tfd);
err3:	free(c);
err2:	close(fd);
err1:	return NULL;
}

COLD int udp_listen_start(void *ctx)
{
	struct udpctx *c=ctx;

	if(pthread_create(&c->th,NULL,udpworker,ctx))return -1;
	return c->pfd[0];
}

COLD void udp_listen_stop(void *ctx,int force)
{
	struct udpctx *c=ctx;

	if(force)pthread_cancel(c->th);
	pthread_join(c->th,NULL);
}

COLD void udp_listen_exit(void *ctx)
{
	struct udpctx *c=ctx;

	close(c->fd);
	close(c->pfd[0]);
	close(c->pfd[1]);
	close(c->tfd);
	close(c->rfd);
	free(c);
}

int udp_accept(int fd)
{
	struct peer p;
	uint64_t dummy=1;

	if(read(fd,&p,sizeof(p))!=sizeof(p))return -1;
	dummy=write(p.rfd,&dummy,sizeof(dummy));
	return p.fd;
}

int udp_connect(void *h,void *dst,int timeout,int sendbuf,int lite,int frag,
	int rtp,int alt,int minport,int maxport,int efd)
{
	int slen;
	int rlen;
	int family;
	int s;
	int x;
	int d;
	int n;
	socklen_t sl;
	struct pollfd p[2];
	union
	{
		struct sockaddr_in a4;
		struct sockaddr_in6 a6;
	} r;
	unsigned char sndbfr[512];
	unsigned char rcvbfr[512];

	family=((struct sockaddr_in *)dst)->sin_family;
	if((s=socket(family,SOCK_DGRAM|SOCK_CLOEXEC|SOCK_NONBLOCK,
		lite?IPPROTO_UDPLITE:0))==-1)goto err1;
	if(family==AF_INET6)
	{
		x=1;
		if(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
			goto err2;
	}
	x=1;
	if(sendbuf)if(setsockopt(s,SOL_SOCKET,SO_SNDBUF,&sendbuf,
		sizeof(sendbuf)))goto err2;
	if(lite)
	{
		if(setsockopt(s,IPPROTO_UDPLITE,UDPLITE_SEND_CSCOV,&lite,
			sizeof(lite)))goto err2;
		if(setsockopt(s,IPPROTO_UDPLITE,UDPLITE_RECV_CSCOV,&lite,
			sizeof(lite)))goto err2;
	}
	x=IP_PMTUDISC_DONT;
	if(frag)if(setsockopt(s,family==AF_INET6?IPPROTO_IPV6:IPPROTO_IP,
		family==AF_INET6?IPV6_MTU_DISCOVER:IP_MTU_DISCOVER,&x,
		sizeof(x)))goto err2;

	if(minport)
	{
		if(crypto_random(h,&d,sizeof(d)))goto err2;
		d&=0xffff;
		d%=maxport-minport;
		memset(&r,0,sizeof(r));
		r.a4.sin_family=family;
		for(x=minport;x<maxport;x++)
		{
			if((n=x+d)>=maxport)n-=maxport-minport;
			if(rtp&&(n&1))continue;
			if(family==AF_INET)r.a4.sin_port=htons(n);
			else r.a6.sin6_port=htons(n);
			if(!bind(s,(struct sockaddr *)&r,sizeof(r)))break;
		}
		if(x==maxport)goto err2;
	}
	if(crypto_portrequest(h,sndbfr,&slen,lite?69:24))goto err2;

	if(sendto(s,sndbfr,slen,MSG_NOSIGNAL|MSG_EOR,dst,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in))<=0)
		goto err2;
	p[0].fd=efd;
	p[0].events=POLLIN;
	p[1].fd=s;
	p[1].events=POLLIN;
	sl=sizeof(r);
	if(poll(p,2,timeout*1000)<1)goto err2;
	if(p[0].revents&POLLIN)goto err2;
	if(!(p[1].revents&POLLIN))goto err2;
	if((rlen=recvfrom(s,rcvbfr,sizeof(rcvbfr),MSG_NOSIGNAL|MSG_WAITALL,
		(struct sockaddr *)&r,&sl))<=0)goto err2;
	else if(family==AF_INET6)
	{
		if(sl<sizeof(struct sockaddr_in6))goto err2;
		if(r.a6.sin6_port!=((struct sockaddr_in6 *)dst)->sin6_port)
			goto err2;
		if(memcmp(r.a6.sin6_addr.s6_addr,
			((struct sockaddr_in6 *)dst)->sin6_addr.s6_addr,
			sizeof(r.a6.sin6_addr.s6_addr)))goto err2;
		if(crypto_portresponse(h,sndbfr,slen,rcvbfr,rlen,
			&r.a6.sin6_port,alt))goto err2;
	}
	else
	{
		if(sl<sizeof(struct sockaddr_in))goto err2;
		if(r.a4.sin_port!=((struct sockaddr_in *)dst)->sin_port)
			goto err2;
		if(memcmp(&r.a4.sin_addr.s_addr,
			&((struct sockaddr_in *)dst)->sin_addr.s_addr,
			sizeof(r.a4.sin_addr.s_addr)))goto err2;
		if(crypto_portresponse(h,sndbfr,slen,rcvbfr,rlen,
			&r.a4.sin_port,alt))goto err2;
	}

	if(connect(s,(struct sockaddr *)&r,family==AF_INET6?
		sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in)))
		goto err2;
	return s;

err2:	close(s);
err1:	return -1;
}

void udpclose(int fd)
{
	close(fd);
}

HOT int udpsend(int fd,void *data,int len)
{
	int l;

	if((l=send(fd,data,len,MSG_NOSIGNAL|MSG_EOR))==-1)switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	else if(l!=len)return -1;
	return l;
}

HOT int udprecv(int fd,void *data,int max)
{
	int l;

	if((l=recv(fd,data,max,MSG_NOSIGNAL|MSG_WAITALL))==1)switch(errno)
	{
	case EINTR:
	case EAGAIN:
		return 0;
	default:return -1;
	}
	if(!l)return -2;
	return l;
}
